--#****************************************************************************
--#**
--#**  New File :  /mods/gilbotsimprovements/ACUcommon.lua
--#**
--#**  Author(s):  Gilbot-X
--#**
--#**  Summary  :  Common ACU Changes Script
--#**
--#****************************************************************************

local GilbotUtils = import('/mods/gilbotsimprovements/utils.lua')

--# This flag sets which log messages are output
local debugAutoToggle = false
local debugAdjacencyCode = false


--#*
--#*  Gilbot-X says:
--#*      
--#*  Use this as a secondary base class in the ACU script files.
--#*  I made this class so I could add extra functionality to all 3 ACUs 
--#*  without mixing my code in the same file as GPG's code. 
--#** 
ACUKnowledge = Class {    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  Call this from an override of OnCreate so I can
    --#*  do variable initialization common to all 3 ACUs at this time.
    --#** 
    InitializeKnowledge = function(self)

        local myArmyId = self:GetArmy()
        local myExpectedCommanderEntityId = (myArmyId -1) * 1048576
        local myEntityId = self:GetEntityId()
        
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        --# I use it to prove that commander entity IDs 
        --# are always a multiple of 1048576.
        --# This was the case up to version 3269.
        if false then
          LOG('ACUcommon.lua: Creating Commander for ' 
           .. ' Army= ' .. repr(myArmyId)
           .. ' with entityid=' .. repr(myEntityId)
          )
        end
        
        --# This warning will alert me as a programmer
        --# that the version of Supreme Commander I am running on has 
        --# a new system for allocating entity IDs and that I must
        --# update my method of getting a reference to the ACU.        
        if myEntityId ~= repr(myExpectedCommanderEntityId) then
          WARN('ACUcommon.lua: Commander entityID for ' 
          .. ' Army= ' .. repr(myArmyId)
          .. ' is ' .. myEntityId
          .. ' but expected ' .. repr(myExpectedCommanderEntityId)
          )     
        end
          
        --# Code to initialize Resource Network system 
        --# was separated into its own function
        --# to make code easier to maintain.
        self:InitializeResourceNetworkSystem()
        
        --# Code to initialize auto-toggle system 
        --# was separated into its own function
        --# to make code easier to maintain.
        self:InitializeAutoToggleSystem()
        
    end,
    
   
   
   
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Had to override this to make sure that tables are 
    --#*  cleaned up when a unit completes an upgrade and 
    --#*  destroys itself.
    --#*
    --#*  This function is defined in Unit.lua
    --#*  and ends by changing state to DeadState
    --#*  so I have to call my extra code before calling
    --#*  the base class code.
    --#**
    CallBefore_OnDestroy = function(self)
        --# Gilbot-X says:  
        --# I added this block of code so that when 
        --# units upgrade they get cleaned out of ResourceNetwork 
        --# tables and the ACU's auto-toggle tables.
        self:NoMoreResourceNetworks() 
    end,
    
    
   
   
   
   
   
--#-------------------------------------------------------------
--#  ADJACENCY NETWORK SPECIFIC CODE
--#-------------------------------------------------------------
 
    
        
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This is called from InitializeKnowledge(), which itself is  
    --#*  called from an override of the ACU's Unit:OnCreate(), so I can
    --#*  do variable initialization common to all 3 ACUs at this time.
    --#** 
    InitializeResourceNetworkSystem = function(self)
        --# Make this value globally available to other scripts in my mod.
        self.MyMaxAdjacencyRangeEncounteredSoFar = 0
        
        --# Set this to a T1 Powergen or MeX
        self.MyMaxSkirtXFromCentre = 1
        self.MyMaxSkirtZFromCentre = 1
        
        --# ACU knows about which networks are running.
        self.NetworkNumber =0
        self.ActiveNetworks = {}
        
        --# ACU knows about which inter-networks are running.
        self.InterNetworkNumber =0
        self.ActiveInterNetworks = {}
        
        --# This variable is used to toggle
        --# if the priority list position numbers 
        --# (set as custom names) are shown.
        --# Start with display switched off.
        self.NetworkNumbersDisplayed = false
    end,
    
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to be called by structure units
    --#*  to update the commander (thus the army) that
    --#*  a new adjacency range is available, so the commamder
    --#*  should now try to link units that are further apart.
    --#** 
    ReceiveACUMessage_SetMaxAdjacencyRangeEncounteredSoFar = function(
                                                      self, 
                                                      maxAdjacencyRangeEncounteredSoFarArg)
        self.MyMaxAdjacencyRangeEncounteredSoFar = 
            math.max(
                maxAdjacencyRangeEncounteredSoFarArg,
                self.MyMaxAdjacencyRangeEncounteredSoFar
            )
    end,
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to be called by structure units
    --#*  to ask the commander (thus the army) what
    --#*  the max adjacency range available is, so the commander
    --#*  should now try to link units that are this far apart.
    --#** 
    ReceiveACUMessage_GetMaxAdjacencyRangeEncounteredSoFar = function(self)
        return self.MyMaxAdjacencyRangeEncounteredSoFar
    end,
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to be called by structure units
    --#*  to update the commander (thus the army) that
    --#*  a new adjacency range is available, so the commamder
    --#*  should now try to link units that are further apart.
    --#** 
    ReceiveACUMessage_SetMaxSkirtBoundsEncounteredSoFar = function(self, 
                                                    maxSkirtXFromCentreArg, 
                                                    maxSkirtZFromCentreArg)
          self.MyMaxSkirtXFromCentre = 
              math.max(
                  maxSkirtXFromCentreArg,
                  self.MyMaxSkirtXFromCentre
              )
              
          self.MyMaxSkirtZFromCentre = 
              math.max(
                  maxSkirtZFromCentreArg,
                  self.MyMaxSkirtZFromCentre
              )
    end,
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to be called by structure units
    --#*  to ask the commander (thus the army) what
    --#*  the max adjacency range available is, so the commander
    --#*  should now try to link units that are this far apart.
    --#** 
    ReceiveACUMessage_GetMaxSkirtBoundsEncounteredSoFar = function(self)
        return self.MyMaxSkirtXFromCentre, self.MyMaxSkirtZFromCentre 
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to be called by ResourceNetwork objects
    --#*  when created to inform the ACU (thus the army) 
    --#*  a new network has been created.
    --#** 
    ReceiveACUMessage_AddNewNetwork = function(self, networkArg)
        self.NetworkNumber = self.NetworkNumber+1
        networkArg.NetworkId = self.NetworkNumber
        --# Keep a reference to the network object
        self.ActiveNetworks[self.NetworkNumber] = networkArg
        
        --# If we were already displaying numbers for 
        --# other networks, then display for this new one too.
        if self.NetworkNumbersDisplayed then 
            networkArg:DisplayNetworkNumbers(true) 
        end
    end,
        
        
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to be called by ResourceInterNetwork objects
    --#*  when created to inform the ACU (thus the army) 
    --#*  a new remote-linked inter-network has been created.
    --#** 
    ReceiveACUMessage_AddNewInterNetwork = function(self, interNetworkArg)
        self.InterNetworkNumber = self.InterNetworkNumber+1
        interNetworkArg.InterNetworkId = self.InterNetworkNumber
        --# Keep a reference to the network object
        self.ActiveInterNetworks[self.InterNetworkNumber] = interNetworkArg
    end,
    
    
        
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to be called by structure units
    --#*  to tell the ACU to dump a list of network srunning.
    --#*  It is for debugging only.
    --#** 
    ReceiveACUMessage_DumpActiveNetworks = function(self)
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if debugAdjacencyCode then
            local armyString = repr(self:GetArmy())
            LOG('Adjacency: a=' .. armyString)
            LOG('Adjacency: a=' .. armyString 
            .. ' ###### Unbroken networks: Start: ###############################')
            for k, vNetwork in self.ActiveNetworks do
                if not vNetwork.IsBroken then
                    vNetwork:DumpNetwork()
                end
            end
            LOG('Adjacency: a=' .. armyString 
            .. ' ###### Unbroken networks: End: #################################')
        end
    end,
    
   
   
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  I added this to be called by the UI via a Sim callback
    --#*  to inform the ACU to toggle display of priority list numbers
    --#** 
    ReceiveACUMessage_ToggleNetworkDisplay = function(self)
        --# If it was toggled on...  then toggle it off...
        --# If it was toggled off... then toggle it back on...
        self.NetworkNumbersDisplayed = not self.NetworkNumbersDisplayed
        
        --# Now deal with new value of toggle.
        if self.NetworkNumbersDisplayed and self.AutoToggleNumbersDisplayed then 
            self.AutoToggleNumbersDisplayed = false
            --# Update displays on all listed auto-toggle units
            self:DisplayAutoToggleNumbers(self.AutoToggleNumbersDisplayed)
        end
        
        --# Defer update to this local function 
        self:DisplayNetworkNumbers(self.NetworkNumbersDisplayed)
    end,  
        


        
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This should be called when the ACU dies so it 
    --#*  is obvious that autotoggle has been disabled on all units.
    --#*  This is safe to be called more than once.
    --#** 
    DisplayNetworkNumbers = function(self, toggleOn)    
        --# Defer call to each active network.
        for kNetworkId, vNetwork in self.ActiveNetworks do
            if not vNetwork.IsBroken then
                vNetwork:DisplayNetworkNumbers(toggleOn)
            end
        end
    end,
    
    
    

    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This should be called when the ACU dies so it 
    --#*  is obvious that autotoggle has been disabled on all units.
    --#** 
    NoMoreResourceNetworks = function(self)    
        
        --# iterate through all inter-networks and break them up.
        for k, vInterNetwork in self.ActiveInterNetworks do
            if not vInterNetwork.IsBroken then
                vInterNetwork.IsBroken = true
            end
        end
        
        --# iterate through all networks and break them up.
        for k, vNetwork in self.ActiveNetworks do
            if not vNetwork.IsBroken then
                vNetwork:BreakNetwork()
            end
        end
        
        --# Make tables totally safe
        self.ActiveNetworks = {}
        self.ActiveInterNetworks = {}
    end,
   
   
  
    
   
   
   
   
   
   
--#-------------------------------------------------------------
--#  AUTO TOGGLE SPECIFIC CODE
--#-------------------------------------------------------------
   
   
    --# This effects how often ACU will check auto-toggle.
    --# It refers to wait time measured in seconds between checks.
    CyclePeriod = 2,
   
   
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This is called from InitializeKnowledge(), which itself is  
    --#*  called from an override of the ACU's Unit:OnCreate(), so I can
    --#*  do variable initialization common to all 3 ACUs at this time.
    --#** 
    InitializeAutoToggleSystem = function(self)
        
        --# Three tables, one for each tech level.
        self.AutoToggleTables = {
            {{},{},{},{}},  --MassFabs
            {{},{},{},{}},  --P2 for units that only get switched on when energy ratio storage is high
            {{},{},{},{}},  --P3 for Intel
            {{},{},{},{}},  --P4 for Stealth/Cloak
            {{},{},{},{}},  --P5 for Mobile Shields
            {{},{},{},{}},  --P6 for Shield Structures
        }
        
        --# This table records which
        --# priority classes of autotoggle units are 
        --# allowed to be switched on.
        self.AllowToBeOn = {
            true, 
            true, 
            true,
            true, 
            true, 
            true
        }
        
        --# Initialise priority list used by
        --# switch on/off cycles.
        self.AutoToggleOffPriorityList = {}
        
        --# This variable is used to toggle
        --# if the priority list position numbers 
        --# (set as custom names) are shown.
        --# Start with display switched off.
        self.AutoToggleNumbersDisplayed = false
            
        --# Launch thread to auto switch on or off
        self:ForkThread(self.AutoToggleManagementThread, self) 
        
    end, 
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This thread periodically checks if this unit should be on or off.
    --#*  The check itself is defered to code in CheckAutoPowerConditions.
    --#** 
    AutoToggleManagementThread = function(self) 
    
        --# Energy bands are a way of recording
        --# what level of energy we are in
        --# and checking if we are making a transition
        --# from one band to another.
        --# Let 4 stand for excess amounts of energy,
        --# Let 3 stand for safe amount of energy but not excess
        --# Let 2 stand for basic amount of energy to keep in balance
        --# Let 1 stand for almost empty and not safe.
        --# When the ACU is created at game start, 
        --# we have full energy and positive energy trend,
        --# so set the band to 4 as we have excess energy.
        self.PreviousEnergyBand = 4
        --# When thread launches assume we already had full mass
        self.PreviousMassStorageRatio = 1
        
        --# Keep doing this while ACU is alive
        while (not self:BeenDestroyed()) or (not self:IsDead()) do
            
            --# Get a snapshot of the economy
            local aiBrain = self:GetAIBrain() 
            local currentMassStorageRatio = aiBrain:GetEconomyStoredRatio('MASS')
            local energyStorageRatio = aiBrain:GetEconomyStoredRatio('ENERGY')
            local energyTrend = aiBrain:GetEconomyTrend('ENERGY') * 10
            local energyStored = aiBrain:GetEconomyStored('ENERGY')
    
            --# If we have full storage for mass, but we didn't before,
            --# then don't allow any to be on.  We don't want to waste 
            --# energy on massfabs or even let economy indicators show 
            --# income with wasteful massfabs on.
            if currentMassStorageRatio == 1 and 
               self.PreviousMassStorageRatio ~=1 then
                --# This makes sure that mass fabs cannot be switched on
                self:AllowUnitsToBeOn(1, false)
                
            --# If we had full storage for mass but now we don't, then..
            elseif currentMassStorageRatio ~= 1 and 
              self.PreviousMassStorageRatio ==1 then
              --# This makes sure that mass fabs can be switched on
                self:AllowUnitsToBeOn(1, true)
            end
            
            
            --# Check which energy band we are in for this cycle
            local currentEnergyBand = nil
            
            --# If energy is pretty much full and trend is positive then
            if energyTrend > 0 and energyStorageRatio > 0.9 then
                --# We have full storage and positive trend so
                --# we need to use some excess energy.
                --# We should turn anything on that we can.
                currentEnergyBand = 4
            
            --# If energy will run out in next 2 cycles ...   
            elseif energyStored + (energyTrend * self.CyclePeriod * 2) < 0 then
                --# We have almost empty storage and a negative trend so
                --# we must switch off the next least essential unit
                currentEnergyBand = 1
            
            --# If energy is increasing and not about to end            
            elseif energyTrend > 0 and energyStorageRatio > 0.2 then
                --# We have almost empty storage but a positive trend so
                --# we can switch on some essential units that are
                --# important to have on as much as possible.
                currentEnergyBand = 3
            
            --# Energy is not near full and decreasing, 
            --# or it is increasing but still almost empty
            else
                --# This is a safe equilibrium state.
                --# We don't want to do anything.
                currentEnergyBand = 2
            end
            
            
            --# Some units are such low priority that
            --# they are only switched on to use up excess energy
            --# when energy storage is pretty much full and 
            --# we have a positive energy trend.
            
            --# If we move into band 4, enable that class of units
            if currentEnergyBand == 4 and self.PreviousEnergyBand ~=4 then 
                self:AllowUnitsToBeOn(2, true)
            end
            --# If we move out of band 4, disable that class of units
            if self.PreviousEnergyBand == 4 and currentEnergyBand ~= 4 then 
                self:AllowUnitsToBeOn(2, false)
            end
            
            
            --# Now decide if a unit muist be toggled on or off, or neither.
            --# If we are gaining energy and we have at least 10% of our storage...
            if currentEnergyBand > 2 then
                --# Turn the next highest priority thing 
                --# back on that won't totally empty out 
                --# our energy reserves in 10 seconds.
                self:PerformCleanCycle(self.PerformToggleOnCycle)
            
            --# If energy will run out in next 4 seconds...                   
            elseif currentEnergyBand == 1 then
                --# Turn 1 thing off at next lowest priority.
                self:PerformCleanCycle(self.PerformToggleOffCycle)
            end
            
            --# Debugging only
            if debugAutoToggle and self.PreviousEnergyBand ~= currentEnergyBand then
              LOG('AutoToggle: Army=' .. repr(self:GetArmy()) 
              .. ' Energy band changed from ' .. repr(self.PreviousEnergyBand)
              .. ' to ' .. repr(currentEnergyBand)
              )
            end
            
            --# Record mass state for next time so we can compare before/after
            self.PreviousMassStorageRatio = currentMassStorageRatio
            self.PreviousEnergyBand = currentEnergyBand
            
            --# Delay between cycles needs to allow 
            --# for economy changes to take effect and 
            --# must reduce burden on CPU, but not be too
            --# slow so that auto-toggle is less desireable than
            --# micro-managing toggles.           
            WaitSeconds(self.CyclePeriod)
        end
        
        --# If the thread gets here, the ACU has died.  Clean up.
        self:NoMoreAutoToggle()
        
    end, 
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This should be called when the ACU dies so it 
    --#*  is obvious that autotoggle has been disabled on all units.
    --#*  This is safe to be called more than once.
    --#** 
    NoMoreAutoToggle = function(self)    
        --# Look at all types of registered auto-toggle units, 
        --# remove auto-toggle name/priority and extra toggle.
        for kUnitType, vAutoToggleUnitTechTables in self.AutoToggleTables do
            for kTechLevel, vAutoToggleUnitTable in vAutoToggleUnitTechTables do
                for kAutoToggleUnit, vAutoToggleUnit in vAutoToggleUnitTable do 
                    if not (vAutoToggleUnit:BeenDestroyed() or vAutoToggleUnit:IsDead()) then
                        --# Take off auto-toggle if it was on so
                        --# it is obvious that autotoggle is off.
                        if self.AutoToggleNumbersDisplayed then
                            vAutoToggleUnit:ClearCustomName()
                        end
                        --# Remove auto-toggle toggle for this unit
                        vAutoToggleUnit:RemoveToggleCap('RULEUTC_WeaponToggle')
                        vAutoToggleUnit:RequestRefreshUI()
                    end
                end
            end
        end
        --# Make it totally safe
        self.AutoToggleTables = {}
        self.AutoToggleOffPriorityList = {}
    end,
    
  
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  Just switch off all units in thsi table unconditionally
    --#*  or mark all of them as allowed to be switched back on.
    --#*  Note: Doesn't do any clean-up.
    --#** 
    AllowUnitsToBeOn = function(self, priorityClass, toggleCanBeOn) 
        
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if debugAutoToggle then
          LOG('AutoToggle: Army=' .. repr(self:GetArmy()) 
           .. ': AllowUnitsToBeOn: ' .. repr(toggleCanBeOn)
           .. ' on class=' .. repr(priorityClass)
          )
        end
    
        --# Record this change so we can set new units added to list 
        self.AllowToBeOn[priorityClass] = toggleCanBeOn
        
        --# Go through all such units registered at all tech levels....
        for kTechLevel, vAutoToggleUnitTable in self.AutoToggleTables[priorityClass] do
            for kUnusedArrayIndex, vAutoToggleUnit in vAutoToggleUnitTable do
                
                --# Make sure it's not dead (don't worry about cleaning up here)
                if not (vAutoToggleUnit:BeenDestroyed() or vAutoToggleUnit:IsDead()) then
                    --# Allow it to be on but don't do any switching on.
                    if toggleCanBeOn then 
                        vAutoToggleUnit.CanAutoToggleOnNow = true
                    else
                        --# switch it off and make sure it can't come back on
                        --# until this function is called again with the opposite
                        --# argument.                        
                        vAutoToggleUnit:SwitchOff()
                        vAutoToggleUnit.CanAutoToggleOnNow = false
                    end
                end 
            end
        end
    end,
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This is repeated code from AutoToggleManagementThread which 
    --#*  I separated into its own function for good code modularity.
    --#** 
    PerformCleanCycle = function(self, functionRef, optionalArg)
        
        --# Use AutoToggleUnitToCleanUp to record reference to any dead unit 
        --# that needs to be cleaned up from auto-toggle table
        self.AutoToggleUnitToCleanUp = true
        while self.AutoToggleUnitToCleanUp do
            
            --# Don't do another pass unless we have 
            --# to stop next one to do a table clean-up.
            self.AutoToggleUnitToCleanUp = false
            
            --# Do cycle which will traverse tables and 
            --# potentially examine all tables and unit entries
            functionRef(self, optionalArg)
            
            --# If a dead unit was found in one of the tables,
            --# it will have ben recorded and traverse exited prematurely.
            if self.AutoToggleUnitToCleanUp then
                --# Do the clean up, i.e. remove the unit from the table
                self:UnregisterAutoToggleEntry(self.AutoToggleUnitToCleanUp, true, true)
                --# Prevent another access to that unit now that its been cleaned up
                self.AutoToggleUnitToCleanUp = true
            end
        end
    end,
    
    
      
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  Turn a single unit off, chosen according to order in priority list.
    --#*
    --#*  This function must be executed via the PerformCleanCycle function
    --#*  so that when any dead entries are encountered, they are cleaned up,
    --#*  and this function will be called again to complete the job.
    --#** 
    PerformToggleOffCycle = function(self) 
  
        --# Go through each unit in priority list
        for kPriority, vAutoToggleUnit in self.AutoToggleOffPriorityList do  
            
            --# Check for dead units.  This happens occaisionally (concurrency issue).
            if vAutoToggleUnit:BeenDestroyed() or vAutoToggleUnit:IsDead() then
                --# This is used to defer clean up of dead table entry,
                --# i.e. a new priority list will be drawn up and then this
                --# function will be called again.
                self.AutoToggleUnitToCleanUp = vAutoToggleUnit
                
                --# This next block is for debugging only.
                --# This can be delete wheh debugging is done.
                if debugAutoToggle then
                  LOG('AutoToggle: PerformToggleOffCycle:' 
                  .. ' Dead unit ' .. vAutoToggleUnit.DebugId
                  .. ' found in ACU AutoToggleOffPriorityList. Stopping cycle to cleanup.'
                  )
                end
                
                --# Abandon this cycle to perform clean-up
                return false
            else
                --# Make sure the unit is actually switched on before trying to switch off...
                --# and don't let units auto-power off when under attack.
                if vAutoToggleUnit.On and not vAutoToggleUnit.UnderAttack then
                    --# Turn the unit off unconditionally
                    vAutoToggleUnit:SwitchOff() 
                    
                    --# Go back to beginning of priority list
                    --# which also gives time for this change 
                    --# on the economy to have an effect.
                    return true
                end
            end 
        end
 
        --# Return true to indicate that this 
        --# priority list is still valid.
        return true
 
    end, 
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  Turn a single unit on, chosen according to order in priority list.
    --#*
    --#*  As switching off priority uses the same ordered list as switching on priorty,
    --#*  then we have to trawl through the nested tables in the opposite order
    --#*  to how we did when switching units off.
    --#*
    --#*  This function must be executed via the PerformCleanCycle function
    --#*  so that when any dead entries are encountered, they are cleaned up,
    --#*  and this function will be called again to complete the job.
    --#** 
    PerformToggleOnCycle = function(self) 
    
        --# Get a snapshot of the economy
        local aiBrain = self:GetAIBrain() 
        local energyTrend = aiBrain:GetEconomyTrend('ENERGY') * 10
        local energyStored = aiBrain:GetEconomyStored('ENERGY')
    
        --# Go through each unit im priority list but in reverse order
        local reversedUnits = GilbotUtils.GetArrayReversed(self.AutoToggleOffPriorityList)
        for kPriority, vAutoToggleUnit in reversedUnits do       
                   
            --# This thread needs to die when the unit dies.
            if vAutoToggleUnit:BeenDestroyed() or vAutoToggleUnit:IsDead() then
                --# This is used to defer clean up of dead table entry
                --# i.e. a new priority list will be drawn up and then this
                --# function will be called again.
                self.AutoToggleUnitToCleanUp = vAutoToggleUnit
                
                --# This next block is for debugging only.
                --# This can be delete wheh debugging is done.
                if debugAutoToggle then
                  LOG('AutoToggle: PerformToggleOnCycle:' 
                  .. ' Dead unit ' .. vAutoToggleUnit.DebugId
                  .. ' found in ACU AutoToggleOffPriorityList. Stopping cycle to cleanup.'
                  )
                end
                
                --# Abandon this cycle to perform clean-up
                return false
            else     
                --# Make sure it's actually off 
                --# before trying to turn back on
                if (not vAutoToggleUnit.On) and 
                  --# This can be cleared, i.e. on Massfabs when there is full mass.
                  vAutoToggleUnit.CanAutoToggleOnNow and
                  --# Don't let units auto-power on or off when under attack.
                  (not vAutoToggleUnit.UnderAttack) then
                    
                    
                    --# What would be the rate per second of energy change 
                    --# if we switched this unit back on?
                    local potentialEnergyTrend = energyTrend - vAutoToggleUnit.EnergyUsage
                    --# How much energy would be left after 2 cycles 
                    --# if we switched this unit back on?
                    local energyLeftAfterTwoCycles = energyStored + 
                        (potentialEnergyTrend * self.CyclePeriod * 2)
                    --# If any energy would be left to stop our shields from cutting...
                    if energyLeftAfterTwoCycles > 0 then 
                        
                        --# Turn the unit on
                        vAutoToggleUnit:SwitchOn() 
                        
                        --# This next block is for debugging only.
                        --# This can be delete wheh debugging is done.
                        if false and debugAutoToggle then
                          LOG('AutoToggle: Army=' .. repr(self:GetArmy()) 
                              .. ': Switched ON: ' .. vAutoToggleUnit.DebugId
                          )
                        end
                          
                        --# Go back to beginning of priority list
                        --# which also gives time for this change 
                        --# on the economy to have an effect.
                        return true
                    else
                        --# What if we can turn off some lower priority units that we 
                        --# turned back on before while skipping this?
                        local energySaved = 
                            self:GetEnergySavedBySwitchingOffLowerPriorityUnits(vAutoToggleUnit)
                        --# What would be the rate per second of energy change of we did this?
                        potentialEnergyTrend =
                            (energyTrend + energySaved) - vAutoToggleUnit.EnergyUsage
                        --# How much energy would we then have left after 2 cycles if we did this?
                        energyLeftAfterTwoCycles = energyStored + 
                              (potentialEnergyTrend * self.CyclePeriod * 2)
                        if energyLeftAfterTwoCycles > 0 then 
                            --# Switch off the units we included in our count
                            self:SwitchOffLowerPriorityUnits(vAutoToggleUnit)
                            --# Turn the current (higher priority) unit on
                            vAutoToggleUnit:SwitchOn() 
                            
                            --# This next block is for debugging only.
                            --# This can be delete wheh debugging is done.
                            if false and debugAutoToggle then
                              LOG('AutoToggle: Army=' .. repr(self:GetArmy()) 
                                  .. ': Switched ON: ' .. vAutoToggleUnit.DebugId
                              )
                            end
                              
                            --# Go back to beginning of priority list
                            --# which also gives time for this change 
                            --# on the economy to have an effect.
                            return true
                        end
                    end
                end
            end
           
        end

        --# Return true to indicate that this 
        --# priority list is still valid.
        return true  
    end, 
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This is modularised code for PerformToggleOnCycle.
    --#*  It get units further down priority list than the unit supplied and sums 
    --#*  EnergyUsage so we can decide if we can afford to switch this back on
    --#*  on the condition that we switch off all those units that are 
    --#*  already back on lower down the priority list for being switched back on.    
    --#** 
    GetEnergySavedBySwitchingOffLowerPriorityUnits = function(self, fromUnitArg)
        
        local result = 0
        --# Traversing through the priority list in switching off order and
        --# counting those before us means that ..
        for kPriority, vAutoToggleUnit in self.AutoToggleOffPriorityList do 
            --# .. we only count units that are a lower priority to switch back on.
            --# Stop counting when we get to ourself in the unit list.
            if vAutoToggleUnit == fromUnitArg then 
                return result

            --# If the unit is on, add its energy usage
            --# to the amount we would save if we switched off
            --# all (cheaper) units that are a lower priority to switch back on.
            elseif vAutoToggleUnit.On and not vAutoToggleUnit.UnderAttack then
                result = result + vAutoToggleUnit.EnergyUsage
            end
        end

    end,
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This is modularised code for PerformToggleOnCycle.
    --#*  It get units further down priority list than the unit supplied and sums 
    --#*  EnergyUsage so we can decide if we can afford to switch this back on
    --#*  on the condition that we switch off all those units that are 
    --#*  already back on lower down the priority list for being switched back on.    
    --#** 
    SwitchOffLowerPriorityUnits = function(self, fromUnitArg)
        
        local result = 0
        --# Traversing through the priority list in switching off order and
        --# counting those before us means that ..
        for kPriority, vAutoToggleUnit in self.AutoToggleOffPriorityList do 
            --# .. we only count units that are a lower priority to switch back on.
            --# Stop counting when we get to ourself in the unit list.
            if vAutoToggleUnit == fromUnitArg then 
                return

            --# If the unit is on, add its energy usage
            --# to the amount we would save if we switched off
            --# all (cheaper) units that are a lower priority to switch back on.
            elseif vAutoToggleUnit.On and not vAutoToggleUnit.UnderAttack then
                vAutoToggleUnit:SwitchOff()
            end
        end

    end,
                        
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to be called by auto-toggle units
    --#*  from their own class when created, to inform the ACU 
    --#*  so they can be used by the ACU's Auto Power-down.
    --#** 
    ReceiveACUMessage_RegisterAutoToggle = function(self, autoToggleUnitArg)
      
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if debugAutoToggle then
          LOG('ReceiveACUMessage_RegisterAutoToggle: ' .. autoToggleUnitArg.DebugId
          .. ' PriorityCategory=' .. repr(autoToggleUnitArg.PriorityCategory) 
          .. ' GetMyTechLevel=' .. repr(autoToggleUnitArg.TechLevel)
          )
        end
        
        --# Keep a reference to the unit object
        local index1 = autoToggleUnitArg.PriorityCategory
        local index2 = autoToggleUnitArg.TechLevel
        table.insert(
            self.AutoToggleTables[index1][index2], 
            autoToggleUnitArg
        )
        
        --# This will generate a new priority list,
        --# removing any dead units from it.
        self:PerformCleanCycle(self.InsertIntoPriorityTable, autoToggleUnitArg)
        
        --# This code checks the ACU's tabkle about whiuch units can be on or off.
        --# It reduces the amount of CPU cycles that the ACU uses to mark things
        --# at each cycle as able to be on or off if we record what can be on
        --# and set acording to the table when we add a unit to the lists.
        if self.AllowToBeOn[autoToggleUnitArg.PriorityCategory] then 
            autoToggleUnitArg.CanAutoToggleOnNow = true
        else
            --# switch it off and make sure it can't come back on
            --# until this function is called again with the opposite
            --# argument.                        
            autoToggleUnitArg:SwitchOff()
            autoToggleUnitArg.CanAutoToggleOnNow = false
        end
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to be called by auto-toggle units
    --#*  from their own class when destroyed or auto-toggle is toggled off, 
    --#*  to inform the ACU, so they won't be used anymore by the ACU's Auto Power-down.
    --#** 
    ReceiveACUMessage_UnregisterAutoToggle = function(self, autoToggleUnitArg)
    
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if debugAutoToggle then
          LOG('ReceiveACUMessage_UnregisterAutoToggle: ' .. autoToggleUnitArg.DebugId
          .. ' PriorityCategory=' .. repr(autoToggleUnitArg.PriorityCategory) 
          .. ' GetMyTechLevel=' .. repr(autoToggleUnitArg.TechLevel)
          )
        end
        
        --# Remove from nested table and priority list
        self:UnregisterAutoToggleEntry(autoToggleUnitArg, true, true) 
                                                                              
        --# If the unit is labeled and it's still alive
        if self.AutoToggleNumbersDisplayed and
           (not autoToggleUnitArg:BeenDestroyed()) and 
           (not autoToggleUnitArg:IsDead())
        then
            --# Remove label
            autoToggleUnitArg:ClearCustomName()
        end
    end,

    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to be called by auto-toggle units
    --#*  from their own class when created, to inform the ACU 
    --#*  to try to move their position in the priority list up
    --#** 
    ReceiveACUMessage_DecreasePositionInPriorityTable = function(self, autoToggleUnitArg)
        self:PerformCleanCycle(self.DecreasePositionInPriorityTable, autoToggleUnitArg)
    end,
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to be called by auto-toggle units
    --#*  from their own class when created, to inform the ACU 
    --#*  to try to move their position in the priority list down
    --#** 
    ReceiveACUMessage_IncreasePositionInPriorityTable = function(self, autoToggleUnitArg)
        self:PerformCleanCycle(self.IncreasePositionInPriorityTable, autoToggleUnitArg)
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  I added this to be called by the UI via a Sim callback
    --#*  to inform the ACU to toggle display of priority list numbers
    --#** 
    ReceiveACUMessage_ToggleDisplayOfAutoToggleNumbers = function(self)
        --# If it was toggled on...  then toggle it off...
        --# If it was toggled off... then toggle it back on...
        self.AutoToggleNumbersDisplayed = not self.AutoToggleNumbersDisplayed
        
         --# Now deal with new value of toggle.
        if self.AutoToggleNumbersDisplayed and self.NetworkNumbersDisplayed then 
            self.NetworkNumbersDisplayed = false
            --# Update displays on all listed auto-toggle units
            self:DisplayNetworkNumbers(self.NetworkNumbersDisplayed)
        end
        
        --# Update displays on all listed auto-toggle units
        self:DisplayAutoToggleNumbers(self.AutoToggleNumbersDisplayed)
    end,


  
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to be called by cleanup code in this class to remove dead entries
    --#*  left in the table by concurrency problems. 
    --#*  It is also called by ReceiveACUMessage_UnregisterAutoToggle 
    --#*  above which can be called by the unit itself
    --#*  when it dies or when the player toggles off its auto-toggle.
    --#** 
    UnregisterAutoToggleEntry = function(self, autoToggleUnitArg, nestedBoolArg, flatBoolArg)
        
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if debugAutoToggle then
          LOG('UnregisterAutoToggleEntry: ' .. autoToggleUnitArg.DebugId
          .. ' PriorityCategory=' .. repr(autoToggleUnitArg.PriorityCategory) 
          .. ' GetMyTechLevel=' .. repr(autoToggleUnitArg.TechLevel)
          )
        end
        
        --# If the user flagged to clean up nested table
        if nestedBoolArg then
            --# Remove from nested table
            local index1 = autoToggleUnitArg.PriorityCategory
            local index2 = autoToggleUnitArg.TechLevel
            self.AutoToggleTables[index1][index2] = 
                GilbotUtils.RemoveFromArrayByValue(
                    self.AutoToggleTables[index1][index2], 
                    autoToggleUnitArg, 
                    true
                )
        end
        
        --# If the user flagged to clean up flat table
        if flatBoolArg then
            
            --# This will generate an updated priority list, 
            --# removing this unit and any other dead units from it.
            local unitsToCleanUpFromNestedTables = 
                self:RemoveFromPriorityTableByValue(autoToggleUnitArg)
              
            --# Copy over items we are keeping from the old list to a new one
            for unusedArrayIndex, vAutoToggleUnit in unitsToCleanUpFromNestedTables do
                --# Remove from nested table
                local index1 = vAutoToggleUnit.PriorityCategory
                local index2 = vAutoToggleUnit.TechLevel
                self.AutoToggleTables[index1][index2] = 
                    GilbotUtils.RemoveFromArrayByValue(
                        self.AutoToggleTables[index1][index2], 
                        vAutoToggleUnit, 
                        true
                    )
            end
            
        end
    end,
    
    

    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  Update the priority list used by the toggle on/off cycles.
    --#*  It mimicks a function table.removeByValue but also resets 
    --#*  priority indicators on the items so user can see what the new 
    --#*  priorities are.
    --#**
    RemoveFromPriorityTableByValue = function(self, autoToggleUnitArg)
        
        local newList = {}
        local newPriority = 1
        local alreadyRemoved = false
        local unitsToCleanUp = {}
        
        --# Copy over items we are keeping from the old list to a new one
        for oldPriorityListPosition, vAutoToggleUnit in self.AutoToggleOffPriorityList do
        
            --# Check if any clean-up needed
            if vAutoToggleUnit:BeenDestroyed() or vAutoToggleUnit:IsDead() then
                
                --# If this is not the unit we intended to clean up
                if (vAutoToggleUnit ~= autoToggleUnitArg) then
                     
                    --# This next block is for debugging only.
                    --# This can be delete wheh debugging is done.
                    if debugAutoToggle then 
                      LOG('AutoToggle: RemoveFromPriorityTableByValue: ' 
                       .. ' Cleaning up additional dead unit ' .. vAutoToggleUnit.DebugId 
                       .. ' found in ACU AutoToggleOffPriorityList.' 
                      )
                    end
                    
                    --# Add this unit to list of units to remove from nested tables.
                    table.insert(unitsToCleanUp, vAutoToggleUnit)
                end
              
            --# If this is not the unit we want to remove...           
            elseif vAutoToggleUnit ~= autoToggleUnitArg then
                --# Have we already removed the unit we wanted to?
                if alreadyRemoved and self.AutoToggleNumbersDisplayed then 
                    --# Need to re-label all units that follow the one removed
                    --# because their priority number will have decreased by one 
                    vAutoToggleUnit:SetCustomName("Auto " .. repr(newPriority))
                end
                
                --# Add it to the end of the new table
                table.insert(newList, vAutoToggleUnit)
                --# increase the priority position for ther next unit  
                newPriority = newPriority + 1
            else
                --# By not adding it to the new list, 
                --# we have just removed the unit we wanted to
                alreadyRemoved = true
            end
        end
        --# Transfer result to our list variable
        self.AutoToggleOffPriorityList = newList 
        --# Return list of dead units that need
        --# to be removed (cleaned up) from nested 
        --# tables to calling code.
        return unitsToCleanUp
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  Update the priority list used by the toggle on/off cycles.
    --#*  It inserts the new autotoggle unit into the ordered list 
    --#*  according to its default priority level, which is set in the mod scripts.
    --#*    
    --#*  This function must be executed via the PerformCleanCycle function
    --#*  so that when any dead entries are encountered, they are cleaned up,
    --#*  and this function will be called again to complete the job.
    --#** 
    InsertIntoPriorityTable = function(self, autoToggleUnitArg) 
        
        local newList = {}
        local newPriority = 1
        local alreadyPlaced = false
        
        --# Copy over items we are keeping from the old list to a new one
        for oldPriorityListPosition, vAutoToggleUnit in self.AutoToggleOffPriorityList do
        
            --# Check if any clean-up needed
            if vAutoToggleUnit:BeenDestroyed() or vAutoToggleUnit:IsDead() then
                
                --# This is used to defer clean up of dead table entry
                self.AutoToggleUnitToCleanUp = vAutoToggleUnit
                
                --# This next block is for debugging only.
                --# This can be delete wheh debugging is done.
                if debugAutoToggle then
                  WARN('AutoToggle: InsertIntoPriorityTable: ' 
                   .. ' Going to clean up dead unit ' .. vAutoToggleUnit.DebugId 
                   .. ' found in ACU AutoToggleOffPriorityList.' 
                  )
                end
                
                --# Abandon this cycle to perform clean-up.
                --# This must be executed with PerformCleanCycle
                --# so that once this entry is cleaned up, 
                --# this function will be called again.
                return
                        
            --# First check for duplicates and abort if found
            elseif vAutoToggleUnit == autoToggleUnitArg then
                WARN('AutoToggle: InsertIntoPriorityTable: '
                 .. ' Attempted to add duplicate entry for ' .. autoToggleUnitArg.DebugId 
                 .. ' to AutoToggleOffPriorityList.'
                )
                return
            
            --# Next, check flag to see if new unit has already been placed.  If it has...
            elseif alreadyPlaced then
                --# Re-label next unit (because priority will be one lower) and then copy over 
                if self.AutoToggleNumbersDisplayed then
                    vAutoToggleUnit:SetCustomName("Auto " .. repr(newPriority))
                end
                vAutoToggleUnit.PriorityListPosition = newPriority
                table.insert(newList, vAutoToggleUnit)
                --# increment index for next entry
                newPriority = newPriority + 1
            
            --# Check if thi sis the right place for insertion.
            --# If this unit in the list already has not had a custom priority set
            elseif vAutoToggleUnit.AutoTogglePositionsMoved == 0 and 
              --# and its normal priority is below the normal priority of the new unit
              (vAutoToggleUnit.PriorityCategory > autoToggleUnitArg.PriorityCategory) 
              or
              ( --# or if they have the same priority category 
               (vAutoToggleUnit.PriorityCategory == autoToggleUnitArg.PriorityCategory) 
                and --# and the new unit has a higher tech level
               (vAutoToggleUnit.PriorityCategory < autoToggleUnitArg.TechLevel)
              )
            then
                --# This next block is for debugging only.
                --# This can be delete wheh debugging is done.
                if false and debugAutoToggle then
                  LOG('AutoToggle: InsertIntoPriorityTable: ' 
                   .. ' Inserted new unit ' .. autoToggleUnitArg.DebugId 
                   .. ' into position= ' .. repr(newPriority)
                  )
                end
                
                --# Insert new auto-toggle unit here and label priority
                if self.AutoToggleNumbersDisplayed then
                    autoToggleUnitArg:SetCustomName("Auto " .. repr(newPriority))
                end
                autoToggleUnitArg.PriorityListPosition = newPriority
                table.insert(newList, autoToggleUnitArg)
                --# Move to next position in list
                newPriority = newPriority + 1
                
                --# Put the unit from the list we compared against just after it
                if self.AutoToggleNumbersDisplayed then
                    vAutoToggleUnit:SetCustomName("Auto " .. repr(newPriority))
                end
                vAutoToggleUnit.PriorityListPosition = newPriority
                table.insert(newList, vAutoToggleUnit)
                --# increment index for next entry
                newPriority = newPriority + 1
                
                --# Future units all get shifted one position along
                alreadyPlaced = true
                
               
            else
                --# Just do a straight copy
                table.insert(newList, vAutoToggleUnit)
                --# increment index for next entry
                newPriority = newPriority + 1
            end
            
            
        end
        
        --# If there were no units already in the list 
        --# that this unit should be inserted after, then 
        if not alreadyPlaced then
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if debugAutoToggle then
              LOG('AutoToggle: InsertIntoPriorityTable: ' 
               .. ' Appended new unit ' .. autoToggleUnitArg.DebugId 
               .. ' at final position= ' .. repr(newPriority)
              )
            end
                
            --# Put this in at the end
            if self.AutoToggleNumbersDisplayed then
                autoToggleUnitArg:SetCustomName("Auto " .. repr(newPriority))
            end
            autoToggleUnitArg.PriorityListPosition = newPriority
            table.insert(newList, autoToggleUnitArg)
        end
        
        --# Transfer result to our list variable
        self.AutoToggleOffPriorityList = newList 
        
    end, 


    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  Update the priority list used by the toggle on/off cycles.
    --#*  It swaps the position of the unit selected with the unit before it.
    --#*    
    --#*  This function must be executed via the PerformCleanCycle function
    --#*  so that when any dead entries are encountered, they are cleaned up,
    --#*  and this function will be called again to complete the job.
    --#** 
    DecreasePositionInPriorityTable = function(self, autoToggleUnitArg) 
        
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if false and debugAutoToggle then
          LOG('AutoToggle: DecreasePositionInPriorityTable: ' 
           .. ' Called on unit ' .. autoToggleUnitArg.DebugId 
           .. ' with old position= ' .. repr(autoToggleUnitArg.PriorityListPosition)
          )
        end
        
        --# The size of the table is also the index of the last position 
        --# in the table because these tables have 1 as the first index.
        local priorityTableSizeAndLastPosition = table.getsize(self.AutoToggleOffPriorityList)
        
        --# Can't swap if there are less than two units in the list!
        if priorityTableSizeAndLastPosition < 2 then return end
        
        --# Can't increase priority list position if the position is already the max value
        if autoToggleUnitArg.PriorityListPosition == 1 
        then return end
      
        --# We can do a swap.
        local newList = {}
        local newPriority = 1
        local swapSuccessful = false
        local doNothingPosition = autoToggleUnitArg.PriorityListPosition
        local doSwapPosition = autoToggleUnitArg.PriorityListPosition - 1
        
        
        --# Copy over items we are keeping from the old list to a new one
        for oldPriorityListPosition, vAutoToggleUnit in self.AutoToggleOffPriorityList do
        
            --# Check if any clean-up needed
            if vAutoToggleUnit:BeenDestroyed() or vAutoToggleUnit:IsDead() then
                
                --# This is used to defer clean up of dead table entry
                self.AutoToggleUnitToCleanUp = vAutoToggleUnit
                
                --# This next block is for debugging only.
                --# This can be delete wheh debugging is done.
                if debugAutoToggle then
                  WARN('AutoToggle: DecreasePositionInPriorityTable: ' 
                   .. ' Going to clean up dead unit ' .. vAutoToggleUnit.DebugId 
                   .. ' found in ACU AutoToggleOffPriorityList.' 
                  )
                end
                
                --# Abandon this cycle to perform clean-up.
                --# This must be executed with PerformCleanCycle
                --# so that once this entry is cleaned up, 
                --# this function will be called again.
                return
                        
            --# Check if this is the right place for insertion.
            --# If this unit in the list already has not had a custom priority set
            elseif oldPriorityListPosition == doNothingPosition then
                --# Do nothing
            elseif oldPriorityListPosition == doSwapPosition then
                
                --# The new unit that used to be before it 
                --# now goes in the position after it
                if self.AutoToggleNumbersDisplayed then
                    autoToggleUnitArg:SetCustomName("Auto " .. repr(newPriority))
                end
                autoToggleUnitArg.PriorityListPosition = newPriority
                table.insert(newList, autoToggleUnitArg)
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAutoToggle then
              LOG('AutoToggle: DecreasePositionInPriorityTable:'
              .. ' Inserted ' .. autoToggleUnitArg.DebugId 
              .. ' at position ' .. repr(autoToggleUnitArg.PriorityListPosition)
              .. ' with decreasing its position.'
              )
            end
                --# increment index for next entry
                newPriority = newPriority + 1
                
                --# This unit goes into the slot we left empty last time
                if self.AutoToggleNumbersDisplayed then
                    vAutoToggleUnit:SetCustomName("Auto " .. repr(newPriority))
                end
                vAutoToggleUnit.PriorityListPosition = newPriority
                table.insert(newList, vAutoToggleUnit)
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAutoToggle then
              LOG('AutoToggle: DecreasePositionInPriorityTable:'
              .. ' Inserted ' .. vAutoToggleUnit.DebugId 
              .. ' at position ' .. repr(vAutoToggleUnit.PriorityListPosition)
              .. ' with increasing its position.'
              )
            end
                --# Move to current position in list
                newPriority = newPriority + 1
                
                --# Future units just get copied 
                swapSuccessful = true
            else
                --# Just do a straight copy
                table.insert(newList, vAutoToggleUnit)
                
                --# This next block is for debugging only.
                --# This can be delete wheh debugging is done.
                if false and debugAutoToggle then
                  LOG('AutoToggle: DecreasePositionInPriorityTable:'
                  .. ' Inserted ' .. vAutoToggleUnit.DebugId 
                  .. ' at position ' .. repr(vAutoToggleUnit.PriorityListPosition)
                  .. ' without changing its position.'
                  )
                end
                --# increment index for next entry
                newPriority = newPriority + 1
            end
            
        end
        
        --# If the swap actually happened..
        if swapSuccessful then
            --# Transfer result to our list variable
            self.AutoToggleOffPriorityList = newList 
            
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAutoToggle then
              LOG('AutoToggle: DecreasePositionInPriorityTable: Completed.' 
              .. ' Table has ' .. repr(table.getsize(self.AutoToggleOffPriorityList))
              .. ' units.'
              )
            end
            
            --# Mark this unit to show how many places its 
            --# custom position differs from its default position.
            autoToggleUnitArg.AutoTogglePositionsMoved = 
                autoToggleUnitArg.AutoTogglePositionsMoved - 1
        end
        
    end, 
    
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  Update the priority list used by the toggle on/off cycles.
    --#*  It swaps the position of the unit selected with the unit after it.
    --#*    
    --#*  This function must be executed via the PerformCleanCycle function
    --#*  so that when any dead entries are encountered, they are cleaned up,
    --#*  and this function will be called again to complete the job.
    --#** 
    IncreasePositionInPriorityTable = function(self, autoToggleUnitArg) 
        
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if false and debugAutoToggle then
          LOG('AutoToggle: IncreasePositionInPriorityTable: ' 
           .. ' Called on unit ' .. autoToggleUnitArg.DebugId 
           .. ' with old position= ' .. repr(autoToggleUnitArg.PriorityListPosition)
          )
        end
                
        --# The size of the table is also the index of the last position 
        --# in the table because these tables have 1 as the first index.
        local priorityTableSizeAndLastPosition = table.getsize(self.AutoToggleOffPriorityList)
        
        --# Can't swap if there are less than two units in the list!
        if priorityTableSizeAndLastPosition < 2 then return end
        
        --# Can't increase priority list position if the position is already the max value
        if autoToggleUnitArg.PriorityListPosition == priorityTableSizeAndLastPosition 
        then return end
      
        --# We can do a swap.
        local newList = {}
        local newPriority = 1
        local swapSuccessful = false
        local doNothingPosition = autoToggleUnitArg.PriorityListPosition
        local doSwapPosition = autoToggleUnitArg.PriorityListPosition + 1
        
        --# Copy over items we are keeping from the old list to a new one
        for oldPriorityListPosition, vAutoToggleUnit in self.AutoToggleOffPriorityList do
        
            --# Check if any clean-up needed
            if vAutoToggleUnit:BeenDestroyed() or vAutoToggleUnit:IsDead() then
                
                --# This is used to defer clean up of dead table entry
                self.AutoToggleUnitToCleanUp = vAutoToggleUnit
                
                --# This next block is for debugging only.
                --# This can be delete wheh debugging is done.
                if debugAutoToggle then
                  WARN('AutoToggle: IncreasePositionInPriorityTable: ' 
                   .. ' Going to clean up dead unit ' .. vAutoToggleUnit.DebugId 
                   .. ' found in ACU AutoToggleOffPriorityList.' 
                  )
                end
                
                --# Abandon this cycle to perform clean-up.
                --# This must be executed with PerformCleanCycle
                --# so that once this entry is cleaned up, 
                --# this function will be called again.
                return
                        
            --# Check if this is the right place for insertion.
            --# If this unit in the list already has not had a custom priority set
            elseif oldPriorityListPosition == doNothingPosition then
                --# Do nothing
            elseif oldPriorityListPosition == doSwapPosition then
                
                --# This unit goes into the slot we left empty last time
                if self.AutoToggleNumbersDisplayed then
                    vAutoToggleUnit:SetCustomName("Auto " .. repr(newPriority))
                end
                vAutoToggleUnit.PriorityListPosition = newPriority
                table.insert(newList, vAutoToggleUnit)
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAutoToggle then
              LOG('AutoToggle: DecreasePositionInPriorityTable:'
              .. ' Inserted ' .. vAutoToggleUnit.DebugId 
              .. ' at position ' .. repr(vAutoToggleUnit.PriorityListPosition)
              .. ' with decreasing its position.'
              )
            end
                --# Move to current position in list
                newPriority = newPriority + 1
                
                --# The new unit that used to be before it 
                --# now goes in the position after it
                if self.AutoToggleNumbersDisplayed then
                    autoToggleUnitArg:SetCustomName("Auto " .. repr(newPriority))
                end
                autoToggleUnitArg.PriorityListPosition = newPriority
                table.insert(newList, autoToggleUnitArg)
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAutoToggle then
              LOG('AutoToggle: DecreasePositionInPriorityTable:'
              .. ' Inserted ' .. autoToggleUnitArg.DebugId 
              .. ' at position ' .. repr(autoToggleUnitArg.PriorityListPosition)
              .. ' with increasing its position.'
              )
            end
                --# increment index for next entry
                newPriority = newPriority + 1
                
                --# Future units just get copied 
                swapSuccessful = true
            else
                --# Just do a straight copy
                table.insert(newList, vAutoToggleUnit)
                
                --# This next block is for debugging only.
                --# This can be delete wheh debugging is done.
                if false and debugAutoToggle then
                  LOG('AutoToggle: DecreasePositionInPriorityTable:'
                  .. ' Inserted ' .. vAutoToggleUnit.DebugId 
                  .. ' at position ' .. repr(vAutoToggleUnit.PriorityListPosition)
                  .. ' without changing its position.'
                  )
                end
                
                --# increment index for next entry
                newPriority = newPriority + 1
            end
            
        end
        
        --# If the swap actially happened..
        if swapSuccessful then
            --# Transfer result to our list variable
            self.AutoToggleOffPriorityList = newList 
            
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAutoToggle then
              LOG('AutoToggle: IncreasePositionInPriorityTable: Completed.' 
              .. ' Table has ' .. repr(table.getsize(self.AutoToggleOffPriorityList))
              .. ' units.'
              )
            end
            
            --# Mark this unit to show how many places its 
            --# custom position differs from its default position.
            autoToggleUnitArg.AutoTogglePositionsMoved = 
                autoToggleUnitArg.AutoTogglePositionsMoved + 1
        end
        
    end, 
    
    
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This should be called when the ACU dies so it 
    --#*  is obvious that autotoggle has been disabled on all units.
    --#*  This is safe to be called more than once.
    --#** 
    DisplayAutoToggleNumbers = function(self, toggleOn)    
        
        --# Copy over items we are keeping from the old list to a new one
        for priorityListPosition, vAutoToggleUnit in self.AutoToggleOffPriorityList do
        
             --# Check if any clean-up needed
            if vAutoToggleUnit:BeenDestroyed() or vAutoToggleUnit:IsDead() then
                
                --# This is used to defer clean up of dead table entry
                self.AutoToggleUnitToCleanUp = vAutoToggleUnit
                
                --# This next block is for debugging only.
                --# This can be delete wheh debugging is done.
                if debugAutoToggle then
                  WARN('AutoToggle: DecreasePositionInPriorityTable: ' 
                   .. ' Going to clean up dead unit ' .. vAutoToggleUnit.DebugId 
                   .. ' found in ACU AutoToggleOffPriorityList.' 
                  )
                end
                
                --# Abandon this cycle to perform clean-up.
                --# This must be executed with PerformCleanCycle
                --# so that once this entry is cleaned up, 
                --# this function will be called again.
                return
                        
            --# Show label
            elseif toggleOn then
                vAutoToggleUnit:SetCustomName("Auto " .. repr(priorityListPosition))
            
            --# Remove label
            else
                vAutoToggleUnit:ClearCustomName()
            end
        end
    end,
    
}




