--#*****************************************************************************
--#* New File : /mods/gilbotsimprovements/activeanimationunit.lua
--#*
--#* Modded By: Gilbot-X
--#*
--#* Summary  : This extra code was originally from my Exponential Mass Extractors mod.
--#*            I made it to remove jumps in animation and to make use of the
--#*            fact that certain positions in the unit's animations make them 
--#*            appear to be either in a defensive or vulnerable state or position.          
--#*
--#*            Adds the following Interface to PauseableProductionUnit:
--#*            ---------------------------------
--#*            EnsureActiveAnimationManipulatorIsCreated
--#*            PauseActiveAnimation
--#*            WaitForActiveAnimationManipulator
--#*            
--#*            Changes this GPG Interface:
--#*            ---------------------------------
--#*            PlayActiveAnimation
--#*            OnActive/OnInactive
--#*
--#*            The Auto-Pack interface can also be added to a unit that has this, 
--#*            but the autopack interface must be added after this one, not before.
--#*            Neither of these interfaces can be applied without the PauseableProductionUnit
--#*            interface being applied beforehand.
--#*
--#*            For example, the ??? units use this interface without the AutoPackUnit
--#*            interface, but all the Aeon mexes use the AutoPackUnit and ActiveAnimationUnit 
--#*            interfaces together.
--#*
--#*****************************************************************************

local GilbotUtils_IsValueInTable = 
    import('/mods/gilbotsimprovements/utils.lua').IsValueInTable
    
--# This allows dynamic class extension:
--# a class can be used on more than one base class
--# so that various classes can be created with it.
local MakePauseableProductionUnit = 
    import('/mods/gilbotsimprovements/pauseableproductionunit.lua').MakePauseableProductionUnit


--# This function is called to create the class 
--# so this class can add its 
--# code to different base classes.
function MakeActiveAnimationUnit(baseClassArg)

local BaseClass = MakePauseableProductionUnit(baseClassArg)  
local resultClass = Class(BaseClass) {

    --# Quick way to check if a unit extends this class.
    IsActiveAnimationUnit = true,

    --# This is all for animation control.
    ActiveAnimationRate = 1,
    ActiveAnimationMaxRate =8, -- might even need to lower this...
    
    --# Extending classes should define this next key
    --# if their active animation is not to
    --# be played in an infinite loop, i.e. they use autopack.
    ActiveAnimationDoesNotUseInfiniteLoop = false,
    
    --# Declare this here in hope that it stays in scope better
    ActiveAnimationManipulator = nil,

    
    --#*    
    --#*  Gilbot-X says:    
    --#*
    --#*  This is just a stub in StructureUnit class, 
    --#*  so no need to call any base class versions.
    --#*  It is called by OnStopBeingBuilt in that class.
    --#*  but it is only called once to start the animation.
    --#*  Pausing/resuming the animation is done from 
    --#*  the update thread and in States defined
    --#*  in the TimeBasedMassCreationUnit class.
    --#**
    PlayActiveAnimation = function(self)
        self:PauseActiveAnimation(false)
    end,
        
  


  
  
    --#*    
    --#*  Gilbot-X says:    
    --#*
    --#*  It is called by PauseActiveAnimation in this class and 
    --#*  in Aeonunits.lua where that class does packing/unpacking.
    --#**
    EnsureActiveAnimationManipulatorIsCreated = function(self)
        
        --# Make sure there is an animation manipulator working.
        if not self.ActiveAnimationManipulator then
            
            --# This is for debugging only.  I intend to delete it.
            if self.AlreadyMadeOneActiveAnimationManipulator then
              WARN('MassCollectionUnit: PauseActiveAnimation: '
                .. 'Creating replacement ActiveAnimationManipulator.')
            end
        
            --# Create a new animation manipulator.
            --# This should last the lifetime of the MeX object
            --# but I have made many observations of that not happening.
            self.ActiveAnimationManipulator = CreateAnimator(self)
            self.Trash:Add(self.ActiveAnimationManipulator)
            self.AlreadyMadeOneActiveAnimationManipulator = true
            
            if not self.ActiveAnimationDoesNotUseInfiniteLoop then
                --# Play animation in a loop
                self.ActiveAnimationManipulator:PlayAnim(
                    self:GetBlueprint().Display.AnimationActivate,       
                    true
                )
            end
        end
    end,
  
  
  
    --#*    
    --#*  Gilbot-X says:    
    --#*
    --#*  This is called once by OnStopBeingBuilt and then 
    --#*  again whenever unit is paused/unpaused.
    --#*  Pausing/resuming the animation should be done inside
    --#*  the context of appropriate states.
    --#**
    PauseActiveAnimation = function(self, toggleToPaused)
        
        --# Make sure there is an animation manipulator working.
        self:EnsureActiveAnimationManipulatorIsCreated()
        
        --# Resume animation at appropriate speed, 
        --# or pause it, depending on toggle argument
        if toggleToPaused and not self.ActiveAnimationDoesNotUseInfiniteLoop then
            --# We are pausing animation that runs in an infinite loop
            self.ActiveAnimationManipulator:SetRate(0)
        else
            
            --# If this unit does its animation one loop at a time
            --# and its at the beginning or end of a loop then
            --# play another loop if we are not pausing it
            if not toggleToPaused and 
                self.ActiveAnimationDoesNotUseInfiniteLoop and 
               (not self.ActiveAnimationLoopRunning) and 
               (not self.Animating)
            then
                --LOG('ActiveAnimationUnit: Playing New Loop.')
                
                --# Play animation once   
                self.ActiveAnimationManipulator:PlayAnim(
                    self:GetBlueprint().Display.AnimationActivate,       
                    false
                )
                --# Don't let a second loop be set going
                --# until this one is finished.
                self.ActiveAnimationLoopRunning = true
                --# The unit is being animated,
                --# so it isn't ready for upgrade.
                self.NotReadyToUpgrade = true
            end
            
            if (not self.ActiveAnimationDoesNotUseInfiniteLoop) or
               self.ActiveAnimationLoopRunning
            then               
                --# Make sure animation is moving forward at whatever speed is appropriate
                --# So current loop can finish or if infinite animation is being unpaused
                self.ActiveAnimationManipulator:SetRate(self.ActiveAnimationRate)
            else
                WARN('ActiveAnimationUnit: PauseActiveAnimation called between loops!!')
            end
        end
    end,
  
  
        
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  What to do when we start producing energy.  
    --#*  This happens when unit's production is unpaused
    --#*  and it has whatever other conditions it needs 
    --#*  to execute production.
    --#**
    OnActive = function(self)
        --# Call base class code first
        baseClassArg.OnActive(self)
        --# Make sure animation is not paused.
        self:PauseActiveAnimation(false)
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  What to do when we stop producing energy.  
    --#*  This happens when unit's production is paused
    --#*  or it doesn't have whatever other conditions 
    --#*  it needs to execute production.
    --#**
    OnInactive = function(self)
        --# Call base class code first
        baseClassArg.OnInactive(self)
    end,
        
        
 

    
        
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  How to behave when production pause is toggled to 'pause'
    --#**
    PausedState = State {
        Main = function(self)
       
            --# Does this unit play its active animation one loop at a time?
            if self.ActiveAnimationDoesNotUseInfiniteLoop then
                self:WaitForActiveAnimationManipulator()
            else
                --# Freeze the animation
                self:PauseActiveAnimation(true)
            end
            
            --# Do Base class stuff later,
            --# i.e. shields and DLS
            BaseClass.PausedState.Main(self)
        end,

        OnProductionUnpaused = BaseClass.PausedState.OnProductionUnpaused,
    }, 

    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  This worked in Supreme Commander but no longer works 
    --#*  in FA due to a new threading bug intoduced into FA by GPG.
    --#*  Now, random warnings are issued here in the log file 
    --#*  when this is called from the UnPausedState,
    --#*  which is the state that loops continually and calls 
    --#*  this function once each loop.  The warning states
    --#*  that the calling thread is not allowed to suspend.
    --#*  possibly the thread has been marked for destruction
    --#*  at a state transition and this occurs as a race condition. 
    --#*  
    --#*  The purpose of this function:
    --#*  Direct all WaitFor calls on active animation here
    --#*  so we record state in a consistent way when a 
    --#*  loop of animation reaches its end.  This should 
    --#*  be called before anything else is carried out
    --#*  on the animation manipulator for units that
    --#*  play the active animation one loop at a time.
    --#**
    WaitForActiveAnimationManipulator = function(self)     
        if self.ActiveAnimationManipulator then
            --# Waiting until the end of this animation loop 
            --# will result in the unit being paused.
            --# If it is also an auto-pack unit, it will be ready to pack.            
         
            --# Wait until we are at beginning of animation.
            WaitFor(self.ActiveAnimationManipulator)
    
            --# We have reached the end of the animation loop
            --# so set this flag which signals this to other code 
            --# so that it might resume what it was waiting to do.
            self.ActiveAnimationLoopRunning = false
        else
            WARN('ActiveAnimationUnit: PausedState: ActiveAnimationManipulator not defined.')
        end
    end,
    
    
}


return resultClass

end