--#****************************************************************************
--#* New File : /mods/gilbotsimprovements/autopackunit.lua
--#*
--#* Modded By: Gilbot-X
--#*
--#* Summary  : Used by weaponless structures (i.e. economy based) 
--#*            that have an open/active animation that we can 
--#*            manipulate to appear in (or out of) a particular defensive 
--#*            position.  This interface allows these units to automatically 
--#*            assume that position when threatened or under fire.
--#*
--#*            Adds the following Interface to PauseableProductionUnit:
--#*            ---------------------------------
--#*            DoPackUpOperation
--#*            WaitForActiveAnimationManipulator (Also in ActiveAnimationUnit Interface)
--#*            WatchState
--#*            CheckAreEnemyUnitsNearby
--#*            DoWatchLoop
--#*            OnDamageWhenPaused
--#*            
--#*            Updates these Interfaces:
--#*            ---------------------------------
--#*            PausedState/UnpausedState
--#*            PreparingToUpgradeState
--#*            OnActive/OnInactive
--#*
--#*            The ActiveAnimationUnit interface can also be added to a unit that has this, 
--#*            but the ActiveAnimationUnit interface must be added before this one, not after.
--#*            Neither of these interfaces can be applied without the PauseableProductionUnit
--#*            interface being applied beforehand.
--#*
--#*            For example, the Aeon HCPP uses this interface without the ActiveAnimationUnit
--#*            interface, but all the Aeon mexes use the AutoPackUnit and ActiveAnimationUnit 
--#*            interfaces together.
--#*
--#****************************************************************************

--# Need this so we can call base class code
--# and bypass any intervening superclass
local Unit = import('/lua/sim/unit.lua').Unit

--# This function is called to create the class 
--# so this class can add its 
--# code to different base classes.
function MakeAutoPackUnit(baseClassArg) 

local resultClass = Class(baseClassArg) {
 
    --# Quick way to check if a unit is an 
    --# autopack unit that extends this class.
    IsAutoPackUnit = true,
    
    --# This sets how often the unit should 
    --# checks to see if it should unpack.
    --# The longer it waits, the safer it will be
    --# but you will lose out on production time
    --# as a consequence to extra safety.
    WatchLoopPeriod = 10,
 
 
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  Code that will be done after the unit 
    --#*  has been built but before the first state is changed
    --#*  in a structure unit that uses states.
    --#**
    DoBeforeAnyStateChanges = function(self)
        --# Do base class versions first
        baseClassArg.DoBeforeAnyStateChanges(self)
    
        --# This isn't important
        --# but I've done it for safety
        self.Damaged = false
            
        --# Switch Auto-pack on.
        --# If it is toggled off.. 
        if not self:GetScriptBit('RULEUTC_WeaponToggle') then
            --# Switch auto-toggle on
            self:SetScriptBit('RULEUTC_WeaponToggle', true)
        else 
            --# or just run the code that switches it on 
            self:OnScriptBit1Set() 
        end
    end, 
    
    
    
    --#* 
    --#*  Gilbot-X says:
    --#* 
    --#*  This function contains code peripheral to the packing
    --#*  unpacking animation; i.e. it handles the state change safely.
    --#**
    DoPackUpOperation = function(self, isPackedRequested)
        --# Do we need to pack or unpack to acheive this?
        if self.IsPacked ~= isPackedRequested then
        
            --# Check if we need to adjust our animation
            --# so that we appear in a defensive state/position.
            --# Aeon Mexes and HCPP do this.
            --# UEF units don't as far as I remember.
            if self.DoPackUpAnimation then
                
                --# Work ot what toggles we need to temporarily disable
                self.HasShieldToggles =  self:TestToggleCaps('RULEUTC_ShieldToggle')
                
                --# There is an animation to do.
                --# Disable toggle buttons and lock while we animate.
                self.Animating = true
                self:RemoveToggleCap('RULEUTC_ProductionToggle')
                self:RemoveToggleCap('RULEUTC_WeaponToggle')
                if self.HasShieldToggles then self:RemoveToggleCap('RULEUTC_ShieldToggle') end
                
                --# This must be defined in extending class
                self:DoPackUpAnimation(isPackedRequested)
                
                --# reenable toggle buttons and unlock
                if self.HasShieldToggles then self:AddToggleCap('RULEUTC_ShieldToggle') end
                self:AddToggleCap('RULEUTC_WeaponToggle')
                self:AddToggleCap('RULEUTC_ProductionToggle')
                self.Animating = false
            end
            --# Record operation is complete.
            self.IsPacked = isPackedRequested
        end
    end,
  

--[[   



Any extending class needs to define this next function to conform to the example below. It must manipulate the open animation to give the effect of the unit packing/unpacking, depending on argument.
This makes it look like the unit can change into a defensive position.
    
    DoPackUpAnimation = function(self, isPackedRequested)
        --# Do we need to pack or unpack to acheive this?
        if isPackedRequested then 
            --# We are deactivating production, so perform pack animation
            self.AnimationManipulator:SetRate(-0.5)
            WaitFor(self.AnimationManipulator)
        else
            --# We are activating production, so perform unpack animation.
            self.AnimationManipulator:PlayAnim(
                self:GetBlueprint().Display.AnimationOpen, false):SetRate(0.5)
            WaitFor(self.AnimationManipulator)
        end
    end,
    
    
    
]]
    
  
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  What to do when we start producing energy.  
    --#*  This happens when unit's production is unpaused
    --#*  and it has whatever other conditions it needs 
    --#*  to execute production.
    --#**
    OnActive = function(self)
        --# This makes sure that damage stabilisation effect is stopped
        if self.DestroyEffects then self:DestroyEffects() end
        
        --# Play unpack animation of there is one
        self:DoPackUpOperation(false)
        
        --# Call base class code now to take care of active animations, 
        --# visual effects and sound effect. 
        baseClassArg.OnActive(self)
    end,
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  What to do when we stop producing energy.  
    --#*  This happens when unit's production is paused
    --#*  or it doesn't have whatever other conditions 
    --#*  it needs to execute production.
    --#**
    OnInactive = function(self)
        --# Call base class first.
        --# That will take care of active animations. 
        baseClassArg.OnInactive(self)
        
        --# Play packing up animation of there is one
        self:DoPackUpOperation(true)
    end,


    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  This worked in Supreme Commander but no longer works 
    --#*  in FA due to a new threading bug intoduced into FA by GPG.
    --#*  Now, random warnings are issued here in the log file 
    --#*  when this is called from the UnPausedState,
    --#*  which is the state that loops continually and calls 
    --#*  this function once each loop.
    --#*  
    --#*  How it works:
    --#*  Direct all WaitFor calls on active animation here
    --#*  so we record state in a consistent way when a 
    --#*  loop of animation reaches its end.  This should 
    --#*  be called before anything else is carried out
    --#*  on the animation manipulator for units that
    --#*  play the active animation one loop at a time.
    --#*  The original version is in ActiveAnimationUnit.lua
    --#*  but this override makes sure that after doing
    --#*  that core function that status flags relevant 
    --#*  to auto-pack units are also set.
    --#**
    WaitForActiveAnimationManipulator = function(self)
        
        --# Safety check first
        if not self.IsActiveAnimationUnit then return end
        
        --# Base class version is in ActiveAnimationUnit class layer.
        baseClassArg.WaitForActiveAnimationManipulator(self)
    
        --# The rest is specific to those units that have active 
        --# animations but also animate (it or something else) to pack up.
        --# First, record if we are packed (or not) now that
        --# we have reached the end of another animation loop.
        self.IsPacked = self.PackedAtAnimationStart
        
        --# Next, check/record if we are ready to upgrade.
        if self.IsPacked and self.UpgradesFromPacked then 
            self.NotReadyToUpgrade = false
        end
    end,  

    
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  How to behave when production pause is toggled to 'unpause'
    --#**
    UnpausedState = State {
        Main = function(self)
            --# Call code from PauseableProductionUnit base class first
            baseClassArg.UnpausedState.Main(self)
          
            --# Special code for units whose animation pauses 
            --# after each loop to do processing.
            if self.IsActiveAnimationUnit and self.ActiveAnimationDoesNotUseInfiniteLoop then
                --# This keeps the animation going and
                --# performs a routine scan for enemies.
                while not self:IsDead() do
                    --# Wait until previous loop is finished.
                    --self:WaitForActiveAnimationManipulator()
                    --# Safety check first
                    if self.IsActiveAnimationUnit then
                    
                        --# Base class version is in ActiveAnimationUnit class layer.
                        --baseClassArg.WaitForActiveAnimationManipulator(self)
                        if self.ActiveAnimationManipulator then
                            --[[
                            --# Debugging code only
                            local thisThread = CurrentThread()
                            if self.LastThread ~= thisThread then
                                LOG('ThreadTest#' .. self:GetEntityId() 
                                .. ':New=' .. repr(thisThread)
                                )
                                self.LastThread = thisThread 
                            end
                            LOG('ThreadTest#' .. self:GetEntityId() 
                            .. ':CallingWaitFor')
                            
                            --#
                            --# Gilbot-X says:  
                            --# The WaitFor command after this block occasionally gives this warning:
                            warning: Error running lua script from destroyed thread: Can't suspend a thread that wasn't created with ForkThread.'
                            ]]
                            --# Wait until we are at beginning of animation.
                            WaitFor(self.ActiveAnimationManipulator)
                            
                            --# We have reached the end of the animation loop
                            self.ActiveAnimationLoopRunning = false
                        else
                            WARN('ActiveAnimationUnit: UnpausedState: '
                              .. 'ActiveAnimationManipulator not defined.')
                        end
                    
                        --# The rest is specific to those units that have active 
                        --# animations but also animate (it or something else) to pack up.
                        --# First, record if we are packed (or not) now that
                        --# we have reached the end of another animation loop.
                        self.IsPacked = self.PackedAtAnimationStart
                        
                        --# Next, check/record if we are ready to upgrade.
                        if self.IsPacked and self.UpgradesFromPacked then 
                            self.NotReadyToUpgrade = false
                        end
                    end
                    
                    --# Don't let a second loop be set going
                    --# until this one is finished.
                    self.ActiveAnimationLoopRunning = false
                    
                    --# Scan for nearby enemy units
                    if self.AutoPackToggledOn and self:CheckAreEnemyUnitsNearby() then
                        --# Pause production
                        self:SetScriptBit('RULEUTC_ProductionToggle', true)    
                    end
                   
                    --# Play another loop of the animation
                    self:PauseActiveAnimation(false) 
                end
            end
        end,

        --# Override to pause production when we get hit by a projectile
        OnDamage = function(self, instigator, amount, vector, damageType) 

            --# For debugging only
            if false then
                LOG('AutoPack: UnpausedState.OnDamage called '
                .. ' on unit ' .. self:GetEntityId()
                .. ' and self.AutoPackToggledOn=' .. repr(self.AutoPackToggledOn)
                )
            end
            
            if instigator == 'SensitiveShield' then
                WARN('SensitiveShield is instigatior in OnDamage when unpaused.')
                return
            end
        
            --# Record that we were damaged so WatchLoop 
            --# (defined in AutoPackUnit abstract class) can query it.
            self.Damaged = true
            
            --# This takes place outside WatchState so pause production
            if (not self.Animating) and self.AutoPackToggledOn then 
                --# Pause production
                self:SetScriptBit('RULEUTC_ProductionToggle', true)
            end
            
            --# Forwarding OnDamage to Unit class to reduce our health
            baseClassArg.OnDamage(self, instigator, amount, vector, damageType)
        end,
        
        
        --# Defer to PauseableProductionUnit base class 
        OnProductionPaused = baseClassArg.UnpausedState.OnProductionPaused
    },

    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  How to behave when production pause is toggled to 'pause'
    --#**
    PausedState = State {
        Main = function(self)
            --# Call code from PauseableProductionUnit base class first
            --# This switches on shield  or DLS so assume packing done already
            baseClassArg.PausedState.Main(self)
            
            --# Check if we are eligeable to go into WatchState for 
            --# automatic unpacking.
            if self.AutoPackToggledOn and 
                (self.Damaged or 
                    (self.PackIfEnemyUnitsNearby and self:CheckAreEnemyUnitsNearby())
                )                                   
            then
                --# WatchState is defined in both of the two extending classes.
                ChangeState(self, self.WatchState)
            end
        end,

        --# Defer to a separate function defined outside the state 
        --# because it runs identically in two states.
        OnDamage = function(self, instigator, amount, vector, damageType)
            self:OnDamageWhenPaused(instigator, amount, vector, damageType)
        end,
        
        --# Defer to PauseableProductionUnit base class 
        OnProductionUnpaused = baseClassArg.PausedState.OnProductionUnpaused
    },

    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  This state was added so that the unit will wait until  
    --#*  it hasn't been notified of a projectile hit for a while
    --#*  before it automatically unpacks.  
    --#* 
    --#*  Units must be in this state for DLS or SensitiveShield to work.
    --#**
    WatchState = State {
        Main = function(self)
            self:DoWatchLoop()
        end,
        
        --# Defer to a separate function defined outside the state 
        --# because it runs identically in two states.
        OnDamage = function(self, instigator, amount, vector, damageType)
            self:OnDamageWhenPaused(instigator, amount, vector, damageType)
        end,
        
        --# Defer to PauseableProductionUnit base class 
        OnProductionUnpaused = baseClassArg.PausedState.OnProductionUnpaused
    },
    
    
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is called by  calls to OnDamage received in PausedState 
    --#*  and WatchState.
    --#*
    --#*  This must be overrided so that damage is taken differently when
    --#*  the unit is packed.
    --#**
    OnDamageWhenPaused = function(self, instigator, amount, vector, damageType)
        --# Record that we were damaged so WatchLoop 
        --# (defined in AutoPackUnit abstract class) can query it.
        self.Damaged = true
        
        WARN('Base class version of OnDamageWhenPaused called.'
         .. ' This should have been overrided.'
        )
        
        --# Forwarding OnDamage to Unit class to reduce our health
        baseClassArg.OnDamage(self, instigator, amount, vector, damageType)
    end,
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  This is the default routine executed by WatchState in an extending class.
    --#*  The unit checks once every 10 seconds to see if it's safe to unpack.
    --#*  The extending class should set self.PackIfEnemyUnitsNearby
    --#*  for this to be used.
    --#**
    DoWatchLoop = function(self)
        while not self:IsDead() and self.AutoPackToggledOn do
            --# Ignore last time we were damaged; 
            --# look for another hit from now on, 
            --# over the next 10 seconds
            self.Damaged = false
            WaitSeconds(self.WatchLoopPeriod)
            --# If we were not damaged again 
            --# within the last 10 seconds
            if (not self.Damaged) and 
                --# and there aren't enemy units in range if we include that 
                (not (self.PackIfEnemyUnitsNearby and self:CheckAreEnemyUnitsNearby()))
            then
                --# Unpause production
                self:SetScriptBit('RULEUTC_ProductionToggle', false)
            end
        end
    end,
  
  
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  Can be used by units to check if it's safe to unpack.
    --#*  The extending class should set self.PackIfEnemyUnitsNearby
    --#*  for this to be used.
    --#**
    CheckAreEnemyUnitsNearby = function(self)
        local Utilities = import('/lua/utilities.lua')
        local nearbyEnemyUnits = Utilities.GetEnemyUnitsInSphere(self, self:GetPosition(), 
                      self:GetBlueprint().Intel.VisionRadius) 
        --# If there were any enemy units nearby..
        for k, vUnit in nearbyEnemyUnits do
            if vUnit:GetBlueprint().Defense.SurfaceThreatLevel > 0.1 and
              not vUnit:IsThisUnitCloaked() then 
                return true
            end
        end
        
        --# No nearby units were dangerous
        return false
    end,
  
  
  
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  Called from OnStartBuild when Mex unit is given an upgrade command.   
    --#**
    PreparingToUpgradeState = State {
        Main = function(self)
       
            --# Try to wait for unit that plays animation loop by loop.
            --self:WaitForActiveAnimationManipulator()
            do--##### Was WaitForActiveAnimationManipulator function
                if self.ActiveAnimationManipulator then
                    --# Wait until we are at beginning of animation.
                    WaitFor(self.ActiveAnimationManipulator)
                    --# We have reached the end of the animation loop
                    self.ActiveAnimationLoopRunning = false
                else
                    WARN('ActiveAnimationUnit: PausedState: '
                      .. 'ActiveAnimationManipulator not defined.')
                end
            
                --# The rest is specific to those units that have active 
                --# animations but also animate (it or something else) to pack up.
                --# First, record if we are packed (or not) now that
                --# we have reached the end of another animation loop.
                self.IsPacked = self.PackedAtAnimationStart
                
                --# Next, check/record if we are ready to upgrade.
                if self.IsPacked and self.UpgradesFromPacked then 
                    self.NotReadyToUpgrade = false
                end
            end--##### Was WaitForActiveAnimationManipulator function
            
            --# Don't let shield run while upgrading.  
            --# This needs to be reenabled on OnFailedToBuild.
            self.ShieldUsePermitted = false
            
            --# Pack or unpack if we need to to be in correct position
            --# for smooth upgrade animation.  If we are already in
            --# position, this function call will do nothing.
            self:DoPackUpOperation(self.UpgradesFromPacked)
            
            --# Call base class code to restart upgrade.
            baseClassArg.PreparingToUpgradeState.Main(self)           
        end,
        
        --# Override to pause production when we get hit by a projectile
        OnDamage = function(self, instigator, amount, vector, damageType) 
            --# This is here as a safety measure.
            --# This shouldn't really happen as shield should be off by this point.
            if instigator == 'SensitiveShield' then
                WARN('SensitiveShield is instigatior in OnDamage when '
                  .. 'preparing to upgrade.')
            else
                --# Forwarding OnDamage to Unit class to reduce our health
                baseClassArg.OnDamage(self, instigator, amount, vector, damageType)
            end
        end,
    },
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is here just as a safety measure.
    --#*  This shouldn't really happen as shield should be off by this point.
    --#*  Without this check we'd get an error.
    --#*  Need to investigate why SensitiveShield is passing calls to
    --#*  OnDamage outside of the PausedState.
    --#**
    OnDamage = function(self, instigator, amount, vector, damageType) 
        if instigator == 'SensitiveShield' then
            WARN('SensitiveShield is instigatior in OnDamage when '
                  .. 'not in any other state.')
        else
            --# Forwarding OnDamage to Unit class to reduce our health
            baseClassArg.OnDamage(self, instigator, amount, vector, damageType)
        end
    end,
        
        
        
    
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  These functions are called when the weapon-toggle button
    --#*  (which I use as the auto-pause toggle button) using bit 1
    --#*  is pressed.  They are called from my override of OnScriptBitSet
    --#*  and OnScriptBitClear defined in my hook of Unit.lua.
    --#**
    OnScriptBit1Set = function(self)
        self.AutoPackToggledOn = true
    end,

    OnScriptBit1Clear = function(self)
        self.AutoPackToggledOn = false
    end,
    
}


return resultClass

end