--#****************************************************************************
--#**
--#**  New File :  /mods/gilbotsimprovements/extendedadjacencyunit.lua
--#**  Modded By:  Gilbot-X
--#**
--#**  Summary  :  My redefinition of the StructureUnit class.
--#**              This imported by defaultunits.lua and the 
--#**              StructureUnit class includes this as a base class.
--#**              This style allows me to separate code for defaultunits 
--#**              into a new file.  There was already too much code in that file!
--#**
--#****************************************************************************

local GilbotUtils = 
    import('/mods/gilbotsimprovements/utils.lua')
local ApplyPositionCorrection = 
    import('/mods/gilbotsimprovements/positioncorrections.lua').ApplyPositionCorrection
local EffectUtil = 
    import('/lua/EffectUtilities.lua')
local ResourceNetwork = 
    import('/mods/gilbotsimprovements/resourcenetwork.lua').ResourceNetwork
local GPGStructureUnit = 
    import('/mods/gilbotsimprovements/gpgstructureunit.lua').GPGStructureUnit
local AdjacencyBonusTypes = 
    import('/mods/gilbotsimprovements/adjacencybonustypes.lua').AdjacencyBonusTypes
local Buff = 
    import('/lua/sim/Buff.lua')
local AdjacencyBuffs = 
    import('/lua/sim/AdjacencyBuffs.lua')    
local SeraphimAdjacencyEffects = 
    import('/lua/EffectTemplates.lua').SAdjacencyAmbient01

local debugAdjacencyCode = true
    
ExtendedAdjacencyUnit = Class(GPGStructureUnit) {

--[[

Gilbot-X says: 
  
In the code that follows, the first 3 fields are settings that only take effect when your ACU has died.  They govern the maximum x or z separation between our unit skirt and the centre of position of a nearby unit that can be checked and then qualify for extended adjacency!

Note that this fetch rectangle is also used for checking if units have Diagonal adjacency, so if you set it to zero then you get no diagonal adjacency.  I recommend 5 as a minimum, that will make sure that all GPG units in patch 3260 can at least get diagonal adjacency.

Keeping this number as small as possible while not skipping any valid adjacency opportunities will optimise performance!

You need to allow for the larger skirt sizes of things you might want to catch in the rectangle, 
i.e. factories, T3 powergens, T3 arty and Quantum gates.

These are some example skirt sizes:
1 = T1 point defense
2 = T1 Powergen, T1 Massfab
6 = T2 Powergen, T3 Mass Fab
8 = Factory, T3 Powergen, T3 Arty
10 = Quantum Gateway
 
My settings:
  
    MaxRemoteAdjacencyAllowedWithoutACU_ModSetting = 8,
    MaxSkirtXFromCentre_ModSetting = 5, 
    MaxSkirtZFromCentre_ModSetting = 5, 
   
Setting above mean that you can give extended adjacency to a Quantum Gateway that is up to 8 o-grids away. Alternatively it means a T1 Massfab that is up to (5+8)-1 = 12 o-grids away.
    
The calculations assume you need half the skirt-size to get the unit's position in the fetch rectangle, because normally position is taken at (or very near) the centre of the skirt. It could even give diagonal adjacency to a unit with skirtsize of (5+8)*2 = 26!  That's 260% of a Quantum Gateway's size!!
 
As a rule of thumb, for skirt settings, take the largest skirt size of any structure in the game and set this value to half of that. There is a +1 added in the calculation just to be safe for units whose reference positions aren't perfectly in the centre of their skirts.

]]
    MaxRemoteAdjacencyAllowedWithoutACU_ModSetting = 8,
    MaxSkirtXFromCentre_ModSetting = 5, 
    MaxSkirtZFromCentre_ModSetting = 5, 
    
    AllowDiagonalAdjacency_ModSetting = true,
--[[    

The 4 fields above are mod settings.  The values can be changed.  See text above.


    
]]
    
    --# Constant variables are declared here.
    --# This next variable makes it easy to check which units 
    --# can get extended adjacency (and which can't).
    IsExtendedAdjacencyUnit = true,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  For debugging only.
    --#** 
    AdjacencyLog = function(self, messageArg)
        if debugAdjacencyCode then 
            LOG('Adjacency: a=' .. self.ArmyIdString .. ' ' .. messageArg)
        end
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I overrided OnCreate so I could 
    --#*  do variable initialization at this time.
    --#** 
    OnCreate = function(self)
        --# Perform original class version first
        GPGStructureUnit.OnCreate(self)
        
        --# Do non-static variable initialization
        self.AdjacentUnits = {
            GPG = {},
            Diagonal = {},
            Remote = {},
            Local = {},
        }
        
        --# T1 Point Defenses and AA Towers occupy same O-Grid as
        --# a T1 power generator so let them qualify for direct adjacency in same way
        if EntityCategoryContains(categories.TECH1 * categories.STRUCTURE * 
            (categories.ANTIAIR + categories.DIRECTFIRE), self) 
        then 
            --# Mark this unit as it needs special treatment
            self.IsT1PDorAA = true 
            if self:GetBlueprint().General.FactionName == 'Seraphim' then
                self.IsSeraphimT1PDorAA = true 
            end
        end
        
        --# This is for adjacency beam effects
        self.AdjacencyBeamsBag = {}
        
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if debugAdjacencyCode then 
            self.DebugId = self:GetUnitId() .. '_' .. self:GetEntityId()
            self.ArmyIdString = repr(self:GetArmy())
        end
    
        --# This needs to be called for ResourceNetworks to function.
        self:InitialiseBonusMarkers()
    
        --# This is the only time we check the BP for this value
        --# but every structure unit does it, so we don't need to do it again.
        --# (the slider mod might check it).
        self.RemoteAdjacencyAllowedSeparationDistance = 
            self:GetBlueprint().Adjacency.AdjacencyExtensionDistance or 0
            
        --# If the Pipeline is no longer working for us,
        --# the callback updates its status.
        self:AddOnCapturedCallback(self.CleanUpExtendedAdjacency)
        
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to be called by OnCreate so I could 
    --#*  do variable initialization at this time.
    --#** 
    InitialiseBonusMarkers = function(self)
    
        --# These tables are used by Resource Networks
        self.IsAdjacencyBonusReceiver = {}
        self.IsAdjacencyBonusGiver = {}
  
        --# How we add the unit depending on its type.
    	if self.IsPipeLineUnit then return end
        
        --# Populate IsAdjacencyBonusReceiver table 
        --# based on what this unit can receive
        self:MarkReceivableBonuses()
        
        --# This unit needs an Adjacency table in its blueprint
        --# in order to give any adjacency bonuses.
        local adjacencyBPName = self:GetBlueprint().Adjacency
        if not adjacencyBPName then return end
        for k,vBuffName in AdjacencyBuffs[adjacencyBPName] do
            local buffDef = Buffs[vBuffName]
            if not buffDef then
                error("*ERROR: Tried to add a buff that doesn\'t exist! Name: ".. buffName, 2)
                return 
            end
        
            local buffAffects = buffDef.Affects
            for bonusTypeName, vals in buffAffects do
                self.IsAdjacencyBonusGiver[bonusTypeName] = true
            end
        end
        
    end,
    
   
   
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  Populate the IsAdjacencyBonusReceiver table with bonus types 
    --#*  based on what adjacency bonuses this unit can receive.
    --#*
    --#*  No known issues with this function.
    --#**
    MarkReceivableBonuses = function(self)
        --# For some types of adjacency bonus, 
        --# unit blueprints are used for checking if the 
        --# unit can receive that type of bonus.
        --# Weapon and shield bonuses are the exception. 
        local bp = self:GetBlueprint()
      
        if (bp.Economy.BuildableCategory and 
            table.getn(bp.Economy.BuildableCategory) > 0 
           )
        or EntityCategoryContains(categories.SILO, self)
        then
            --# For debugging only, delete after
            if debugAdjacencyCode then 
                self:AdjacencyLog('Added active bonus on ' .. self.DebugId)
            end
            
            self.IsAdjacencyBonusReceiver['EnergyActive'] = true
            self.IsAdjacencyBonusReceiver['MassActive'] = true
        end
        
        if bp.Economy.ProductionPerSecondEnergy and 
           bp.Economy.ProductionPerSecondEnergy > 0 then
            self.IsAdjacencyBonusReceiver['EnergyProduction'] = true
        end
        
        if bp.Economy.ProductionPerSecondMass and 
           bp.Economy.ProductionPerSecondMass > 0 then
            self.IsAdjacencyBonusReceiver['MassProduction'] = true
        end
    
        if bp.Economy.MaintenanceConsumptionPerSecondEnergy or 
          self.EnergyMaintenanceConsumptionOverride then
            self.IsAdjacencyBonusReceiver['EnergyMaintenance'] = true
        end

        if bp.Economy.MaintenanceConsumptionPerSecondMass or
          self.MassMaintenanceConsumptionOverride then
            self.IsAdjacencyBonusReceiver['MassMaintenance'] = true
        end
        
        local weaponCount = self:GetWeaponCount()
        if weaponCount > 0 then
            self.IsAdjacencyBonusReceiver['RateOfFire'] = true
            for i = 1, weaponCount do
                local wep = self:GetWeapon(i)
                if wep:WeaponUsesEnergy() then
                    self.IsAdjacencyBonusReceiver['EnergyWeapon'] = true
                end
            end
        end
        
        --# Testing self.MyShield didn't work here - 
        --# perhaps a race condition where shield wasn't created yet?
        if EntityCategoryContains(categories.SHIELD, self) then 
            self.IsAdjacencyBonusReceiver['ShieldStrength'] = true
        end

    end,
    
    
        
        
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Had to override this to make sure that tables are 
    --#*  cleaned up when a unit completes an upgrade and 
    --#*  destroys itself.
    --#*
    --#*  This function is defined in Unit.lua
    --#*  and ends by changing state to DeadState
    --#*  so I have to call my extra code before calling
    --#*  the base class code.
    --#**
    OnDestroy = function(self)
        --# Gilbot-X says:  
        --# I added this block of code so that when 
        --# units upgrade they get cleaned out of ResourceNetwork 
        --# tables and the ACU's auto-toggle tables.
        self:CleanUpExtendedAdjacency() 
             
        --# Finally the rest is GPG code.
        GPGStructureUnit.OnDestroy(self)
    end,
    
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I overrided OnStopBeingBuilt so I could 
    --#*  check for extended adjacency at this time.
    --#** 
    OnStopBeingBuilt = function(self,builder,layer)
        --# Perform original class version first
        GPGStructureUnit.OnStopBeingBuilt(self,builder,layer)
        
         --# This will give us a MySkirtBounds variable.
        self:GetSkirtBounds()
        --# For debugging only
        if debugAdjacencyCode then self:DumpSkirtBounds() end
        
        --# If ACU is not alive then don't allow any more adjacency.
        if self:GetMyCommander() then 
            --# Try to add extended adjacency
            self:AddExtendedAdjacencyToNearbyUnits()
        end
    end,
    
    
    
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  by ResourceNetwork objects so that when they
    --#*  give this unit a reference to themselves,
    --#*  the unit can update its label if the display 
    --#*  is toggled on.
    --#**
    SetNetwork = function(self, networkArg, networkIdDisplayIsOn_Arg)
        
        --# Store the network reference set
        self.MyNetwork = networkArg
        
        --# Work out if the network display is on or not
        local networkIdDisplayIsOn = networkIdDisplayIsOn_Arg
        --# If we have a network, that overrides. 
        if self.MyNetwork then 
            networkIdDisplayIsOn = self.MyNetwork.NetworkNumbersDisplayed
        end
        
        --# Only update label if the network
        --# display is switched on.
        if networkIdDisplayIsOn then
            
            --# If this unit is in a valid network
            if self.MyNetwork and 
              (not self.MyNetwork.IsBroken)
            then
                --# Add the label
                self:DisplayNetworkNumbers(true)
                
            --# Otherwise, take old network label off
            else
                self:DisplayNetworkNumbers(false)
            end
        end
    end,
    
    
    
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  by ResourceNetwork objects they can update 
    --#*  this unit's network ID label if the display 
    --#*  is toggled on.  It is also called indirectly
    --#*  through calls to SetNetwork defined above.
    --#**
    DisplayNetworkNumbers = function(self, toggleOn)
        
        --# Put a label on if requested
        if toggleOn then
            --# Put network id in label
            local labelstring = 'N=' .. repr(self.MyNetwork.NetworkId) 
            
            --# Add internetwork ID if it has one
            if self.MyNetwork.MyInterNetwork and 
              not self.MyNetwork.MyInterNetwork.IsBroken 
            then 
                labelstring = 
                    labelstring .. ' I=' .. repr(self.MyNetwork.MyInterNetwork.InterNetworkId)
            end
            
            --# Show the label
            self:SetCustomName(labelstring)
            
        --# Otherwise, take any old label off
        --# i.e. restore to normal display.
        else
            self:ClearCustomName()
        end
    end,
    
    
    
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  from my override of OnStopBeingBuilt to apply
    --#*  adjacency as soon as the structures are built.
    --#*  This only runs once in a structure's lifetime.
    --#*  
    --#*  This version records which units our structure 
    --#*  has adjacency with already and adds diagonal 
    --#*  adjacency where allowed/applicable.
    --#**
    AddExtendedAdjacencyToNearbyUnits = function(self)
    
        local structureTable = self:GetNearbyFriendlyStructures()
        for kNearbyUnit, vNearbyUnit in structureTable do
                
            --# Next we will check if any skirt boundaries coincide.
            local hasGPGAdjacency, hasDiagonalAdjacency = 
                    self:CheckForDirectAdjacency(vNearbyUnit)
            
            --# We record these so we don't give them extended adjacency 
            --# because that would be giving them adjacency a second time!
            if hasGPGAdjacency then
                self:AddToEachOthersAdjacencyTables(vNearbyUnit, 'GPG')
            
            --# We give these extended adjacency if the mod setting is set.
            elseif hasDiagonalAdjacency and self.AllowDiagonalAdjacency_ModSetting then 
            
                --AdjacencyLog('Unit ' .. self.DebugId 
                --    .. ' is diagonally Adjacent to ' .. vNearbyUnit.DebugId)
                --# Make sure there are appropriate tables.
                self:AddToEachOthersAdjacencyTables(vNearbyUnit, 'Diagonal') 
                --# Apply adjacency as if these were actually adjacent
                self:OnAdjacentTo(vNearbyUnit)  
            end
	end
    end,
    
    
    
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  from AddExtendedAdjacencyToNearbyUnits below.
    --#*  It separates the code into resuable sections
    --#*  and makes the code more readable.
    --#**
    GetDirectAdjacencyBoundsForT1PDorAAStructure = function(self)
          
        --# Don't recalculate, as structures don't move!
        if not self.MyDirectAdjacencySkirtBounds then 
        
            --# This will give us a MySkirtBounds variable.
            local mySkirtBounds = self:GetSkirtBounds()
  
            --# Need to extend skirtbounds by 0.5 in each direction
            --# so they can link with each other and larger structures
            self.MyDirectAdjacencySkirtBounds = {
                UpperLeft = {
                    x = mySkirtBounds.UpperLeft.x  - 0.5,
                    z = mySkirtBounds.UpperLeft.z  - 0.5,
                },
                BottomRight = {
                    x = mySkirtBounds.BottomRight.x + 0.5,
                    z = mySkirtBounds.BottomRight.z + 0.5,
                },
            }
        end
        
        --# Debugging Only
        if debugAdjacencyCode then
          self:AdjacencyLog(self.DebugId
            .. ': My skirtbound2: UL:'
              .. ' x=' .. self.MyDirectAdjacencySkirtBounds.UpperLeft.x
              .. ' z=' .. self.MyDirectAdjacencySkirtBounds.UpperLeft.z .. ' BR:'
              .. ' x=' .. self.MyDirectAdjacencySkirtBounds.BottomRight.x
              .. ' z=' .. self.MyDirectAdjacencySkirtBounds.BottomRight.z)
        end
              
        --# This result is used by CheckForDirectAdjacency
        return self.MyDirectAdjacencySkirtBounds
    end,              
          
        
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  from AddExtendedAdjacencyToNearbyUnits below.
    --#*  It separates the code into resuable sections
    --#*  and makes the code more readable.
    --#**
    CheckForDirectAdjacency = function(self, nearbyUnitArg)
           
          --# Next we will check if any skirt boundaries coincide.
          local hasGPGAdjacency = false
          local hasDiagonalAdjacency = false
          
          --# If both structures are T1 PD or AA or neither of them are then
          if (self.IsT1PDorAA and nearbyUnitArg.IsT1PDorAA)
            or ((not self.IsT1PDorAA) and (not nearbyUnitArg.IsT1PDorAA))
          then
              --# Use their normal skirt bounds
              hasGPGAdjacency, hasDiagonalAdjacency = 
                  self.CheckForTouching(
                      self:GetSkirtBounds(), 
                      nearbyUnitArg:GetSkirtBounds()
                  )
          --# Otherwise if one is a T1 PD/AA but not the other,
          elseif self.IsT1PDorAA then
              --# Adjust the skirt bounds of the T1 PD/AA
              hasGPGAdjacency, hasDiagonalAdjacency = 
                  self.CheckForTouching(
                      self:GetDirectAdjacencyBoundsForT1PDorAAStructure(), 
                      nearbyUnitArg:GetSkirtBounds()
                  )
              --# GPG code doesn't automatically give adjacency from a pipeline to a T1 PD/AA
              --# so we mark it as diagonal adjacency so we manage it ourselves
              if hasGPGAdjacency and nearbyUnitArg.IsPipeLineUnit then 
                  hasGPGAdjacency, hasDiagonalAdjacency = false, true
              end
          else
              --# Adjust the skirt bounds of the T1 PD/AA
              hasGPGAdjacency, hasDiagonalAdjacency = 
                  self.CheckForTouching(
                      self:GetSkirtBounds(),
                      nearbyUnitArg:GetDirectAdjacencyBoundsForT1PDorAAStructure()
                  )
              --# GPG code doesn't automatically give adjacency from a pipeline to a T1 PD/AA
              --# so we mark it as diagonal adjacency so we manage it ourselves
              if hasGPGAdjacency and self.IsPipeLineUnit then 
                  hasGPGAdjacency, hasDiagonalAdjacency = false, true
              end
          end
          
          --# Respond to calling function
          return hasGPGAdjacency, hasDiagonalAdjacency
          
      end,
    
    
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  from AddExtendedAdjacencyToNearbyUnits below.
    --#*  It separates the code into resuable sections
    --#*  and makes the code more readable.
    --#**
    CheckForTouching = function(mySkirtBounds, nearbyUnitSkirtBounds)
           
          --# Next we will check if any skirt boundaries coincide.
          local oneSideIsFullyTouchingAnother = false
          local onlyCornersTouchOrPartOfOneSideIsTouchingAnother = false
          
          --# Check if the top of one touches the bottom of another
          if (mySkirtBounds.BottomRight.z == nearbyUnitSkirtBounds.UpperLeft.z) 
          or (mySkirtBounds.UpperLeft.z == nearbyUnitSkirtBounds.BottomRight.z) then
              --# if my left is equal to or right of your left,
              if (mySkirtBounds.UpperLeft.x >= nearbyUnitSkirtBounds.UpperLeft.x) then
                  --# and if my right is left or equal to your right
                  if (mySkirtBounds.BottomRight.x <= nearbyUnitSkirtBounds.BottomRight.x) then
                      --# I am smaller than you and within your skirtbounds.  
                      --# You will have received adjacency from GPG code.
                      oneSideIsFullyTouchingAnother = true
                  --# but if my left is at least left of your right or level with it
                  elseif (mySkirtBounds.UpperLeft.x <= nearbyUnitSkirtBounds.BottomRight.x) then
                      --# then we overlap or our corners touch 
                      onlyCornersTouchOrPartOfOneSideIsTouchingAnother = true
                  end
              end
              --# if your left is equal to or right of my left,
              if (nearbyUnitSkirtBounds.UpperLeft.x >= mySkirtBounds.UpperLeft.x) then
                  --# and if your right is left or equal to my right
                  if (nearbyUnitSkirtBounds.BottomRight.x <= mySkirtBounds.BottomRight.x) then
                      --# You are smaller than me and within my skirtbounds.  
                      --# You will have received adjacency from GPG code.
                      oneSideIsFullyTouchingAnother = true
                  --# but if your left is at least left of my right or level with it
                  elseif (nearbyUnitSkirtBounds.UpperLeft.x <= mySkirtBounds.BottomRight.x) then
                      --# then we overlap or our corners touch 
                      onlyCornersTouchOrPartOfOneSideIsTouchingAnother = true
                  end
              end

          --# Check if the left side of one touches the right side of another
          --# NB: Use elseif because the only way we can touch two sides is 
          --# to be touching the corner so we save ourselves checks and don't
          --# give structures with touching corners adjacency twice
          elseif (mySkirtBounds.BottomRight.x == nearbyUnitSkirtBounds.UpperLeft.x) 
          or (mySkirtBounds.UpperLeft.x == nearbyUnitSkirtBounds.BottomRight.x) then
              --# if my top is equal to or below your top,
              if (mySkirtBounds.UpperLeft.z >= nearbyUnitSkirtBounds.UpperLeft.z) then
                  --# and if my bottom is higher or level with your bottom
                  if (mySkirtBounds.BottomRight.z <= nearbyUnitSkirtBounds.BottomRight.z) then
                      --# I am smaller than you and within your skirtbounds.  
                      --# You will have received adjacency from GPG code.
                      oneSideIsFullyTouchingAnother = true
                  --# but if my top is at least above your bottom or level with it
                  elseif (mySkirtBounds.UpperLeft.z <= nearbyUnitSkirtBounds.BottomRight.z) then
                      --# then we overlap or our corners touch 
                      onlyCornersTouchOrPartOfOneSideIsTouchingAnother = true
                  end
              end
               --# if your top is equal to or below my top,
              if (nearbyUnitSkirtBounds.UpperLeft.z >= mySkirtBounds.UpperLeft.z) then
                  --# and if your bottom is higher or level with my bottom
                  if (nearbyUnitSkirtBounds.BottomRight.z <= mySkirtBounds.BottomRight.z) then
                      --# You are smaller than me and within my skirtbounds.  
                      --# You will have received adjacency from GPG code.
                      oneSideIsFullyTouchingAnother = true
                  --# but if your top is at least above my bottom or level with it
                  elseif (nearbyUnitSkirtBounds.UpperLeft.z <= mySkirtBounds.BottomRight.z) then
                      --# then we overlap or our corners touch 
                      onlyCornersTouchOrPartOfOneSideIsTouchingAnother = true
                  end
              end
          end
          
          --# Respond to calling function
          return oneSideIsFullyTouchingAnother, onlyCornersTouchOrPartOfOneSideIsTouchingAnother
          
      end,
    
    
  

    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  I added this.  It is called by AddExtendedAdjacencyToNearbyUnits
    --#*  defined above.  Each unit keeps a reference to the other.
    --#**
    AddToEachOthersAdjacencyTables = function(self, adjacentUnitArg, tableKeyArg)
        --# Each unit keeps a reference to the other
        table.insert(self.AdjacentUnits[tableKeyArg], adjacentUnitArg)
        table.insert(adjacentUnitArg.AdjacentUnits[tableKeyArg], self)
    end,
    
  
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  I added this.  It is required by OnNotAdjacentTo.
    --#*
    --#*  The last boolean argument to RemoveFromArrayByValue is optional.  
    --#*  It means don't destroy the empty table
    --#*  if last value is removed.  I think it's needed if we are 
    --#*  iterating through the table as it will stop the loop at the end of
    --#*  the table or when table is empty?
    --#**
    RemoveFromEachOthersAdjacencyTables = function(self, adjacentUnitArg, tableKeyArg)
    
        --# Take them out of our table.
        self.AdjacentUnits[tableKeyArg] = 
            GilbotUtils.RemoveFromArrayByValue(self.AdjacentUnits[tableKeyArg], 
                                                adjacentUnitArg, true)
        --# Take us out of their table.
        adjacentUnitArg.AdjacentUnits[tableKeyArg] = 
            GilbotUtils.RemoveFromArrayByValue(adjacentUnitArg.AdjacentUnits[tableKeyArg], 
                                                self, true)
    end,
    
    
    
     
    
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This overrided function is only called from inside
    --#*  function calls to OnAdjacentTo and OnNotAdjacentTo.
    --#*  It creates a single linking beam effect between two units.
    --#*
    --#*  I overrided this because GetEntityId doesn't work 
    --#*  in the way GPG tried to use it here.
    --#*  I had to remove code that tries to use the entity id as if it was a GUID.
    --#*  Unlike a GUID, entity IDs are reused as units die.
    --#*    
    --#*  Anyway it will only try to create an effect twice (once each way) maximum.
    --#*  If two units give and receive adjacency to each other, or are pipelines,
    --#*  then OnAdjacentTo is called twice.
    --#*
    --#**
    CreateAdjacentEffect = function(self, adjacentUnitArg)
                
        --# Create trashbag to hold all these entities and beams
        if not self.AdjacencyBeamsBag then
            WARN('Recreating adjacency beams bag.')
            self.AdjacencyBeamsBag = {}
        end
        
        --# This is an effect for seraphim only who don't get beams/hubs between T1 PD or AA
        if self.IsSeraphimT1PDorAA and adjacentUnitArg.IsSeraphimT1PDorAA then 
            self:CreateSeraphimPDToPDEffect()
            adjacentUnitArg:CreateSeraphimPDToPDEffect()
        else
            --# Make the beam.
            EffectUtil.CreateAdjacencyBeams(self, adjacentUnitArg, self.AdjacencyBeamsBag)
        end
    end,
    
   
   
   
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is required to get rid of beams between units when 
    --#*  the adjacency is switched off. It's only called from inside
    --#*  class to OnAdjacentTo and OnNotAdjacentTo.
    --#*
    --#*  No known issues with this function.
    --#*
    --#** 
    CleanUpAdjacentEffectsWith = function(self, adjacentUnitArg)
    
        --# Safety check
        if not adjacentUnitArg then return end
    
        --# Remove any beam effect from you to me.
        --# Make sure they have a beam bag
        if adjacentUnitArg.AdjacencyBeamsBag then 
            --# Iterate through it till we find our beam
            for kBeamBagEntry, vBeamBagEntry in adjacentUnitArg.AdjacencyBeamsBag do
                if vBeamBagEntry then 
                    if vBeamBagEntry.Unit == self then
                        --# Destroy our beam
                        vBeamBagEntry.Trash:Destroy()
                        adjacentUnitArg.AdjacencyBeamsBag[kBeamBagEntry] = nil
                    end
                end
            end
        end
        
        --# Now destroy any beam effects originating from me to you.
        if self.AdjacencyBeamsBag then
            --# Iterate through it till we find our beam
            for kBeamBagEntry, vBeamBagEntry in self.AdjacencyBeamsBag do
                if vBeamBagEntry then 
                    if vBeamBagEntry.Unit == adjacentUnitArg then
                        --# Destroy our beam
                        vBeamBagEntry.Trash:Destroy()
                        self.AdjacencyBeamsBag[kBeamBagEntry] = nil
                    end
                end
            end
        end
        
        --# Remove any reference to each other in the tables
        self:RemoveFromEachOthersAdjacencyTables(adjacentUnitArg, 'Local')
        self:RemoveFromEachOthersAdjacencyTables(adjacentUnitArg, 'GPG')
        self:RemoveFromEachOthersAdjacencyTables(adjacentUnitArg, 'Diagonal')
    	self:RemoveFromEachOthersAdjacencyTables(adjacentUnitArg, 'Remote')
        
        --# This is for Seraphim units only.  I added these effects.
        if self.HasT1PDAdjacencyEffect then 
            self:CleanUpSeraphimPDToPDAdjacencyEffects() 
        end
        if adjacentUnitArg.HasT1PDAdjacencyEffect then 
            adjacentUnitArg:CleanUpSeraphimPDToPDAdjacencyEffects()
        end
    end,
    
    
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  I give an effect to T1 Seraphim PD and AA units
    --#*  when they become adjacent to one another.
    --#*  They can be linked so closely that there isn't room
    --#*  to put hubs between them, so they themselves get
    --#*  a radiating effect to show they have adjacency
    --#*
    --#**
    CleanUpSeraphimPDToPDAdjacencyEffects = function(self)
   
        --# This links to code in effect until for seraphim
        if self.IsSeraphimT1PDorAA then
            --# Check if it's adjacent to another Seraphim T1 PD or AA
            local hasSeraphimT1PDInLocalAdjacencyTable = false
            for kUnusedArrayIndex, vUnit in self.AdjacentUnits['Local'] do
                if vUnit.IsSeraphimT1PDorAA then
                    hasSeraphimT1PDInLocalAdjacencyTable = true
                    break
                end
            end
            if not hasSeraphimT1PDInLocalAdjacencyTable then
                --# Allow effect to be added next time
               self:RemoveSeraphimPDToPDEffect()
            end
        end
    end,
   
   
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  I give an effect to T1 Seraphim PD and AA units
    --#*  when they become adjacent to one another.
    --#*  They can be linked so closely that there isn't room
    --#*  to put hubs between them, so they themselves get
    --#*  a radiating effect to show they have adjacency
    --#*
    --#**
    CreateSeraphimPDToPDEffect = function(self)
        --# Avoid duplicate effects
        if not self.HasT1PDAdjacencyEffect then
            self.SeraphimT1PDAdjacencyEffects = TrashBag()
            local army = self:GetArmy()
            for k, vEmit in SeraphimAdjacencyEffects do
                emit = CreateAttachedEmitter( self, 0, army, vEmit )
                self.SeraphimT1PDAdjacencyEffects:Add(emit)
                self.Trash:Add(emit)
            end
            self.HasT1PDAdjacencyEffect = true
        end
    end,
    
    
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  I give an effect to T1 Seraphim PD and AA units
    --#*  when they become adjacent to one another.
    --#*  They can be linked so closely that there isn't room
    --#*  to put hubs between them, so they themselves get
    --#*  a radiating effect to show they have adjacency
    --#*
    --#**
    RemoveSeraphimPDToPDEffect = function(self)
        if self.HasT1PDAdjacencyEffect then
            --# Destroy our effect
            self.SeraphimT1PDAdjacencyEffects:Destroy()
            self.SeraphimT1PDAdjacencyEffects = nil
            self.HasT1PDAdjacencyEffect = false
        end
    end,
    
    
    
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Had to mod this because OnNotAdjacentTo is called by the engine
    --#*  on the units it gave adjacency to, but not the ones I give    
    --#*  diagonal and remote adjacency to.
    --#*
    --#**
    CleanUpExtendedAdjacency = function(self)
        
        --# Stop this from being called twice on same object.
        if not self.WasAlreadyCalled_CleanUpExtendedAdjacency then

            --# For debugging only, delete after
            if debugAdjacencyCode then 
                self:AdjacencyLog('Cleaning Up Extended Adjacency on ' .. self.DebugId)
            end
            
            --# This unit won't be rejoining any networks
            self.DontRejoinNetwork = true
                
            --# Destroy beam effects from units giving me an adjacency bonus
            for kAdjacentUnit, vAdjacentUnit in self.AdjacentUnits['Diagonal'] do
                --# This is why I have a separate table for 'Diagonal' adjacency.
                --# because OnNotAdjacentTo always gets called by engine code
                --# on units that it decided were adjacent.  I have to call it 
                --# explicitly where I have decided additional units qualify.
                self:OnNotAdjacentTo(vAdjacentUnit)  
            end
            
            --# Destroy beam effects from units giving me an adjacency bonus
            for kAdjacentUnit, vAdjacentUnit in self.AdjacentUnits['Remote'] do
                self:OnNotAdjacentTo(vAdjacentUnit) 
            end
            
            --# Just for safety, 
            --# Destroy effects from me to any adjacent units I gave a bonus to.
            if self.AdjacencyBeamsBag then
                for k, v in self.AdjacencyBeamsBag do
                    v.Trash:Destroy()
                end
            end
            
            --# Stop this from being called twice on same object.
            self.WasAlreadyCalled_CleanUpExtendedAdjacency = true
            
        end
        
    end,
    
    
    
    
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I moved the code here from StructureUnit class and add an entry
    --#*  for my new type of adjacency.  Call on thsi unit to see what 
    --#*  bonus value it will give to the other unit (supplied as an argument)
    --#*  for the type of bonus specified (as the other argument).
    --#*
    --#*  The original version returned 1 when there was no bonus in 
    --#*  the blueprint.  Now it returns 0 because the ResourceNetwork 
    --#*  class actually sums up the values that this function returns
    --#*  to get a modifier to apply to all units in the network 
    --#*  of this type.
    --#*
    --#*  No known issues with this function.
    --#**
    GetAdjacentBonus = function(self, adjacentUnit, bonusTypeName)
        local resultAdd, resultMult = 0, 0
        
        --# Pauseable production units (such as MassFabs, HCPP and 
        --# shield strengthening units) do not give bonuses when output is paused.
        if self.IsProductionPaused then return resultAdd, resultMult end
        
        --# This unit needs an Adjacency table in its blueprint
        --# in order to give any adjacency bonuses.
        local adjacencyBPName = self:GetBlueprint().Adjacency
        if not adjacencyBPName then return resultAdd, resultMult end
        for k,vBuffName in AdjacencyBuffs[adjacencyBPName] do
            local buffDef = Buffs[vBuffName]
            if buffDef and buffDef.EntityCategory then
                local cat = ParseEntityCategory(buffDef.EntityCategory)
                if EntityCategoryContains(cat, adjacentUnit) then
                   
                    local buffAffects = buffDef.Affects
                    if buffAffects[bonusTypeName] then 
                        resultAdd = buffAffects[bonusTypeName].Add
                        resultMult = buffAffects[bonusTypeName].Mult
                        --# The self.ModifierDifference key is set by stat slider controls
                        --# so we don't actually have to change the blueprint values
                        if self.ModifierDifference then 
                            resultAdd = resultAdd + 
                                (self.ModifierDifference[buffDef.EntityCategory].Add or 0)
                            resultMult = resultMult + 
                                (self.ModifierDifference[buffDef.EntityCategory].Mult or 0)
                        end
                    end    
                end
            end
        end
        
        --# This is new for FA.
        --# It means that directly connected units give each other full bonuses
        --# but indirectly connected units on the same local network get
        --# half the bonus, where as units connected remotely on 
        --# separate local networks give 25% of full bonus.
        local attenuationFactor = 1
        if not self:CheckIfAlreadyConnected(adjacentUnit, 'Local') then 
            if self.MyNetwork.NetworkId == adjacentUnit.MyNetwork.NetworkId
            then attenuationFactor = 0.5
            else attenuationFactor = 0.25
            end
        end
         
        --# We should have worked out the bonus now.
        return resultAdd * attenuationFactor, math.pow(resultMult, attenuationFactor)
    end,
    
    
    
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I overrided this function so we are not adding
    --#*  or taking away from the previous bonus, we are
    --#*  setting it outright.  I can do this because
    --#*  each unit can only receive bonuses from one source: 
    --#*  i.e. the network it is attached to.
    --#*
    --#*  No known issues with this function.
    --#**
    SetAdjacencyBonus = function(self, bonusTypeName, bonusAdd)
        
        if bonusTypeName == 'MassActive' then
            self.MassBuildAdjMod = 1 + bonusAdd
            self:UpdateConsumptionValues()
        
        elseif bonusTypeName == 'EnergyActive' then
            self.EnergyBuildAdjMod = 1 + bonusAdd
            self:UpdateConsumptionValues()
        
        elseif bonusTypeName == 'MassMaintenance' then
            self.MassMaintAdjMod = 1 + bonusAdd
            self:UpdateConsumptionValues()
        
        elseif bonusTypeName == 'EnergyMaintenance' then
            self.EnergyMaintAdjMod = 1 + bonusAdd
            self:UpdateConsumptionValues()
        
        elseif bonusTypeName == 'MassProduction' then
            self.MassProdAdjMod = 1 + bonusAdd
            self:UpdateProductionValues()
        
        elseif bonusTypeName == 'EnergyProduction' then
            self.EnergyProdAdjMod = 1 + bonusAdd
            self:UpdateProductionValues()
        
        elseif bonusTypeName == 'EnergyWeapon' then
            for i = 1, self:GetWeaponCount() do
                local wep = self:GetWeapon(i)
                if wep:WeaponUsesEnergy() then
                    wep.AdjEnergyMod = 1 + bonusAdd
                end
            end
            
        elseif bonusTypeName == 'RateOfFire' then
            for i = 1, self:GetWeaponCount() do
                local wep = self:GetWeapon(i)
                local wepbp = wep:GetBlueprint()
                local weprof = wepbp.RateOfFire
                local delay = 1/wepbp.RateOfFire
                local newdelay = (1+bonusAdd) * delay
                local newRateOfFire = 1/newdelay 
                wep:ChangeRateOfFire(newRateOfFire)
                wep.LastRateOfFireSet = newRateOfFire
            end
            
        --# This is a new type of bonus I added.
        elseif bonusTypeName == 'ShieldStrength' then
            if self.MyShield then
                local oldShieldStrengthMod = self.ShieldStrengthMod
                self.ShieldStrengthMod = 1 + bonusAdd
                                 
                --# If the adjacency bonus on shield strength changed
                if true and (oldShieldStrengthMod ~= self.ShieldStrengthMod) then
                    --# Show message to user
                    self.MyShield:ShowShieldStrengthMessage()
                end
  
            end
        end
    end,
    
    

    
    
    
    
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Had to mod this to make sure tables are cleaned up.
    --#*  Base class version cancels bonuses and cleans up effects.
    --#*  
    --#**
    OnAdjacentTo = function(self, adjacentUnitArg, triggerUnit)
        
        --# Safety checks are gropuped together in another function
        if not self:CanLinkWith(adjacentUnitArg) then return end

        --# Call code that deals with resource networks and internetworks
        self:AddLinkToNetwork(adjacentUnitArg)
        
        --# Create link effect and register connection with connected units
        self:CreateLink(adjacentUnitArg)
        
        --# This (amongst other purposes) will stop this being executed more than once. 
        self:AddToEachOthersAdjacencyTables(adjacentUnitArg, 'Local')
    end,

    
    
    
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Called by OnAdjacentTo.
    --#*  Contains safety checks that can abort linking in OnAdjacentTo.
    --#**
    CanLinkWith = function(self, adjacentUnitArg)
    
        --# This is a to protect against applying GPG adjacency
        --# to units that do not inherit this class.
        if not adjacentUnitArg.IsExtendedAdjacencyUnit then return false end
        
        --# This is a safety GPG uses to protect against
        --# duplicate applications of GPG adjacency.
        if self:IsBeingBuilt() then return false end
        if adjacentUnitArg:IsBeingBuilt() then return false end
        
        --# This next call is a safety against duplicate links.
        if self:CheckIfAlreadyConnected(adjacentUnitArg, 'Local') then return false end
        if self:CheckIfAlreadyConnected(adjacentUnitArg, 'Remote') then return false end
        
        --# Extra safety
        if self:BeenDestroyed() or adjacentUnitArg:BeenDestroyed() then return false end
        
        --# If ACU is not alive then don't allow any more adjacency.
        if not self:GetMyCommander() then return false end
        
        --# Passed all safety checks
        return true
    end,
    
    
 
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Called by OnAdjacentTo.
    --#*  Deals with ResourceNetwork objects.
    --#**
    AddLinkToNetwork = function(self, adjacentUnitArg)
        
        --# For safety, although not sure this is necessary.
        self.DontRejoinNetwork = false
        adjacentUnitArg.DontRejoinNetwork = false
        
        --# Only create and associate with a brand new network supervsior
        --# if we haven't had one already or the last one was broken up and is invalid.
        local useMyNetwork = true
        --# First check if we have a valid network.
        if not self.MyNetwork or self.MyNetwork.IsBroken then
            --# if the adjacent unit has a functioning network
            if adjacentUnitArg.MyNetwork and (not adjacentUnitArg.MyNetwork.IsBroken) then
                --# Add it to the adjacent unit's network because 
                --# has a valid one and we don't.
                adjacentUnitArg.MyNetwork:AddLinkToNetwork(adjacentUnitArg, self)
                useMyNetwork = false
            else
                --# We'll make a new network for this unit
                --# because neither unit had a network.
                --# This is the only time/place we create
                --# a new network.                 
                self:SetNetwork(ResourceNetwork(self))
            end
        end
        --# Add the link to this unit's (possibly new) network if we didn''t
        --# already add it to the other unit's network.
    	if useMyNetwork then self.MyNetwork:AddLinkToNetwork(self, adjacentUnitArg) end
    
    end,
        
        
    
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Called by OnAdjacentTo.
    --#*  Deals with ResourceNetwork objects.
    --#**
    CreateLink = function(self, adjacentUnitArg)
    
        --# Always create/remove a linking beam effect regardless
        --# of whether or not bonus is applied.
        --# Create adjacency beam effect
    	ForkThread(
            function(self, adjacentUnitArg) 
                --# You get errors if it tries to create effects too quickly.
                WaitSeconds(1)
                self:CreateAdjacentEffect(adjacentUnitArg)
            end, self, adjacentUnitArg
        )
     
    end,
    
    
    
    
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Had to mod this to make sure tables are cleaned up.
    --#*  Base class version cancels bonuses and cleans up effects.
    --#**
    OnNotAdjacentTo = function(self, adjacentUnitArg)

        --# Safety checks are gropuped together in another function
        if not self:CanDisconnectFrom(adjacentUnitArg, 'Local') then return end
        
        --# We break the network if we are removing a bridging link
        local unitRemoved1, breakNetwork1 = self:RemoveMeFromMyNetworkIfIAmDead()
        local unitRemoved2, breakNetwork2 = adjacentUnitArg:RemoveMeFromMyNetworkIfIAmDead()    
        local unitRemoved = unitRemoved1 or unitRemoved2
        local breakNetwork = breakNetwork1 or breakNetwork2
        
        --# Always remove a linking beam effect regardless
        --# of whether or not bonus is applied:
        --# Remove any adjacency beam effects between this pair.
        --# Also removes references to each other from all connection tables,
        --# which stops this function from being called a second time on the same pair.
        self:CleanUpAdjacentEffectsWith(adjacentUnitArg)
        
        
        --# Both units will have same network object becausde they were attached
        local network = self.MyNetwork or adjacentUnitArg.MyNetwork
        --# This call will tell the network to update itself after units are lost.
        --# If there are units left in it, the network will either recalulate bonuses 
        --# or, if told it must by the argument, it will split into subnetworks
        --# and then recalculate bonuses.  If there are no units left in it, it just dies.
        if network and unitRemoved then network:OnLinkRemovedFromNetwork(breakNetwork) end
        
    end,
    
    
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Called by OnNotAdjacentTo.
    --#*  Contains code that would have been 
    --#*  repeated if it was not kept separately.
    --#**
    RemoveMeFromMyNetworkIfIAmDead = function(self)
    
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if false then
          self:AdjacencyLog('OnNotAdjacentTo: ' 
           .. ' Unit ' .. self.DebugId
           .. ' has health '.. repr(self:GetHealth()) 
           .. ' IsDead=' .. repr(self:IsDead())  
          )
        end
    
        --# If this unit is dead and hasn't already been taken out of its network,
        --# We will contact the network to remove this unit from network.
        --# First check if we have the correct variables to make the call.
    	if self.MyNetwork then 
            --# When a unit is destroyed, the OnKilled event is processed after this event, 
            --# so we must check the health of the unit.
            if self:BeenDestroyed() or self:IsDead() 
              or (self:GetHealth() <= 0) then
                self.DontRejoinNetwork = true
                self.MyNetwork:RemoveUnitFromNetwork(self) 
                --# We record that we must break the network 
                --# later if we are removing a bridging link.
                self:CheckIfUnitIsBridgeInNetwork()
                return true, self.IsNonRedundantBridgeInNetwork
            end
        end
        
        --# Otherwise unit was not removed.  Don't call
        --# OnLinkRemovedFromNetwork and don't break network.
        return false, false
        
    end,
        
        
    
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Called by OnNotAdjacentTo.
    --#*  Contains safety checks that can abort discionnecting in OnNotAdjacentTo.
    --#**
    CanDisconnectFrom = function(self, adjacentUnitArg, connectionTypeArg)
    
        --# This is a to protect against trying to remove GPG adjacency
        --# from units that do not inherit this class.
        if not adjacentUnitArg.IsExtendedAdjacencyUnit then return false end
        
        --# These next call is a safety against duplicate calls.
        if not self:CheckIfAlreadyConnected(adjacentUnitArg, connectionTypeArg) then return false end
    
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if debugAdjacencyCode then
            self:AdjacencyLog('CanDisconnectFrom: ' .. self.DebugId
              .. ' can disconnect linktype=' .. connectionTypeArg 
              .. ' from ' .. adjacentUnitArg.DebugId
            )
        end
        
        --# Passed all safety checks
        return true
    end,
    
    
    
  
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  from RebuildRemoteAdjacencyConnections.
    --#*  It separates the code into resuable sections
    --#*  and makes the code more readable.
    --#*
    --#*  It's also used as a safety in OnNotAdjacentTo.
    --#*
    --#**
    CheckIfAlreadyConnected = function(self, nearbyUnitArg, tableKeyArg)
        return
          (        
            GilbotUtils.IsValueInTable(self.AdjacentUnits[tableKeyArg], nearbyUnitArg)
              or 
            GilbotUtils.IsValueInTable(nearbyUnitArg.AdjacentUnits[tableKeyArg], self)
          )
    end,
    
    
    
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  from IsNonRedundantBridgeInNetwork below.
    --#**
    GetNumberOfConnectedUnits= function(self, tableKeyArg)
        return table.getsize(self.AdjacentUnits[tableKeyArg])
    end,
    
    
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  from the ResourceNetwork class.
    --#**
    CheckIfUnitIsBridgeInNetwork = function(self)
        
        --# Need to be connected to at least 2 other units to be a bridge.
        if self:GetNumberOfConnectedUnits('Local') < 2 then 
            self.IsNonRedundantBridgeInNetwork = false
            return false 
        end
        
        --# For every unit we are connected to in this network,
        for unusedArrayIndex1, vAdjacentUnit1 in self.AdjacentUnits['Local'] do
            --# For every other unit we can pair it with in this network,
            for unusedArrayIndex2, vAdjacentUnit2 in self.AdjacentUnits['Local'] do
                --# (Don't pair a unit with itself)
                if vAdjacentUnit1 ~= vAdjacentUnit2 then
                    --# If these units are not connected to each other directly then
                    if not vAdjacentUnit1:CheckIfAlreadyConnected(vAdjacentUnit2, 'Local') then 
                        
                        --# We found a bridge.
                        --# This next block is for debugging only.
                        --# This can be delete wheh debugging is done.
                        if false and debugAdjacencyCode then 
                            self:AdjacencyLog('Unit ' .. self.DebugId 
                            .. ' is a bridge in a network' 
                            .. ' between ' .. vAdjacentUnit1.DebugId 
                            .. ' and ' .. vAdjacentUnit2.DebugId
                            )
                        end
                        
                        --# Check if these two units are bridged indirectly by another unit.
                        if not self:CheckIfUnitsHaveAlternateBridgeInNetwork(
                                  vAdjacentUnit1, vAdjacentUnit2
                               ) 
                        then
                            --# Alert user, calling code and record result
                            self.IsNonRedundantBridgeInNetwork = true
                            --# This next block is for debugging only.
                            --# This can be delete wheh debugging is done.
                            if false and debugAdjacencyCode then 
                                self:AdjacencyLog('Unit ' .. self.DebugId 
                                .. ' is Non-redundant bridge in a network '
                                .. ' between ' .. vAdjacentUnit1.DebugId 
                                .. ' and ' .. vAdjacentUnit2.DebugId
                                )
                            end
                            return true
                        end
                    end
                end
            end
        end
        
        --# Didn't find any bridges
        self.IsNonRedundantBridgeInNetwork = false
        return false
    end,
         

         
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  from the ResourceNetwork class.
    --#**
    CheckIfUnitsHaveAlternateBridgeInNetwork = function(self, unitArg1, unitArg2)
        
        --# look at every unit that the first node is connected to in this network,
        --# as each one can be a bridge from the first unit to the second.
        for unusedArrayIndex1, vPotentialBridgingUnit in unitArg1.AdjacentUnits['Local'] do
            --# (Ignore the first bridge found before calling this function)
            if vPotentialBridgingUnit ~= self then
                --# Look at all units it could be connecting indirectly to via this bridge
                for unusedArrayIndex2, vIndirectlyConnectedUnit in
                                              vPotentialBridgingUnit.AdjacentUnits['Local'] do
                    --# If this potential bridge is indirectly connecting 
                    --# the first unit to the second, then                     
                    if vIndirectlyConnectedUnit == unitArg2 then
                        
                        --# This next block is for debugging only.
                        --# This can be delete wheh debugging is done.
                        if false and debugAdjacencyCode then 
                            self:AdjacencyLog('Unit ' .. vPotentialBridgingUnit.DebugId 
                            .. ' Is alternate bridge bewteen units ' .. unitArg1.DebugId
                            .. ' and ' .. unitArg2.DebugId .. ' In Network.')
                        end
                        --# There is an alternate bridge in the network,
                        --# in other words, we are a *redundant* bridge.
                        return true
                    end
                end
            end
        end
        
        --# Didn't find any alternate bridges
        return false
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  I added this.  It is called by 
    --#   AddExtendedAdjacencyToNearbyUnits
    --#*  and RebuildRemoteAdjacencyConnections.
    --#*
    --#*  No known issues with this function.
    --#*
    --#**
    GetNearbyFriendlyStructures = function(self)
    
        local maxSkirtXFromCentre, maxSkirtZFromCentre = 4, 4
        local offset = self.RemoteAdjacencyAllowedSeparationDistance
    
        --# Experimental approach to improving efficiency 
        --# that works while ACU is alive : Need to test ???????
        local myCommander = self:GetMyCommander()
        --# If commander is alive...
        if myCommander then
            --# Give this to ACU who will decide if thsi is the largest range
            myCommander:ReceiveACUMessage_SetMaxAdjacencyRangeEncounteredSoFar(
                self.RemoteAdjacencyAllowedSeparationDistance
            )
            --# ACU tells us the largest offset this army is capable of.
            offset = 
                math.max(myCommander:ReceiveACUMessage_GetMaxAdjacencyRangeEncounteredSoFar() or 0, offset)
            
            --# ACU tells us highest values for skirt sizes
                maxSkirtXFromCentre, maxSkirtZFromCentre = 
                    myCommander:ReceiveACUMessage_GetMaxSkirtBoundsEncounteredSoFar()
        else
            --# Without ACU alive, use the defaults defined at the top of the file, 
            --# unless the unit looking for adjacency has higher values itself.
            --# Thus we see the new role of ACU as information coordinator
            --# who can improve army efficiency.  Some connections might not be made
            --# automatically without the ACU alive.
            offset = math.max(offset, self.MaxRemoteAdjacencyAllowedWithoutACU_ModSetting)
            maxSkirtXFromCentre = math.max(maxSkirtXFromCentre, self.MaxSkirtXFromCentre_ModSetting)
            maxSkirtZFromCentre = math.max(maxSkirtZFromCentre, self.MaxSkirtZFromCentre_ModSetting)
        end
        
        --# The table variable we will return.
        local resultTable = {}
        
        --# This will give us the MySkirtBounds variable to use below.
        mySkirtBounds = self:GetSkirtBounds()
 
        --# Get nearby units to check for touching adjacency
	local x1 = mySkirtBounds.UpperLeft.x - (1+offset+maxSkirtXFromCentre)
	local z1 = mySkirtBounds.UpperLeft.z - (1+offset+maxSkirtZFromCentre)
	local x2 = mySkirtBounds.BottomRight.x + (1+offset+maxSkirtXFromCentre)
	local z2 = mySkirtBounds.BottomRight.z + (1+offset+maxSkirtZFromCentre)
	local UnitsinRec = GetUnitsInRect( Rect(x1, z1, x2, z2) ) or {}
        
        --# Now filter through the units found (shouldn't be too many)
        for kNearbyUnit, vNearbyUnit in UnitsinRec do
        
            --# Don't connect to ourself!!
            if self ~= vNearbyUnit and 
              --# unit has to be in same army
              self:GetArmy() == vNearbyUnit:GetArmy()
              --# Both units must inherot this class!
              and vNearbyUnit.IsExtendedAdjacencyUnit
            then
                --# Check that it's not a structure contained within a structure,
                if (not self.ContainingStructure) or 
                    --# or if it is then that we are not trying to 
                    --# connect to our containing unit!
                   (self.ContainingStructure ~= vNearbyUnit)
                then  
                    --# This unit qualifies
                    table.insert(resultTable, vNearbyUnit)
                end
            end
        end
        
        --# Return result to calling code
        return resultTable
        
    end,
        
        
        
    
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  I added this.  It is called by AddExtendedAdjacencyToNearbyUnits
    --#*  defined above.
    --#*
    --#*  No known issues with this function.
    --#*
    --#**
    GetSkirtSeparationDistance = function(self, adjacentUnitArg)
          
          local xSeparation = 0
          local zSeparation = 0
          
          local SB1 = self:GetSkirtBounds()
          local SB2 = adjacentUnitArg:GetSkirtBounds()
          
          --# If my top is below your bottom
          if SB1.UpperLeft.z > SB2.BottomRight.z then
              --# I am below you
              zSeparation = SB1.UpperLeft.z - SB2.BottomRight.z
          --# but if your top is below my bottom
          elseif SB2.UpperLeft.z > SB1.BottomRight.z then
              --# You are below me!!
              zSeparation = SB2.UpperLeft.z - SB1.BottomRight.z
          end 
          
          --# If my left side is to the right of your right side
          if SB1.UpperLeft.x > SB2.BottomRight.x then
              --# I am to the right of you
              xSeparation = SB1.UpperLeft.x - SB2.BottomRight.x
          --# But if your left side is to the right of my right side
          elseif SB2.UpperLeft.x > SB1.BottomRight.x then
              --# You are to the right of me
              xSeparation = SB2.UpperLeft.x - SB1.BottomRight.x
          end
          
          --# Use Pythagoras Theorem to find separation
          return math.sqrt(math.pow(xSeparation,2) + math.pow(zSeparation,2))
              
    end,
      
              
          
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  from AddExtendedAdjacencyToNearbyUnits below.
    --#*  It is to save on repeated code.
    --#*
    --#*  No known issues with this function.
    --#*
    --#**
    GetSkirtBounds = function(self)
        
        --# Don't recalculate, as structures don't move!
        if not self.MySkirtBounds then 
          
            local myBP = self:GetBlueprint()
    
            --# Debugging only
            if (not myBP.Physics.SkirtSizeX) or 
               (not myBP.Physics.SkirtSizeZ) then
                   WARN('Adjacency: unit ' .. self:GetUnitId() .. 
                   ' does not have SkirtSize Defined.')
            end
                
            --# bounding box overrided by skirt size if set
            local myBPSkirtXFromCentre = (myBP.Physics.SkirtSizeX or myBP.SizeX) * 0.5
            local myBPSkirtZFromCentre = (myBP.Physics.SkirtSizeZ or myBP.SizeZ) * 0.5
            
            --# Correction for custom units
            self.MyCorrectedPosition = ApplyPositionCorrection(self, self:GetPosition())
            
            --# Get edge corner positions, { TOP, LEFT, BOTTOM, RIGHT }
            self.MySkirtBounds = {
                UpperLeft = {
                    z = self.MyCorrectedPosition.z - myBPSkirtZFromCentre,
                    x = self.MyCorrectedPosition.x - myBPSkirtXFromCentre,
                },
                BottomRight = {
                    z = self.MyCorrectedPosition.z + myBPSkirtZFromCentre,
                    x = self.MyCorrectedPosition.x + myBPSkirtXFromCentre,
                }
            }
            
            --# New approach to improving efficiency 
            --# that works while ACU is alive.
            local myCommander = self:GetMyCommander()
            --# If commander is alive...
            if myCommander then
                --# Give new skirt values to ACU 
                --# who will then decide if these are the largest so far
                myCommander:ReceiveACUMessage_SetMaxSkirtBoundsEncounteredSoFar(
                    myBPSkirtXFromCentre, myBPSkirtZFromCentre 
                )
            end
                
        end
        
        --# Return value just in case a copy is needed.
        --# Might make calling code easier to read.        
        return self.MySkirtBounds
    end,
    
    
    
    
    
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  from AddExtendedAdjacencyToNearbyUnits below.
    --#*  It is only for debugging.
    --#*
    --#*  No known issues with this function.
    --#*
    --#**
    DumpSkirtBounds = function(self)
        
        local myposition = self:GetPosition()
        --[[
        self:AdjacencyLog(self.DebugId
            .. ': position:'
            .. ' x=' .. myposition.x 
            .. ' z=' .. myposition.z
            .. '    correted position:'
            .. ' x=' .. self.MyCorrectedPosition.x 
            .. ' z=' .. self.MyCorrectedPosition.z
            .. '    skirt sizes:'
            .. ' x=' .. self:GetBlueprint().Physics.SkirtSizeX 
            .. ' z=' .. self:GetBlueprint().Physics.SkirtSizeZ)
        ]]
        self:AdjacencyLog(self.DebugId
            .. ': My skirtbounds: UL:'
            .. ' x=' .. self.MySkirtBounds.UpperLeft.x
            .. ' z=' .. self.MySkirtBounds.UpperLeft.z .. ' BR:'
            .. ' x=' .. self.MySkirtBounds.BottomRight.x
            .. ' z=' .. self.MySkirtBounds.BottomRight.z)
        
        if math.mod(myposition.x,1) ~= 0.5 then
            if not self.IsPipeLineUnit then 
              WARN('Adjacency: unit ' .. self:GetUnitId() .. 
                   ' does not have valid x position.')
            end
        end
        if math.mod(myposition.z,1) ~= 0.5 then
            if not self.IsPipeLineUnit then 
              WARN('Adjacency: unit ' .. self:GetUnitId() .. 
                   ' does not have valid z position.')
            end
        end
        
    end,
  
}