do --(start of non-destructive hook)
--#****************************************************************************
--#**
--#**  Hook File:  /lua/sim/unit.lua
--#**
--#**  Summary  :  The Unit lua module
--#**
--#**  Modded By:  Gilbot-X, with some code from CovertJaguar
--#**
--#**  Changes  : 
--#**    Maintenance Consumption Breakdown code added.
--#**    Adjacency code added.
--#**    Cloak Effect code added (adapted from code by CovertJaguar)
--#**
--#**  Notes    :  
--#**    The original version of unit.lua (in patch 3269) has 3608 lines.
--#**    To find line that caused error, subtract line number given by 
--#**    log file by 3649 to get the line number you need to look at 
--#**    in this file.
--#**
--#****************************************************************************

--# This flag switches on a debugging feature
--# where units have their entity ID as a label
--# that can be used for the custom naming system.
local debugAllUnits = false

local PreviousVersion = Unit

Unit = Class(PreviousVersion) {

    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I overrided OnCreate so I could 
    --#*  do variable initialization at this time.
    --#** 
    OnCreate = function(self)
        --# Perform original class version first
        PreviousVersion.OnCreate(self)
        
         --# This next block is for network display system.
        self.BaseLabel = nil
        --# This next block is for debugging only.
        --# This can be delete when debugging is done.
        if debugAllUnits then 
            self.BaseLabel = 'e' .. self:GetEntityId()
            self:ClearCustomName()
        end
    end,
    
        
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I overrided SetCustomName so I could 
    --#*  call GetCustomName from the user side.
    --#** 
    SetCustomName = function(self, stringArg)
        --# The empty string is always converted to nil
        --# False is also converted to nil.  There must be
        --# only one value that means 'no custom name', 
        --# otherwise this code will break because of the next 
        --# comparison which uses the ~= operator.
        if (not stringArg) or stringArg == "" then stringArg = nil end
        --# Try to keep the unit's original name
        --# as long as we can.
        --# Don't update if new value is the same as old        
        if stringArg ~= self.CurrentCustomName then
            --# Don't actually change display name
            --# yet if a message is being flashed.
            --# The message thread will do that for us
            --# when it is finished if we just store the label now.
            if not self.FlashMessageLabel then
                --# Original class version is a moho method
                local stringArgToPass = (stringArg or "")
                if type(stringArgToPass) ~= "string" then 
                    WARN('Unit.lua: SetCustomName: argument is not a string')
                    return
                end
                PreviousVersion.SetCustomName(self, stringArgToPass)
            end
            --# Store value so we can query from sim side
            --# The empty string is always converted to nil
            --# so this can never be ""
            self.CurrentCustomName = stringArg
        end
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I created a version of this to query from sim side.
    --#*  This function already exists on Class UserUnit.
    --#** 
    GetCustomName = function(self)
        --# We kept this variabe so we can query from sim side
        return self.CurrentCustomName 
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I overrided OnCreate so I could 
    --#*  do variable initialization at this time.
    --#** 
    ClearCustomName = function(self)
        self:SetCustomName(self.BaseLabel)
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This is called to flash a message on a unit that 
    --#*  a player can see.
    --#* 
    --#*  Calling this temporarily delays other threads 
    --#*  from changing the custom name text that you see.
    --#** 
    FlashMessage = function(self, stringArg, secondsArg)

        --# Do this for safety
        if not (type(stringArg)=="string") then return end
        if not (type(secondsArg)=="number") then secondsArg=2 end
        
        --# Show our label for a couple seconds
        ForkThread(
            function(self, stringArg, secondsArg)
                
                --# Wait for any FlashMessage thread 
                --# ahead of it to finish.
                while(self.FlashMessageLabel) do
                    WaitSeconds(secondsArg)
                end
                
                --# Setting this stops other threads from 
                --# changing the label until we are done.
                self.FlashMessageLabel = stringArg
                
                --# Original class version is a moho method
                --# We call that to chnage the display label on the unit
                PreviousVersion.SetCustomName(self, self.FlashMessageLabel)
            
                --# Leave it showing according to what 
                --# calling code specified
                WaitSeconds(secondsArg)
                
                --# Now try to restore whatever the label should be now.
                --# Unit may have been destroyed in the meantime
                if not (self:BeenDestroyed() or self:IsDead()) then 
                    --# Original class version is a moho method
                    PreviousVersion.SetCustomName(self, (self:GetCustomName() or ""))
                    --# Setting this to nil allows
                    --# other threads to change the display name again.
                    self.FlashMessageLabel = nil
                end
            end, self, stringArg, secondsArg
        )
    end,
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This is called from file debuggingcallbacks.lua
    --#*  when a player selects units and hits ALT-F.
    --#*  It displays each unit's Rate Of Fire over the
    --#*  unit (as its custom name) for 3 seconds.
    --#** 
    ShowRateOfFireMessage = function(self)
        local weaponCount = self:GetWeaponCount()
        if weaponCount == 0 then return end
        local messageText = 'ROF='
        --# This block will make a comma separated list
        --# listing the ROF of all wepons this unit has
        for i = 1, weaponCount do
            --# Get weapon object from SIM
            local weapon = self:GetWeapon(i)
            --# Look for cache of last rate of fire set from 
            --# a buff or adjacency bonus
            local weaponROF = weapon.LastRateOfFireSet or false
            if not weaponROF then
                --# Otherwise use default value from its blueprint
                weaponROF = weapon:GetBlueprint().RateOfFire
            end
            --# Format number as a string with 2 d.p.
            local weaponROFtext = self.NumberToStringWith2DPMax(weaponROF)
            messageText = messageText .. weaponROFtext
            --# Add a comma to list if there's another weapon to do
            if i < weaponCount then messageText = messageText .. ', ' end
        end
        
        --# Notify user
        self:FlashMessage(messageText, 3)
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This is called from ShowRateOfFireMessage above.
    --#*  It formats a number as a string up to 2 d.p.
    --#*  and returns it.
    --#** 
    NumberToStringWith2DPMax = function(numberArg)
        --# safety check first
        if not type(numberArg) == 'number' then return end
        local resultText = repr(numberArg)
        if string.find(resultText, "%d+.%d%d") then 
            local s = string.gfind(resultText, "%d+.%d%d")
            resultText = s()
        elseif string.find(resultText, "%d+.%d") then 
            local s = string.gfind(resultText, "%d+.%d")
            resultText = s()
        end
        return resultText
    end,
    
--[[

Message from Gilbot-X:

To use this mod and use the blueprint keys, your unit's script file must declare the EnabledResourceDrains table as shown below.   

    OnStopBeingBuilt = function(self, builder, layer)
        
        self.EnabledResourceDrains = {
            Cloak  = false,
            Intel  = false,
            Jammer = false,
            Shield = false,
            Stealth= false,
            ProductionCosts = false,
            WeaponFiringCost = false,
        },
        
        BaseClass.OnStopBeingBuilt(self, builder, layer)
        
    end,
    
If you declare this table in an override of OnStopBeingBuilt, You have to declare this table before calling the base class version.  Note: You could alternatively declare the table in an override of OnCreate because OnCreate is called before OnStopBeingBuilt.
    
The keys in the table above show examples of what you can use.  Your unit mod must override the value of this in its script file from within OnCreate or OnStopBeingBuilt.  Do not just declare it outside of a function. Why not?  Because sometimes class member variables are treated as static member variables, especially likely if the value has been set outside a function, even more likely if the value is a table and the table was not set inside a member function of that class. 
  
If you override OnStopBeingBuilt you must call the base class version, otherwise the effects of this mod are lost and your ResourceDrainBreakDown blueprint keys will be ignored.

  ]]
  
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I override this to set up the maintenance consumption 
    --#*  breakdown code.  It also launches the thread that manages
    --#*  cloak effects.
    --#** 
    OnStopBeingBuilt = function(self, builder, layer)
        --# Call superclass version first
        PreviousVersion.OnStopBeingBuilt(self, builder, layer)
     
        --# Set up the maintenance consumption breakdown code.
        self:InitResourceDrainBreakDown()
        
        --# Cloak Effect management starts now
        self:ForkThread(self.CloakEffectControlThread)
    end,
    
    
    
    --# This will remain false for units that 
    --# were not designed for use with this mod. 
    --# They will function as normal
    --# without any changes to their BP files.
    ResourceDrainBreakDown = false,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  I added this to set up the maintenance consumption 
    --#*  breakdown code.  It is called from OnStopBeingBuilt.
    --#** 
    InitResourceDrainBreakDown = function(self)
        --# get energy consumption values from blueprint
        local tempTable = self:GetBlueprint().Economy.ResourceDrainBreakDown or false
        --# do this because ResourceDrainBreakDown was synching with BP!
        if tempTable then 
            self.ResourceDrainBreakDown = table.deepcopy(tempTable)
        end
        
        --# If the fetch and table copy succeeded then...
        if self.ResourceDrainBreakDown then
            --# This next block is just a safeguard so we don't have to keep checking
            --# if the keys were set correctly in the blueprint file...
            for k,v in self.EnabledResourceDrains do 
                if not self.ResourceDrainBreakDown[k] then 
                    self.ResourceDrainBreakDown[k] = { Energy = 0, Mass = 0 }
                    v = false
                else
                    if not self.ResourceDrainBreakDown[k].Energy then
                        self.ResourceDrainBreakDown[k].Energy = 0
                    end
                    if not self.ResourceDrainBreakDown[k].Mass then
                        self.ResourceDrainBreakDown[k].Mass = 0
                    end
                end
            end
  
            --# We have a valid ResourceDrainBreakDown table,
            --# so let's use it as the unit is ready to start.
            self:UpdateConsumptionValues()
        end
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  I added this as the missing complement to 
    --#*  SetEnergyMaintenanceConsumptionOverride which
    --#*  is defined in the origional Unit.lua file.
    --#**
    SetMassMaintenanceConsumptionOverride = function(self, override)
        self.MassMaintenanceConsumptionOverride = override or 0
    end,
    
    
--[[

Message from Gilbot-X:
 
WARNING!!!  This is a destructive hook of the UpdateConsumptionValues function - I do not call original code, I override it!  For that reason it may need to be updated if any patch released by GPG changes the code used in this member function.
  
]]
    --# GPG: work out a total for the energy rate and override the single value in the blueprint
    --# Here we calculate it dynamically for what the toggle/enabled functions should consume
    --# Then call the superclass verision (defined in class Unit)
    UpdateConsumptionValues = function(self)
    
        --###########################
        --# START OF GILBOT-X CODE
        --# We have to treat GPG blueprints as
        --# they were treated in the original unit.lua
        --# so skip all my code if the blueprint file 
        --# isn't set up to use my extra features
        if self.ResourceDrainBreakDown then

            --# Calculate rates
            local energy_rate = 0
            local mass_rate = 0
        
            for k,v in self.ResourceDrainBreakDown do
                if self.EnabledResourceDrains[k] then
                    energy_rate = energy_rate + self.ResourceDrainBreakDown[k].Energy 
                    mass_rate =     mass_rate + self.ResourceDrainBreakDown[k].Mass 
                end
            end
            
            --# set amount of energy drain
            self.EnergyMaintenanceConsumptionOverride = energy_rate
            self.MassMaintenanceConsumptionOverride = mass_rate
            
            --# call code from GPG version that works out bonus etc
            self.MaintenanceConsumption = not (energy_rate == 0 and mass_rate == 0)
        end
        --#  END OF GILBOT-X CODE
        --###########################
        
        --################################################
        --# START OF GPG CODE
        --# The rest is GPG code from unit.lua
        --# but I change a couple lines so that 
        --# MassMaintenanceConsumptionOverride works.
        local myBlueprint = self:GetBlueprint()
        local energy_rate = 0
        local mass_rate = 0
 
        if self.ActiveConsumption then
            local focus = self:GetFocusUnit()
            local time = 1
            local mass = 0
            local energy = 0
            if self.WorkItem then
                time, energy, mass = Game.GetConstructEconomyModel(self, self.WorkItem)
            elseif focus and focus:IsUnitState('SiloBuildingAmmo') then
                --# If building silo ammo; create the energy and mass 
                --# costs based on build rate of the silo
                --# against the build rate of the assisting unit
                time, energy, mass = focus:GetBuildCosts(focus.SiloProjectile)
                local siloBuildRate = focus:GetBuildRate() or 1
                energy = (energy / siloBuildRate) * (self:GetBuildRate() or 1)
                mass = (mass / siloBuildRate) * (self:GetBuildRate() or 1)
            elseif focus then
                --# bonuses are already factored in by GetBuildCosts
                time, energy, mass = self:GetBuildCosts(focus:GetBlueprint())
                --# Gilbot-X:  I added this next block:
                --# It only really applies to upgrading T1/T2 Mexes
                --# see timebasedoutputunit.lua for linking code
                if self:IsUnitState('Upgrading') and self.IsTimeBasedOutputUnit then 
                    energy = energy * self.UpgradeDiscountFactor
                    mass = mass * self.UpgradeDiscountFactor
                end
            end
            energy = energy * (self.EnergyBuildAdjMod or 1)
            if energy < 1 then
                energy = 1
            end
            mass = mass * (self.MassBuildAdjMod or 1)
            if mass < 1 then
                mass = 1
            end
   
            energy_rate = energy / time
            mass_rate = mass / time
        end

        
        if self.MaintenanceConsumption then
            local mai_energy = (self.EnergyMaintenanceConsumptionOverride or
                      myBlueprint.Economy.MaintenanceConsumptionPerSecondEnergy)  or 0
            --###########################
            --#  START OF GILBOT-X MOD
            local mai_mass = (self.MassMaintenanceConsumptionOverride or              
                      myBlueprint.Economy.MaintenanceConsumptionPerSecondMass) or 0
            --#  END OF GILBOT-X MOD
            --###########################
            
            --# apply bonuses
            mai_energy = mai_energy * (100 + self.EnergyModifier) * (self.EnergyMaintAdjMod or 1) * 0.01
            mai_mass = mai_mass * (100 + self.MassModifier) * (self.MassMaintAdjMod or 1) * 0.01

            energy_rate = energy_rate + mai_energy
            mass_rate = mass_rate + mai_mass
        end

        --# apply minimum rates
        energy_rate = math.max(energy_rate, myBlueprint.Economy.MinConsumptionPerSecondEnergy or 0)
        mass_rate = math.max(mass_rate, myBlueprint.Economy.MinConsumptionPerSecondMass or 0)

        self:SetConsumptionPerSecondEnergy(energy_rate)
        self:SetConsumptionPerSecondMass(mass_rate)

        if (energy_rate > 0) or (mass_rate > 0) then
            self:SetConsumptionActive(true)
        else
            self:SetConsumptionActive(false)
        end
        
        --#
        --# END OF GPG CODE
        --################################################
    end,
    
    
    
    
   

    
    
    --##########################################################################################
    --## TOGGLES
    --##########################################################################################
    
--[[

Message from Gilbot-X:

WARNING!!!  These are a destructive hook of the OnScriptBitSet and OnScriptBitClear functions - I do not call original code, I override it!  For that reason it may need to be updated if any patch released by GPG changes the code used in this member function.  Seems to work fine up to patch 3260 though...
  
In these overrides, we first check if the unit whose toggles have been pressed is actually designed to use the extr afeatures of this Maintenance Consumption Breakdown mod.  If it is, instead of calling SetMaintenanceConsumptionActive and SetMaintenanceConsumptionInActive we set the value of keys inside EnabledResourceDrains and make calls to UpdateConsumptionValues.  This is safe for all units, if they were designed for this mod or not.

For units that use the mod, we assume they might have more than one kind of intel and we call OnIntelEnbaled and OnIntelDisabled explicitly when toggle bits 2,3,5,8 are used (these are all classed as intel by GPG).  This was needed to make TransparentCloaks mod compatible with this mod for example.

Now changed so that cloak fields work.  If you enable the Cloak Intel, the CloakField intel won't work.  So now we don't enable cloak if the unit has CloakField define in the blueprint.  If the unit gains CloakField as an enhancement, they can set 'self.HasCloakField = true' or they will have to override this as CloakField won't be defined in the blueprint.

 ]]
    --#*
    --#*  This function does what is needed when 
    --#*  the player toggles shields on or anything else off.  
    --#**     
    OnScriptBitSet = function(self, bit)
    
        if bit == 0 then --# shield toggle
            if self.IsSensitiveShieldUser then
                self.MyShieldToggledOn = true
                --# Try to enable our personal shield if allowed
                if self.ShieldUsePermitted and self.IsPacked then 
                    --# This calls the overrided version
                    --# in the SensitiveShield class
                    self:EnableShield() 
                end
            --# This unit has a normal shield
            else
                --# So we should just turn it on.
                self:PlayUnitAmbientSound( 'ActiveLoop' )
                self:EnableShield()
                if self.ResourceDrainBreakDown then 
                    self.EnabledResourceDrains.Shield = true
                    self:UpdateConsumptionValues()
                end
            end
            
        elseif bit == 1 then --# weapon toggle
            --# Perform this function if it is defined.
            if self.OnScriptBit1Set then 
                self:OnScriptBit1Set() 
            end
            
        elseif bit == 2 then --# jamming toggle
            --LOG("Maintenance Mod: Turning off jammer with bit 2")
            self:StopUnitAmbientSound( 'ActiveLoop' )
            self:DisableUnitIntel('Jammer')
            if self.ResourceDrainBreakDown then 
                --LOG("Maintenance Mod: Removing jammer cost only")
                self.EnabledResourceDrains.Jammer = false
                self:UpdateConsumptionValues()
                --# OnIntelDisabled often doesn't get called
                --# if we have multiple intel types
                --# so we do it explicitly for units that use this mod
                self:OnIntelDisabled()
            else
                self:SetMaintenanceConsumptionInactive()
                --LOG("Maintenance Mod: This unit doesn't use the mod: " .. 
                --    "Setting consumption off from jammer toggle")
            end
            
        elseif bit == 3 then --# intel toggle
            --LOG("Maintenance Mod: Turning off intel with bit 3")
            self:StopUnitAmbientSound( 'ActiveLoop' )
            self:DisableUnitIntel('Sonar')
            self:DisableUnitIntel('Omni')
            self:DisableUnitIntel('Spoof')
            self:DisableUnitIntel('Radar')
            if self.ResourceDrainBreakDown then 
                --LOG("Maintenance Mod: Removing intel cost only")
                self.EnabledResourceDrains.Intel = false
                self:UpdateConsumptionValues()
                --# OnIntelDisabled often doesn't get called
                --# if we have multiple intel types
                --# so we do it explicitly for units that use this mod
                self:OnIntelDisabled()
            else
                --# GPG units that weren't designed for this mod
                --# are used to having these enabled when bit 3 is used! 
                self:DisableUnitIntel('Jammer')
                self:DisableUnitIntel('RadarStealth')
                self:DisableUnitIntel('RadarStealthField')
                self:DisableUnitIntel('SonarStealth')
                self:DisableUnitIntel('SonarStealthField')
                self:DisableUnitIntel('Cloak')
                self:DisableUnitIntel('CloakField')
                self:SetMaintenanceConsumptionInactive()
                --LOG("Maintenance Mod: This unit doesn't use the mod: " .. 
                --    "Setting consumption off from intel toggle")
            end
            
        elseif bit == 4 then --# production toggle
            --# see below for my override of this
            self:OnProductionPaused()
            
        elseif bit == 5 then --# stealth toggle
            --LOG("Maintenance Mod: Turning off stealth with bit 5")
            self:StopUnitAmbientSound( 'ActiveLoop' )
            self:DisableUnitIntel('RadarStealth')
            self:DisableUnitIntel('RadarStealthField')
            self:DisableUnitIntel('SonarStealth')
            self:DisableUnitIntel('SonarStealthField')
            if self.ResourceDrainBreakDown then 
                --LOG("Maintenance Mod: Removing stealth cost only")
                self.EnabledResourceDrains.Stealth = false 
                self:UpdateConsumptionValues() 
                --# OnIntelDisabled often doesn't get called
                --# if we have multiple intel types
                --# so we do it explicitly for units that use this mod
                self:OnIntelDisabled()
            else
                self:SetMaintenanceConsumptionInactive()
                --LOG("Maintenance Mod: This unit doesn't use the mod: " .. 
                --    "Setting consumption off from stealth toggle")
            end
            
        elseif bit == 6 then --# generic pause toggle
            self:SetPaused(true)
            
        elseif bit == 7 then --# special toggle
            self:EnableSpecialToggle()
            
        elseif bit == 8 then --# cloak toggle
            --LOG("Maintenance Mod: Turning off cloak with bit 8")
            self:StopUnitAmbientSound( 'ActiveLoop' )
            if self:GetBlueprint().Intel.CloakField or 
                self.HasCloakField then
                --LOG("Maintenance Mod: Disabling cloak field.")
                self:DisableUnitIntel('CloakField')
            else
                --LOG("Maintenance Mod: Disabling personal cloak.")
                self:DisableUnitIntel('Cloak')
            end
            if self.ResourceDrainBreakDown then
                --LOG("Maintenance Mod: Removing cloak cost only")
                self.EnabledResourceDrains.Cloak = false
                self:UpdateConsumptionValues()
                --# OnIntelDisabled often doesn't get called
                --# if we have multiple intel types
                --# so we do it explicitly for units that use this mod
                self:OnIntelDisabled()
            else
                self:SetMaintenanceConsumptionInactive()
                --LOG("Maintenance Mod: This unit doesn't use the mod: " .. 
                --    "Setting consumption off from cloak toggle")
            end
        end
    end,
    
    --#*
    --#*  This function does what is needed when 
    --#*  the player toggles shields off or anything else on.  
    --#**   
    OnScriptBitClear = function(self, bit)
    
        if bit == 0 then --# shield toggle off
            if self.IsSensitiveShieldUser then
                self.MyShieldToggledOn = false
                --# Disable our personal shield if it is enabled
                if self.MyShieldIsEnabled then 
                    --# This calls the overrided version
                    --# in the SensitiveShield class
                    self:DisableShield() 
                end
            --# This unit has a normal shield
            else
                --# so turn it off
                self:StopUnitAmbientSound( 'ActiveLoop' )
                self:DisableShield()
                if self.ResourceDrainBreakDown then 
                    self.EnabledResourceDrains.Shield = false 
                    self:UpdateConsumptionValues() 
                end
            end
            
        elseif bit == 1 then --# weapon toggle
            --# Perform this function if it is defined.
            if self.OnScriptBit1Clear then 
                self:OnScriptBit1Clear() 
            end
            
        elseif bit == 2 then --# jamming toggle on
            --LOG("Maintenance Mod: Turning on jammer with bit 2")
            self:PlayUnitAmbientSound( 'ActiveLoop' )
            self:EnableUnitIntel('Jammer')
            if self.ResourceDrainBreakDown then 
                --LOG("Maintenance Mod: Adding jammer cost only")
                self.EnabledResourceDrains.Jammer = true
                self:UpdateConsumptionValues()
                --# OnIntelEnabled often doesn't get called
                --# if we have multiple intel types
                --# so we do it explicitly for units that use this mod
                self:OnIntelEnabled()
            else
                self:SetMaintenanceConsumptionActive()
            end
            
        elseif bit == 3 then --# intel toggle on
            --LOG("Maintenance Mod: Turning on intel with bit 3")
            self:PlayUnitAmbientSound( 'ActiveLoop' )
            self:EnableUnitIntel('Radar')
            self:EnableUnitIntel('Sonar')
            self:EnableUnitIntel('Omni')
            self:EnableUnitIntel('Spoof')
            if self.ResourceDrainBreakDown then 
                --LOG("Maintenance Mod: Adding intel cost only")
                self.EnabledResourceDrains.Intel = true
                self:UpdateConsumptionValues()
                --# OnIntelEnabled often doesn't get called
                --# if we have multiple intel types
                --# so we do it explicitly for units that use this mod
                self:OnIntelEnabled()
            else
                --# GPG units that weren't designed for this mod
                --# are used to having these enabled when bit 3 is used! 
                self:EnableUnitIntel('Jammer')
                self:EnableUnitIntel('RadarStealth')
                self:EnableUnitIntel('RadarStealthField')
                self:EnableUnitIntel('SonarStealth')
                self:EnableUnitIntel('SonarStealthField')
                self:EnableUnitIntel('Cloak')
                self:EnableUnitIntel('CloakField')
                self:SetMaintenanceConsumptionActive()
            end
            
        elseif bit == 4 then --# production toggle on
            --# See below for my override of this
            self:OnProductionUnpaused()
            
        elseif bit == 5 then --# stealth toggle on
            --LOG("Maintenance Mod: Turning on stealth with bit 5")
            self:PlayUnitAmbientSound( 'ActiveLoop' )
            self:EnableUnitIntel('RadarStealth')
            self:EnableUnitIntel('RadarStealthField')
            self:EnableUnitIntel('SonarStealth')
            self:EnableUnitIntel('SonarStealthField')
            if self.ResourceDrainBreakDown then 
                --LOG("Maintenance Mod: Adding stealth cost only")
                self.EnabledResourceDrains.Stealth = true 
                self:UpdateConsumptionValues()
                --# OnIntelEnabled often doesn't get called
                --# if we have multiple intel types
                --# so we do it explicitly for units that use this mod
                self:OnIntelEnabled()
            else
                self:SetMaintenanceConsumptionActive()
            end
            
        elseif bit == 6 then --# generic pause toggle on
            self:SetPaused(false)
            
        elseif bit == 7 then --# special toggle off
            self:DisableSpecialToggle()
            
        elseif bit == 8 then --# cloak toggle on
            --# If the unit has this sound then play it
            self:PlayUnitAmbientSound('ActiveLoop')
            
            --# Enable Cloak or Cloakfield
            if self:GetBlueprint().Intel.CloakField or 
               self.HasCloakField then
                self:EnableUnitIntel('CloakField')
            else
                self:EnableUnitIntel('Cloak')
            end
            
            --# take care of power usage
            if self.ResourceDrainBreakDown then 
                --LOG("Maintenance Mod: Adding cloak cost only")
                self.EnabledResourceDrains.Cloak = true
                self:UpdateConsumptionValues()
                --# OnIntelEnabled often doesn't get called
                --# if we have multiple intel types
                --# so we do it explicitly for units that use this mod
                self:OnIntelEnabled()
            else 
                self:SetMaintenanceConsumptionActive()
            end
        end
    end,
    
    --#*
    --#*  I overrided this to use my ResourceDrainBreakDown key.
    --#**   
    OnProductionPaused = function(self)
        --# Change this to use ResourceDrainBreakDown
        if self.ResourceDrainBreakDown then 
            --# Switch off costs because production is paused
            self.EnabledResourceDrains.ProductionCosts = false
            self:UpdateConsumptionValues()
        else
            self:SetMaintenanceConsumptionInactive()
        end
        --# GPG version does this
        self:SetProductionActive(false)
    end,

    --#*
    --#*  I overrided this to use my ResourceDrainBreakDown key.
    --#**   
    OnProductionUnpaused = function(self)
        --# Change this to use ResourceDrainBreakDown
        if self.ResourceDrainBreakDown then 
            --# Switch costs on because production is active
            self.EnabledResourceDrains.ProductionCosts = true
            self:UpdateConsumptionValues()
        else
            self:SetMaintenanceConsumptionActive()
        end
        --# GPG version does this
        self:SetProductionActive(true)
    end,
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  I moved this code here as it should not be elsewhere.
    --#*  This happens when unit's production is unpaused
    --#*  and it has whatever other conditions it needs 
    --#*  to execute production.
    --#*  The consequence of moving it here is that you may
    --#*  sometimes hear the sound twice.
    --#**
    OnActive = function(self)
        --# Play activate sound if there is one
        local myBlueprint = self:GetBlueprint()
        if myBlueprint.Audio and myBlueprint.Audio.Activate then
            self:PlaySound(myBlueprint.Audio.Activate)
        end
    end,
    
    
    
    
--[[

Message from Gilbot-X:

This function is new, I added it.  Units with enhancements that use this mod should call this function at the end of their override of CreateEnhancement if any of their enhancements consume power!!

]]
    FinishCreatingEnhancement = function(self, enh)
        if not enh then 
            WARN("MaintenanceConsumption Mod: FinishCreatingEnhancement: " .. 
                 "Expected 2 args but got 1!! for unit " 
                  .. self:GetUnitId() .. " and entity " .. self:GetEntityId())
            return
        end
    
        --LOG("MaintenanceConsumption Mod: FinishCreatingEnhancement: " .. 
        --        "Applying ResourceDrainChanges for enh=" .. enh ..
        --        "for unit " .. self:GetUnitId() .. " and entity " .. self:GetEntityId())
        
        --# Override existing energy consumption values 
        --# with any new values provided in the enhancement blueprint
        local enhancementBP = self:GetBlueprint().Enhancements[enh] 
        if not enhancementBP then 
            WARN("MaintenanceConsumption Mod: FinishCreatingEnhancement: " .. 
                 "BP not found for " .. enh .. " for unit " 
                  .. self:GetUnitId() .. " and entity " .. self:GetEntityId())
            return
        end
        
        
        if enhancementBP.ResourceDrainChanges then
            for drainsourcekey, drainsourcetable in enhancementBP.ResourceDrainChanges do
                if self.ResourceDrainBreakDown[drainsourcekey] 
                and drainsourcetable then 
                    --# Note, errors in the BP file, 
                    --# no extra keys will be inserted
                    for draintype, drainamount in drainsourcetable do 
                        if draintype == 'Energy' then 
                            self.ResourceDrainBreakDown[drainsourcekey].Energy = drainamount
                        elseif draintype == 'Mass' then 
                            self.ResourceDrainBreakDown[drainsourcekey].Mass = drainamount
                        else
                            WARN("MaintenanceConsumption Mod: FinishCreatingEnhancement: " .. 
                                "Found bad keys in ResourceDrainChanges in enhancement BP: " .. 
                                drainsourcekey .. "." .. draintype .. " for unit " 
                                .. self:GetUnitId() .. " and entity " .. self:GetEntityId())
                        end
                    end 
                else
                    WARN("MaintenanceConsumption Mod: FinishCreatingEnhancement: " .. 
                          "Found bad keys in ResourceDrainChanges in enhancement BP: " .. 
                          drainsourcekey .. " for unit " 
                          .. self:GetUnitId() .. " and entity " .. self:GetEntityId())
                end                
            end
        else
            --LOG("MaintenanceConsumption Mod: FinishCreatingEnhancement: " .. 
            --    "No ResourceDrainChanges found.")
        end
        
        --# Call this so unit will start to consume 
        --# energy according to newly loaded values 
        self:UpdateConsumptionValues()
    end,
    
  
  
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  Called only by UpdateConsumptionWhenAbilityChanges.
    --#*  Also called by slider control callbacks. 
    --#*  Get reference values for consumption from unit blueprints.
    --#*  
    --#*  Note:  No point looking at enhancement blueprints.
    --#*  We would need to sync to find out which enhancement we had, 
    --#*  so we would have got data from sync table anyway.
    --#**  
    GetReferenceConsumptionFromBlueprint = function(self, resourceDrainId) 
        local econbp = self:GetBlueprint().Economy
        local referenceConsumption = {Energy=0, Mass=0}
        
        --# Does this unit use the ResourceDrainBreakDown key 
        --# to manage multiple resource drains?
        if self.ResourceDrainBreakDown then
            --# Perform safety check
            if econbp.ResourceDrainBreakDown[resourceDrainId] then
                --# Get values from BP
                referenceConsumption.Energy =            
                    econbp.ResourceDrainBreakDown[resourceDrainId].Energy or 0
                referenceConsumption.Mass =            
                    econbp.ResourceDrainBreakDown[resourceDrainId].Mass or 0
            else
                --# Warn unit programmer they have made a mistake
                --# when trying to use this mod.
                WARN('GetReferenceConsumptionFromBlueprint: Bad resource drain ID ' 
                  .. repr(resourceDrainId) .. ' supplied.'
                )
            end
        else
            --# This unit was not designed for use with 
            --# my Maintenance Consumption Breakdown code.
            referenceConsumption.Energy =  econbp.MaintenanceConsumptionPerSecondEnergy or 0 
            referenceConsumption.Mass = econbp.MaintenanceConsumptionPerSecondMass or 0
        end
        
        --# return result
        return referenceConsumption
    end,
        
        
        
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  Called by slider control callbacks and also in unit scripts by other 
    --#*  units who change how much something drains depending on its situation.
    --#**          
    UpdateConsumptionWhenAbilityChanges = function(self, newAbilityValue, 
              referenceAbilityValue, resourceDrainId, referenceConsumption)
          
        local changeFactor = 0
        local newConsumption = {Energy=0, Mass=0}
        
        --# If we didn't get it as an argument...
        if not referenceConsumption then
            --# then get it from the blueprint.
            referenceConsumption = self:GetReferenceConsumptionFromBlueprint(resourceDrainId)
        end
               
        --# Without a reference value we can't update the consumption to anything meaningful
        if referenceAbilityValue > 0 then 
            --# Warning, this change factor can be zero if the newAbilityValue is zero
            changeFactor = newAbilityValue / referenceAbilityValue    
        
             --# Calculate new resource consumption based on linear scaling
            newConsumption.Energy = referenceConsumption.Energy * changeFactor
            newConsumption.Mass   = referenceConsumption.Mass   * changeFactor
      
            --# Alter amount of each resource consumed now that 
            --# abilty value is at this setting.
            if self.ResourceDrainBreakDown then
                --# If resourceDrainId was not defined it means
                --# there should be no resource drain for this ability.
                --# Should this have been called in that case?
                if resourceDrainId then
                    --# Perform safety check to see unit has values
                    if self.ResourceDrainBreakDown[resourceDrainId] then
                        --# Set values in self
                        self.ResourceDrainBreakDown[resourceDrainId].Energy = newConsumption.Energy
                        self.ResourceDrainBreakDown[resourceDrainId].Mass = newConsumption.Mass
                        self:UpdateConsumptionValues()
                    else
                        --# Warn self programmer they have made a mistake
                        --# when trying to use this mod.
                        WARN('UpdateConsumptionWhenAbilityChanges: Bad resource drain ID ' 
                          .. repr(resourceDrainId) .. ' supplied.'
                        )
                    end
                else
                    --# Warn self programmer they have made a mistake
                    --# when trying to use this mod.
                    WARN('UpdateConsumptionWhenAbilityChanges: Called on' 
                     .. ' ResourceDrainBreakDown unit with no resourceDrainId supplied.'
                    )
                end
            else
                --# This self was probably not designed for use with 
                --# my Maintenance Consumption Breakdown code, or 
                --# it only has one resource consuming ability.
                if newConsumption.Energy then
                    self:SetEnergyMaintenanceConsumptionOverride(newConsumption.Energy)
                    self:SetConsumptionPerSecondEnergy(newConsumption.Energy)
                end
                if newConsumption.Mass then
                    self:SetMassMaintenanceConsumptionOverride(newConsumption.Mass)
                    self:SetConsumptionPerSecondMass(newConsumption.Mass)
                end
            end
            
            --# This next block is for debugging only.
            --# This can be deleted when debugging is finished.
            if debugAllUnits then
              LOG('UpdateConsumptionWhenAbilityChanges:' ..
                    '  resourceDrainId='        .. repr(resourceDrainId) ..
                    '  newConsumption.Energy='  .. repr(newConsumption.Energy) ..
                    '  newConsumption.Mass='    .. repr(newConsumption.Mass) ..
                    '  referenceAbilityValue='  .. repr(referenceAbilityValue) ..
                    '  newValue='               .. repr(newAbilityValue)
              )
            end
        end
    end,

--[[

Message from Gilbot-X:

This function is new, I added it when making Adjacency v2 to support ACUs carrying information about the army.

]]  
     
     
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  Commanders can store information for the army
    --#*  and be referenced throughout their lifetime.
    --#*  If you are playing non-assassin mode and you 
    --#*  lose your commander, then tough, you lose your 
    --#*  army's data!
    --#*
    --#*  If you are playing with Victory conditions set to 
    --#*  Supremacy, Annihilation or Sandbox, then whatever is
    --#*  built first after the ACU does will get its entity id.
    --#*  This function returns nil if that has happened.
    --#** 
    GetMyCommander= function(self)
        
        --# Safety check to prevent script errors.
        if self:BeenDestroyed() then return nil end
        
        --# This would cause a script error on a destroyed unit
        local myArmyId = self:GetArmy()
        
        --# In patch 3260, 1048576 was the multiple for each army.
        --# these numbers get reused, but once you lose the commander 
        --# you can't build another one, so does its entity ID get used
        --# by another unit when you are not playing Assassin mode?
        local myCommanderEntityId = (myArmyId -1) * 1048576
        
        --# This is for debugging only
        if debugAllUnits then  
          LOG('Adjacency mod: GetMyCommander: Commander entityID for ' 
          .. ' Army= ' .. repr(myArmyId)
          .. ' is ' .. repr(myCommanderEntityId)
          )
        end
        
        --# Now try to get the unit itself.        
        local myCommander = GetUnitById(myCommanderEntityId)
        
        --# If no unit came back at all...
        --# (Note that this is unlikely to happen often, 
        --# as entity IDs get reused when a unit dies.)
        if not myCommander then 
            --# This is for debugging only
            if debugAllUnits then
              LOG('Adjacency mod: GetMyCommander: Tried to get Commander at unit ' 
              .. ' entity Id=' .. repr(myCommanderEntityId) 
              .. ' but nothing was returned.  Maybe this Commander was killed?'
              )
            end
            --# Calling code must check for nil
            --# in case commander can't be found
            return nil
            
            
        --# This is more likely to happenm if your ACU has died...
        elseif not (EntityCategoryContains(categories.COMMAND, myCommander)) then
            --# This is for debugging only
            if debugAllUnits then
              LOG('Adjacency mod: GetMyCommander: Tried to get Commander at unit'
              .. ' entity Id=' .. repr(myCommanderEntityId) 
              .. ' but something else was returned: '
              .. ' UnitId=' .. repr(myCommander:GetUnitId())
              )
            end
            --# Calling code must check for nil
            --# in case commander can't be found
            return nil
        
        --# We found the ACU for this unit's army,        
        else
            --# Return a unit object reference to it
            --# that allows calling code to call special
            --# functions defined in the ACU's unit script.
            return myCommander
        end
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*      
    --#*  This is used by the ACUKnowledge class for auto power-down.
    --#** 
    GetMyTechLevel = function(self)
        if EntityCategoryContains(categories.TECH1, self) then return 1 end
        if EntityCategoryContains(categories.TECH2, self) then return 2 end
        if EntityCategoryContains(categories.TECH3, self) then return 3 end
        if EntityCategoryContains(categories.EXPERIMENTAL, self) then return 4 end
    end,
    
    
--------------------------------------------------------------------------------------------------
--  New Functions For Cloak Effect 
--  (Based on code by CovertJaguar)                                                  	--------------------------------------------------------------------------------------------------

    --# Here you set the time in seconds you want between 
    --# each time a cloakfield unit checks its surroundings 
    --# and marks units as being in  a cloak field.
    --# All units will unmark themselves at half the frequency that units are marked.   
    CloakUpdatePeriod = 0.25,
    CloakEffectEnabled = 0,

    --# This thread runs constantly in the background for all units. 
    --# It ensures that the cloak effect and cloak field are always in the correct state
    CloakEffectControlThread = function(self)
        while not (self == nil or self:GetHealth() <= 0 or self:IsDead()) do
            self.InCloakField = false
            --# In the elapsed time we might have been marked 
            --# as being in a cloak feild
            WaitSeconds(self.CloakUpdatePeriod)
            
            --# Units with an active CloakField mark other units that they cloak
            if self:IsIntelEnabled('CloakField') then self:MarkUnitsInRange() end
            
            --# In the elapsed time we might have been marked 
            --# as being in a cloak feild
            WaitSeconds(self.CloakUpdatePeriod)
            
            --# Units with an active CloakField mark other units that they cloak
            if self:IsIntelEnabled('CloakField') then self:MarkUnitsInRange() end
            
            --# Every unit a) unmarks itself, and 
            --# b) updates its own cloak effect, 
            --# at half the frequency that units with cloak fields 
            --# mark other units as being cloaked.
            self:UpdateCloakEffect()
        end
    end,

    MarkUnitsInRange = function(self)
        local position = self:GetPosition(0)
        local range = self:GetIntelRadius('CloakField') 
            or self:GetBlueprint().Intel.CloakFieldRadius
        --# Range must be (radius - 2) because 
        --# it seems GPG did that for the actual field for some reason.
        --# Anything beyond (radius - 2) is not cloaked by the cloak field
        range = range - 2
        --LOG('Transparent Cloak Effect Mod: MarkUnitsInRange: ' ..
        --    ' for ' .. self:GetUnitId() .. ' entity ' .. self:GetEntityId() ..
        --    ' cloak field range is ' .. repr(range))
            
        local brain = self:GetAIBrain()
        local UnitsInRange = 
            brain:GetUnitsAroundPoint( categories.ALLUNITS, position, range, 'Ally' )
        for num, unit in UnitsInRange do
            if unit ~= self then
                --LOG('Transparent Cloak Effect Mod: MarkUnitsInRange: ' ..
                --    'marking unit ' .. unit:GetUnitId() .. ' and entity ' .. unit:GetEntityId())
                unit.InCloakField = true
            end
        end
    end,
    
    -- This is the core of the entire mod. The effect is actually applied here.
    UpdateCloakEffect = function(self)
        --# Next block is for debug only
        if self.InCloakField then 
            --LOG('Transparent Cloak Effect Mod: UpdateCloakEffect: ' ..
            --        ' unit ' .. self:GetUnitId() .. ' and entity ' .. self:GetEntityId() ..
            --        ' was in a cloak field.')
        end
        
        local cloaked = self:IsIntelEnabled('Cloak') or 
                        self:IsIntelEnabled('CloakField') or 
                        self.InCloakField
        if (not cloaked) and self.CloakEffectEnabled > 0 then
            self:DeactivateTransparentCloakEffect()
        elseif cloaked and (self.CloakEffectEnabled < 3) and self:GetHealth() > 0 then
            self:ActivateTransparentCloakEffect()
        end
    end,

    ActivateTransparentCloakEffect = function(self)
        --LOG('*DEBUG: Cloak Effect Enabled')
        self.CloakEffectEnabled = self.CloakEffectEnabled + 1
        self:DestroyIdleEffects()
        self:DestroyMovementEffects()
        self:DestroyBeamExhaust()
        
        --# Ask unit to get rid of any conflicting effects
        --# that shouldn't be on when cloak effect is on
        if self.OnCloakEffectEnabled then 
            self:OnCloakEffectEnabled(true)
        end
        
        self:SetMesh(self:GetBlueprint().Display.CloakMeshBlueprint, true)
    end,
    
    DeactivateTransparentCloakEffect = function(self)
        --LOG('*DEBUG: Cloak Effect Disabled')
        self.CloakEffectEnabled = 0

        self:SetMesh(self:GetBlueprint().Display.MeshBlueprint, true)
        
        --# Ask unit to renable any conflicting effects
        --# that should come back on when cloak effect is switched off
        if self.OnCloakEffectEnabled then 
            self:OnCloakEffectEnabled(false)
            --# i.e. make calls to these:
            --self:CreateIdleEffects()
            --self:CreateMovementEffects()
            --self:UpdateBeamExhaust('Idle')
        end
    end,

    
    
    --#*
    --#*  Gilbot-X says:
    --#* 
    --#*  I added this function to support use of cloak effect.
    --#**
    IsThisUnitCloaked = function(self)
        
        --# First check if we are in someone else's claok field.
        if self.InCloakField then return true end
        
        --# If not, then next check if the unit itself has cloak powers.
        --# If a unit is using my Maintenance Consumption Breakdown code,
        --# then there is an alternative way to test that does not rely on 
        --# the GPG function Unit.IsIntelEnabled.  There were issues about 
        --# the accuracy of that function, for some of my modded units
        --# that have more than one type of intel. So if it isn't trusted, we
        --# can opt (before hook time) to use the alternative method for those
        --# units.        
        local dontTrustIsIntelEnabled = false
        if dontTrustIsIntelEnabled and self.ResourceDrainBreakDown then
            
            --# Yes, Maintenance Consumption Breakdown mod IS being used by this unit
            local cloaked = false
            --# If cloak is switched on
            if self.EnabledResourceDrains.Cloak then
                --# Look at the economy and our unit's energy income
                local aiBrain = self:GetAIBrain()
                local fraction = self:GetResourceConsumed()
                --#  If we are not getting full energy we need to cloak 
                --#  or our economy is at zero energy then we can't be cloaked
                if fraction ~= 1 and aiBrain:GetEconomyStored('ENERGY') <= 0 then
                    cloaked = false
                else
                --# otherwise assume we are cloaked!
                    cloaked = true
                end
            end
          
            return cloaked
        else
            --# No, Maintenance Consumption Breakdown mod 
            --# is NOT being used by this unit
            --# so pray that this GPG code actually works!!
            return self:IsIntelEnabled('Cloak') or self:IsIntelEnabled('CloakField')
        end
    end,
    
    
--------------------------------------------------------------------------------------------------
--  Overridden Functions For Cloak Effect
--   (Based on code by CovertJaguar)    	--------------------------------------------------------------------------------------------------
	

    --# Overridden Function:
    --# Just an attempt to ensure that a dying unit decloaks, may not be 100% effective
    OnKilled = function(self, instigator, type, overkillRatio)
        if not self:BeenDestroyed() then 
            self:DeactivateTransparentCloakEffect() 
        end
        PreviousVersion.OnKilled(self, instigator, type, overkillRatio)
    end,

    --# Overridden Function:
    --# While the CloakEffectControlThread will activate the cloak effect eventually,
    --# this method tries to provide faster a response time to intel changes
    OnIntelEnabled = function(self)
        PreviousVersion.OnIntelEnabled(self)
        self:UpdateCloakEffect()
    end,

    --# Overridden Function:
    --# While the CloakEffectControlThread will deactivate the cloak effect eventually,
    --# this method tries to provide faster a response time to intel changes
    OnIntelDisabled = function(self)
        PreviousVersion.OnIntelDisabled(self)
        self:UpdateCloakEffect()
    end,

    --# Overridden Function:
    --# Fixes a problem with the default code and multiple types of intel on a unit.
    IntelWatchThread = function(self)
        WaitSeconds(0.5)
        local energy = self:GetConsumptionPerSecondEnergy()
        local mass = self:GetConsumptionPerSecondMass()
        local fraction = self:GetResourceConsumed()
        if energy == 0 and mass == 0 then
            fraction = 1
        end
        while fraction == 1 do
            WaitSeconds(0.5)
            energy = self:GetConsumptionPerSecondEnergy()
            mass = self:GetConsumptionPerSecondMass()
            if energy == 0 and mass == 0 then
                fraction = 1
            else
                fraction = self:GetResourceConsumed()
            end
        end
        self:DisableUnitIntel()
    end,
    
    
    
--------------------------------------------------------------------------------------------------
--  Overridden Functions For Pipelines
--     	--------------------------------------------------------------------------------------------------


    --#*
    --#*  Gilbot-X says:
    --#* 
    --#*  I added this function to support use of x and z offsets
    --#*  in the shield blueprints.
    --#**
    CreateShield = function(self, shieldSpec)
        local bp = self:GetBlueprint()
        local bpShield = shieldSpec
        if not shieldSpec then
            bpShield = bp.Defense.Shield
        end
        if bpShield then
            self:DestroyShield()
            self.MyShield = Shield {
                Owner = self,
                Mesh = bpShield.Mesh or '',
                MeshZ = bpShield.MeshZ or '',
                ImpactMesh = bpShield.ImpactMesh or '',
                ImpactEffects = bpShield.ImpactEffects or '',    
                Size = bpShield.ShieldSize or 10,
                ShieldMaxHealth = bpShield.ShieldMaxHealth or 250,
                ShieldRechargeTime = bpShield.ShieldRechargeTime or 10,
                ShieldEnergyDrainRechargeTime = bpShield.ShieldEnergyDrainRechargeTime or 10,
                ShieldVerticalOffset = bpShield.ShieldVerticalOffset or -1,
                --# Gilbot-X says:
                --# The next 2 lines are new.
                ShieldXOffset = bpShield.ShieldXOffset or 0,
                ShieldZOffset = bpShield.ShieldZOffset or 0,
                ShieldRegenRate = bpShield.ShieldRegenRate or 1,
                ShieldRegenStartTime = bpShield.ShieldRegenStartTime or 5,
                PassOverkillDamage = bpShield.PassOverkillDamage or false,
            }
            self:SetFocusEntity(self.MyShield)
            self:EnableShield()
            self.Trash:Add(self.MyShield)
        end
    end,
}


   



end --(end of non-destructive hook)