--#*****************************************************************************
--#* New File : /mods/gilbotsimprovements/pauseableproductionunit.lua
--#*
--#* Modded By: Gilbot-X
--#*
--#* Summary  : This file is to add an interface to particular structure units,
--#*            namely those that produces energy or mass.  It gives the unit a
--#*            set of States that helps the programmer manage their behaviour
--#*            consistently, so that less code (if any) is needed in individual
--#*            unit scripts.  This is efficient as many of these units benefit
--#*            from having similar functionality. 
--#*
--#*            Defines the following Interface:
--#*            ---------------------------------
--#*            DoBeforeAnyStateChanges
--#*            UnpausedState
--#*            PausedState
--#*            PreparingToUpgradeState
--#*            UpgradingState
--#*
--#*            This interface facilitates upgrade delaying and blocking.
--#*
--#*****************************************************************************

--# Need this so we can call base class code
--# and bypass any intervening superclass
local Unit = import('/lua/sim/unit.lua').Unit
local messageDialog=nil

--# This function is called to create the class 
--# so this class can add its 
--# code to different base classes.
function MakePauseableProductionUnit(baseClassArg)


local resultClass = Class(baseClassArg) {

    --# Quick way to check if a unit extends this class.
    IsPauseableProductionUnit = true,

    --# This function does initialisation of non-static variables.
    --# We set up effect size values and fork the update thread,
    OnStopBeingBuilt = function(self,builder,layer)
        --# Call base class code first
        baseClassArg.OnStopBeingBuilt(self,builder,layer)
    
        --# This was done in GPG code for units that
        --# had production.
        self:SetMaintenanceConsumptionActive()
    
        --# This is code that will be done after the unit has 
        --# been built but before the first state is changed
        --# in a structure unit that uses states.
        self:DoBeforeAnyStateChanges()
        
        --# The rest is the same code we call when 
        --# we are in the PausedState and changing
        --# into the initial state.
        Unit.OnProductionUnpaused(self)           
        self.IsProductionPaused = false
        ChangeState(self, self.UnpausedState)
    end,
  
 
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  This serves as an interface/template for inheriting classes to override.
    --#*  Override this with  code that will be done after the unit 
    --#*  has been built but before the first state is changed
    --#*  in a structure unit that uses states.
    --#**
    DoBeforeAnyStateChanges = function(self)
    
    end,
    
    
 
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  How to behave when production pause is toggled to 'unpause'.
    --#*  This serves as an interface/template for inheriting classes to override.
    --#*  It contains 'dormant code' just for use by certain of my inheriting classes.
    --#*  Dormant code is code that only run in inheriting classes that set certain 
    --#*  flags that indicate the code should be run.
    --#**
    UnpausedState = State {
        Main = function(self)
            --# Keep a reference to which state we are in
            self.PreviousState = self.UnpausedState
            --# Do code for sensitive shield users (Aeon)
            if self.IsSensitiveShieldUser then
                if self.MyShieldIsEnabled then self:DisableShield() end
            end
            --# Update effecst to match current state
            self:OnActive()
        end,

        OnProductionPaused = function(self)
            --# Now do base calss after we have packed up
            Unit.OnProductionPaused(self)
            --# Change state
            self.IsProductionPaused = true
            ChangeState(self, self.PausedState)
        end,
    },
    
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  How to behave when production pause is toggled to 'pause'.
    --#*  This serves as an interface/template for inheriting classes to override.
    --#*  It contains 'dormant code' just for use by certain of my inheriting classes.
    --#*  Dormant code is code that only run in inheriting classes that set certain 
    --#*  flags that indicate the code should be run.
    --#**
    PausedState = State {
        Main = function(self)
            --# Keep a reference to which state we are in
            self.PreviousState = self.PausedState
            --# Update effecst to match current state
            self:OnInactive()

            --# Do code for sensitive shield users (Aeon)
            --# Could this be moved to a base class?
            if self.IsSensitiveShieldUser then
                --# Remove next line if shield toggles are added
                --# to production units
                self.MyShieldToggledOn = true
                self:EnableShield()
                
            --# Do code for damage limitiation system (DLS) users (UEF)
            --# Could this be moved to base class?
            elseif self.IsDamageLimitationSystemUser then
                self:AddDLSEffects()
            end
        end,

        OnProductionUnpaused = function(self) 
            --# Call base class version (outside of state) 
            --# which sets production active.
            Unit.OnProductionUnpaused(self)           
            --# Change state
            self.IsProductionPaused = false
            ChangeState(self, self.UnpausedState)
        end,
    }, 


    
    
    
    
    
    
    --#-----------------------------------------------------------
    --#      Special Upgrading Behaviour
    --#-----------------------------------------------------------
    --#*
    --#*  Gilbot-X says:  
    --#*
    --#*  The following 3 keys are used here and can
    --#*  be set by extending classes wishing to 
    --#*  use this extra functionality.
    --#*
    --#*  self.NotReadyToUpgrade = true/false   |  Use to delay upgrade so 
    --#*                                        |   an animatuon can be done first
    --#*  self.UpgradeBlocked = true/false      |  Use to prevent any upgrading so 
    --#*                                        |   a condition can be met first
    --#*  self.UpgradeBlockedMessage = string   |  Message to show when upgarding is 
    --#*                                        |   blocked, stating which condition 
    --#*                                        |   must be met first.
    --#*      
    --#*
    --#*  Notes:
    --#*
    --#*  When Unit.OnStartBuild is called, 
    --#*  some code that you can't access/modify will create a mesh for
    --#*  the unit we are upgrading to.  This can sometimes be undersireable.
    --#*  This override allows calls to ExtendedAdjacencyUnit.OnStartBuild to 
    --#*  be deferred until later, by intercepting and cancelling the order 
    --#*  so we have time to do anything first that we might want to do before
    --#*  the mesh of the new unit is created.
    --#*  
    --#*  For example, waiting until the Aeon MeX activeAnimation is in the 
    --#*  correct position that corresponds with the start of its upgrade animation. 
    --#*
    --#*  In the GPG base classes, OnStartBuild calls ChangeState to go into UpgradingState.
    --#*  Now, I override the UpgradingState to call this BaseClass version of
    --#*  OnStartBuild which will create the mesh, and it then executes the base class
    --#*  version of UpgradingState (main) to execute the rest of the upgarde animating.
    --#**
    OnStartBuild = function(self, unitBeingBuilt, order) 
        --# First, is this upgrade blocked due to a condition not being met?
        --# i.e. the unit trying to upgrade is not old enough yet?
        if self.UpgradeBlocked then 
            --# It is blocked, so simply cancel 
            --# the upgrade command, show message (optional) and then return.
            IssueStop({self})
            IssueClearCommands({self})
            if self.UpgradeBlockedMessage then
                if self.BuildableCategoryName then
                    self:AddBuildRestriction(categories[self.BuildableCategoryName])
                    self.RestrictionNeedsLifting = true
                end
                self:FlashMessage(self.UpgradeBlockedMessage, 5)
            end
        
        --# Upgrade was not blocked...
        else
            --# Check flags to see if the unit is ready to upgrade,
            --# i.e. flags can be set when the animation position is aligned
            --# ready for the upgarde animation manipulator to take over
            --# so that the animation doesn't jump, as for Aeon MeXes.
            if self.PreparingToUpgradeState and self.NotReadyToUpgrade then
                --# Unit isn't ready, so call code to prepare unit for upgrade.
                --# That state will call us back when its done.
                IssueStop({self})
                IssueClearCommands({self})
                ChangeState(self, self.PreparingToUpgradeState)
            --# We are ready to upgrade.
            else
                --# This links to code in timebasedresourceoutputunit.lua
                if self.IsTimeBasedOutputUnit then
                    self:DecideIfCanPassOnProductionPercentageIncreaseToUpgrade(unitBeingBuilt)
                end
                
                --# Store these values (so that 
                --# Unit.OnStartBuild can be called later)
                --# and start upgrading.             
                self.UnitBeingBuilt = unitBeingBuilt
                self.BuildOrder = order
                ChangeState(self, self.UpgradingState)
            end
        end
    end,

    
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  Called when Mex unit is given an upgrade command.
    --#*  This default version does nothing, it just acts as an interface template.
    --#*  Override this to make use of the functionality.
    --#*  For example, this is overrided by Autopack units where the 
    --#*  unit needs to put its animation in a certain position
    --#*  before it starts the upgrade command.
    --#**
    PreparingToUpgradeState = State {
        Main = function(self)
            --# Now we are ready to upgrade so let's try again!
            self.NotReadyToUpgrade = false
            local upgradeBPName = self:GetBlueprint().General.UpgradesTo or false
            if upgradeBPName then IssueUpgrade( {self}, upgradeBPName )  end
        end,
    },
    
    
          
    --#*
    --#*  Gilbot-X says:
    --#*        
    --#*  Normally this state gets initiated directly from OnStartBuild
    --#*  when a Mex unit is given an upgrade command.
    --#*  Now it's called from PreparingToUpgradeState so 
    --#*  units can do something first to prepare for upgrading
    --#*  when they are issued an upgrade command.
    --#**  
    UpgradingState = State {
    
        Main = function(self)
            --# In this class, UpgradingState is called by our own override
            --# of OnStartBuild which doesn't call its own base class version.
            --# so we must go back and call the base class version of OnStartBuild now.
            baseClassArg.OnStartBuild(self, self.UnitBeingBuilt, self.BuildOrder)
            
            --# Otherwise the upgrading state main function
            --# is the same as its base class.
            --# This next call gets the thread to 
            --# do the upgrading animation while the upgrade
            --# takes place.            
            baseClassArg.UpgradingState.Main(self)
        end,

        --# Defer this to base class code.
        OnStopBuild = baseClassArg.UpgradingState.OnStopBuild,

        --#*
        --#*  Gilbot-X says:
        --#*        
        --#*  I had to override this so if the upgrade is cancelled
        --#*  we go back into one of our states.
        --#**
        OnFailedToBuild = function(self)
            --# This next block is all GPG code.
            --# except I removed a call to PlayAnimation
            --# because that is dealt with by the appropriate states,
            --# and the call to ChangeState at the end.
            do --GPG Code from FA StructureUnit/Unit in defaultunits.lua
              Unit.OnFailedToBuild(self)
              self:EnableDefaultToggleCaps()
              if self.AnimatorUpgradeManip then self.AnimatorUpgradeManip:Destroy() end
              if self:GetCurrentLayer() == 'Water' then
                  self:StartRocking()
              end
              self:PlayUnitSound('UpgradeFailed')
              self:CreateTarmac(true, true, true, 
                                self.TarmacBag.Orientation, 
                                self.TarmacBag.CurrentBP)
            end --of GPG Code from FA StructureUnit in defaultunits.lua
            
            --# I added this.
            --# We disable sensitive shields when upgrading starts,  
            --# so they need to be reenabled here.  We can't disable
            --# it in this class because it needs to be disabled 
            --# before any potential unpack operation is executed.
            --# Therefore it is disabled in the autopack abstract class.
            if self.IsSensitiveShieldUser then 
                self.ShieldUsePermitted = true 
            end
            
            --# I also changed this.
            --# Instead of going to 'idle' state, we go to
            --# the 'paused' or 'unpaused' state.            
            ChangeState(self, self.PreviousState)
        end,
            
    },
    
  
}


return resultClass

end
