--#****************************************************************************
--#**
--#**  New File :  /mods/gilbotsimprovements/remotepipelines.lua
--#**
--#**  Author(s):  Gilbot-X
--#**
--#**  Summary  :  Definition of remote-link pipeline units 
--#**              common to all factions.
--#** 
--#****************************************************************************


--# These are new files I created.
local PipeLineUnit = 
    import('/mods/gilbotsimprovements/pipelines.lua').PipeLineUnit
local GilbotUtils = 
    import('/mods/gilbotsimprovements/utils.lua')
local ResourceNetwork = 
    import('/mods/gilbotsimprovements/resourcenetwork.lua').ResourceNetwork
local ResourceInterNetwork = 
    import('/mods/gilbotsimprovements/resourceinternetwork.lua').ResourceInterNetwork

local debugAdjacencyCode = false


--#*
--#*
--#*  Gilbot-X says:
--#*
--#*  Thsi class is for the T2 version of pipeline 
--#*  and versions higher up the tech tree.
--#*  
--#*  These must be build within a certain distance of each other
--#*  but do not need to have touching skirts.  They cannot 
--#*  link directly to units, but the can join up smaller networks made of 
--#*  T1 pipelines.
--#*  
--#*  These form intelligent networks that only make enough connections
--#*  to form a closed network, and they re-link to different connections
--#*  if the loss of a link isolates any units from a network.
--#*
--#***
RemotePipeLineUnit = Class(PipeLineUnit) {

    --# This constant shared variable provides a convenient means 
    --# of testing any object to see if it uses this class.
    IsRemotePipeLineUnit = true,

    
    OnCreate = function(self)
        --# Call superclass version
        PipeLineUnit.OnCreate(self)
        
        --# Gilbot-X says: 
        --# This is for my Maintenance Consumption Breakdown Mod
        --# which keep tracks of which / how many types of resource 
        --# consuming abilities are currently active in the unit.
        self.EnabledResourceDrains = {
            Adjacency = true,
        }
        
    end,

    
        
    --#*
    --#*  Gilbot-X says: 
    --#*  
    --#*  I added this function in ExtendedAdjacencyUnit to be called 
    --#*  from my override of OnStopBeingBuilt to apply
    --#*  adjacency as soon as the structures are built.
    --#*  This only runs once in a structure's lifetime.
    --#    
    --#*  This override is the same as the version in the 
    --#*  ExtendedAdjacencyUnit class, except here we try for remote
    --#*  Adjacency at the end.
    --#**
    AddExtendedAdjacencyToNearbyUnits = function(self)
        --# If direct or diagonal adjacency is dealt with immediately
        --# then hopefully there won't be too many duplicate joins to 
        --# the same network.
        PipeLineUnit.AddExtendedAdjacencyToNearbyUnits(self)

        --# Fork a thread so there can be  a pause after between each 
        --# remote adjacency is connected.
        ForkThread(self.FindAndCreateRemoteAdjacencyConnections, self)
        
    end,
    
    
      
 
        
    --#*
    --#*  Gilbot-X says: 
    --#*  
    --#*  Call this to examine all potential pipeline connections
    --#*  and then try to connect to the best.  Only connects to 
    --#*  pipelines in other networks and waits for other networks to 
    --#*  finish any reconstruction work.
    --#*
    --#*  I added this function to be called 
    --#*  from DoStatValueUpdateFunction for my adjacency range slider menu.
    --#**
    FindAndCreateRemoteAdjacencyConnections = function(self)
    
        --# If ACU is not alive then don't allow any more adjacency.
        if not self:GetMyCommander() then return end
    
        --# Concurrency protection.
        while self.WaitForSwitchOffRemoteLinkGiving do WaitTicks(5) end
        --# All salve operations assumed complete.
        --# Now we can rejoin networks.
        self.DontRejoinInterNetwork = false
        
        --# We are going to look at all possible connections 
        --# before we connect to anything.
        local PotentialConnections = {}
        --# This gets a list of units we can consider connecting to
        local structureTable = self:GetNearbyFriendlyStructures()
        
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if false and debugAdjacencyCode then
          LOG('Adjacency: FindAndCreateRemoteAdjacencyConnections:' 
              .. ' Found ' .. repr(table.getsize(structureTable)) .. ' units nearby.'
          )
        end
        
        --# Start looking for potential connections
        for kNearbyUnit, vNearbyUnit in structureTable do
      
            --# Next we will check if there are any pipelines in range
            --# that we haven't already connected to.
            local skirtSeparation = 
                self:CheckIfEligibleForRemoteAdjacency(vNearbyUnit)
            
            --# If this unit is eligible to connect remotely with us then
            if skirtSeparation then
            
                --# Both units need to have a valid 
                --# ResourceNetwork object in order to do this.
                self:EnsureUnitHasValidNetwork()
                vNearbyUnit:EnsureUnitHasValidNetwork()
        
                --# First wait for any network reconstruction to finish
                --# before recording the network Id of the unit 
                --# we will potentially connect to.
                while vNearbyUnit.MyNetwork.IsStillBeingReconstructed do
                    
                    WaitTicks(2)
                    
                    --# This next block is for debugging only.
                    --# This can be delete wheh debugging is done.
                    if debugAdjacencyCode then
                      WARN('Adjacency: FindAndCreateRemoteAdjacencyConnections:' 
                          .. ' waiting for network ' .. repr(vNearbyUnit.MyNetwork.NetworkId)
                          ..  ' to be reconstructed..' 
                      )
                    end
                end
                
                --# Now it's safe to query network ID. 
                local networkID = vNearbyUnit.MyNetwork.NetworkId
                
                --# This safety was necessary because we 
                --# were getting an error here about invalid table arguments
                --# caused by the network id - this seems to happen after 
                --# an ACU explosion (will possibly happen after a nuke).
                if networkID and networkID > 0 then 
                
                    --# Put all potential connections in a nested table grouped by network id.
                    if not PotentialConnections[networkID] then
                        
                        --# This next block is for debugging only.
                        --# This can be delete wheh debugging is done.
                        if false and debugAdjacencyCode then
                          LOG('Adjacency: FindAndCreateRemoteAdjacencyConnections:' 
                              .. ' first potential connection for network=' .. repr(networkID)
                              .. ' is ' .. vNearbyUnit.DebugId
                          )
                        end
                        
                        PotentialConnections[networkID] = {
                            Unit = vNearbyUnit,
                            SkirtSeparation = skirtSeparation,
                        }
                    else
                        --# This next block is for debugging only.
                        --# This can be delete wheh debugging is done.
                        if false and debugAdjacencyCode then
                          LOG('Adjacency: FindAndCreateRemoteAdjacencyConnections:' 
                              .. ' another potential connection for network=' .. repr(networkID)
                              .. ' is ' .. vNearbyUnit.DebugId
                          )
                        end
                        
                        --# Compare this potential connection and the previous
                        --# one.  The best one is kept.
                        PotentialConnections[networkID] = 
                            self:ChooseBetterConnection(
                                PotentialConnections[networkID],
                                {Unit = vNearbyUnit, 
                                 SkirtSeparation = skirtSeparation}
                            )                            
                    end
                
                else
                    --# This next block is for debugging only.
                    --# This can be delete wheh debugging is done.
                    if debugAdjacencyCode then
                      WARN('Adjacency: FindAndCreateRemoteAdjacencyConnections: ' 
                          .. vNearbyUnit.DebugId .. ' has no network ID.'
                      )
                    end
                end
            end
	end
        
        for kNetworkId, vConnection in PotentialConnections do
            --# This will check if the units qualify for remote adjacency.
            --# If they qualify, they will also be given it.
            self:TryToGiveOrReceiveRemoteAdjacency(vConnection.Unit)
        end
        
    end,
    
    
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  from FindAndCreateRemoteAdjacencyConnections in this class.
    --#*  That function is always forked as a thread, and so this runs in that thread.
    --#*
    --#*  Unlike the calling function, this code gets executed 
    --#*  more than once: if networks are broken it gets called again.
    --#*
    --#*  Therefore this function separates the code into reusable sections
    --#*  and makes the code more readable.
    --#**
    CheckIfEligibleForRemoteAdjacency = function(self, nearbyUnitArg)
        
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if false and debugAdjacencyCode then
          LOG('Adjacency: CheckIfEligibleForRemoteAdjacency: Called by '
              .. self.DebugId .. ' onto ' .. nearbyUnitArg.DebugId
          )
        end
        
        --# Make sure other unit is a pipeline.  We don't need to check if 
        --# we are a RemotePipeLineUnit, because this function is defined 
        --# in the RemotePipeLineUnit class, so we must be.
        if not nearbyUnitArg.IsRemotePipeLineUnit then 
            
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAdjacencyCode then
              LOG('Adjacency: CheckIfEligibleForRemoteAdjacency: '
                  .. nearbyUnitArg.DebugId .. ' is not a remote pipeline.'
              )
            end
            
            --# Notify calling code what it wants to know;
            --# These two units cannot get a remote adjacency connection.
            return false
        end
        
        --# Don't connect to a unit or network that we are already connected to.
        if self:CheckIfAlreadyConnected(nearbyUnitArg, 'Local') 
        or self:CheckIfAlreadyConnected(nearbyUnitArg, 'Remote') 
        or
          --# Don't join a local network we are already part of!
          ( self.MyNetwork and nearbyUnitArg.MyNetwork and 
              (self.MyNetwork == nearbyUnitArg.MyNetwork)
          )
          --# Don't join an internetwork we are already part of
        or 
          ( self.MyNetwork and self.MyNetwork.MyInterNetwork and
            nearbyUnitArg.MyNetwork and nearbyUnitArg.MyNetwork.MyInterNetwork and
              (self.MyNetwork.MyInterNetwork == nearbyUnitArg.MyNetwork.MyInterNetwork)
          )
        then 
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAdjacencyCode then
              LOG('Adjacency: CheckIfEligibleForRemoteAdjacency: '
                  .. self.DebugId .. ' and ' 
                  .. nearbyUnitArg.DebugId .. ' are already conected.'
              )
            end
            
            --# Notify calling code what it wants to know;
            --# These two units cannot get a remote adjacency connection.
            return false
        end
        
        --# Calculate maximum allowed separation based 
        --# on the remote adjacency ranges these units have.
        local myRange, yourRange = 0, 0
        
        --# Check if other pipeline unit is paused.
        if not nearbyUnitArg.IsRemoteAdjacencyPaused then 
            --# The other pipeline is not paused so use its max range.
            yourRange = nearbyUnitArg.RemoteAdjacencyAllowedSeparationDistance
        end
        --# Check if I'm paused.
        if not self.IsRemoteAdjacencyPaused then 
            --# I'm not paused so use my max range.
            myRange = self.RemoteAdjacencyAllowedSeparationDistance
        end
        --# Take the bigger of the two units' ranges as max.  
        local maxSkirtSeparationForAdjacency = math.max(myRange, yourRange)
                      
              
        --# If these units are in each other's range then
        skirtSeparation = self:GetSkirtSeparationDistance(nearbyUnitArg)        
        if maxSkirtSeparationForAdjacency > 0 and 
          skirtSeparation <= maxSkirtSeparationForAdjacency then
          
            if false and debugAdjacencyCode then
              LOG('Adjacency: CheckIfEligibleForRemoteAdjacency: '
                  .. self.DebugId .. ' and ' 
                  .. nearbyUnitArg.DebugId .. ' can be connected.'
                  .. ' Skirts are ' .. repr(skirtSeparation) .. ' O-grids apart.'
              )
            end
            
            --# Notify calling code what it wants to know;
            --# These two units can get a remote adjacency connection.
            return skirtSeparation
        else
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAdjacencyCode then
              LOG('Adjacency: CheckIfEligibleForRemoteAdjacency: '
                  .. self.DebugId .. ' and ' 
                  .. nearbyUnitArg.DebugId .. ' are out of each others range.'
              )
            end
            
            --# Notify calling code what it wants to know;
            --# These two units cannot get a remote adjacency connection.
            return false
        end 
        
    end,
    
    
    --#*
    --#*  Gilbot-X says: 
    --#*  
    --#*  Given two possible connections to the same network as arguments,
    --#*  it returns the pipeline that is a better way to connect to that network.
    --#*  If they are equally desireable connections then just return the first by default.
    --#*
    --#*  Both connection arguments must be a table with two string-key fields:
    --#*     { Unit = ??, SkirtSeparation = ?? }
    --#*
    --#*  One of those arguments will be returned.
    --#*
    --#*  I added this function to be called from
    --#*  FindAndCreateRemoteAdjacencyConnections in this class, which is where 
    --#*  I extracted the code from.
    --#*  That function is always forked as a thread, and so this runs in that thread.
    --#*
    --#*  This function gets executed with higher frequency than its calling function.
    --#*  If new pipelines are built or networks are broken then it gets called again.
    --#*
    --#*  I moved this code into this function to make the code more readable.
    --#**
    ChooseBetterConnection = function(self, connectionArg1, connectionArg2)
      
        --# Check if they are different tech levels
        mytl = self:GetMyTechLevel()
        tl1 = connectionArg1.Unit:GetMyTechLevel()
        tl2 = connectionArg2.Unit:GetMyTechLevel()
        
        --# Prefer to link with pipeline of the same tech level
        --# or fewer tech levels away
        if math.abs(mytl - tl1) < math.abs(mytl - tl2) then return connectionArg1 end
        if math.abs(mytl - tl1) > math.abs(mytl - tl2) then return connectionArg2 end
        
        --# if both potential connections are the same number of tech levels away 
        --# (i.e. we are a T2 pipeline and we choose between T1 and T3 pipelines)
        --# then choose to connect to the higher tech level.
        if (mytl - tl1) > (mytl - tl2) then return connectionArg1 end
        if (mytl - tl1) < (mytl - tl2) then return connectionArg2 end
        
        --# Find out how far away they both are
        local d1 = connectionArg1.SkirtSeparation
        local d2 = connectionArg2.SkirtSeparation
        
        --# Choose the closest or the first if they are equal.
        if d1 < d2 then return connectionArg1 else return connectionArg2 end
        
    end,
    
    
    
    
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  from FindAndCreateRemoteAdjacencyConnections in this class.
    --#*  That function is always forked as a thread, and so this runs in that thread.
    --#*
    --#*  Unlike the calling function, this code gets executed 
    --#*  more than once: if networks are broken it gets called again.
    --#*
    --#*  Therefore this function separates the code into reusable sections
    --#*  and makes the code more readable.
    --#**
    TryToGiveOrReceiveRemoteAdjacency = function(self, nearbyUnitArg)
        
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if false and debugAdjacencyCode then
          LOG('Adjacency: TryToGiveOrReceiveRemoteAdjacency: Called by '
              .. self.DebugId .. ' onto ' .. nearbyUnitArg.DebugId
          )
        end
        
        --# This makes sure we don't join a network twice while 
        --# it is being rebuilt from a broken one and more nodes 
        --# are being added.
        while(nearbyUnitArg.MyNetwork.IsStillBeingReconstructed) do
            WaitTicks(5)
        end
        
        --# Connect units with a beam and swap adjacency bonuses
        self:OnRemotelyAdjacentTo(nearbyUnitArg) 
                   
        --# Now wait while our network and theirs are being merged.
        while(self.MyNetwork.IsStillBeingReconstructed) do
            WaitTicks(5)
        end
  
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  I added this.  It is called by AddExtendedAdjacencyToNearbyUnits
    --#*  defined above.  Each unit keeps a reference to the other.
    --#**
    OnRemotelyAdjacentTo = function(self, adjacentUnitArg)
         
        --# Safety checks are grouped together in another function
        if not self:CanLinkWith(adjacentUnitArg) then return end

        --# Call code that deals with resource networks and internetworks
        self:AddLinkToInterNetwork(adjacentUnitArg)
        
        --# Create link effect and register connection with connected units
        self:CreateLink(adjacentUnitArg)
        
        --# Make sure there are appropriate tables.       
        self:AddToEachOthersAdjacencyTables(adjacentUnitArg, 'Remote') 
    end,
    
    
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Called by OnAdjacentTo.
    --#*  Deals with ResourceNetwork objects.
    --#**
    AddLinkToInterNetwork = function(self, adjacentUnitArg)
        
        --# For safety, although not sure this is necessary.
        self.MyNetwork.RemoveMeFromMyInterNetwork = false
        adjacentUnitArg.MyNetwork.RemoveMeFromMyInterNetwork = false
        
        --# Only create and associate with a brand new network supervsior
        --# if we haven't had one already or the last one was broken up and is invalid.
        local useMyInterNetwork = true
        --# First check if we have a valid internetwork.
        if (not self.MyNetwork.MyInterNetwork) or self.MyNetwork.MyInterNetwork.IsBroken then
            --# if the adjacent unit has a functioning internetwork
            if adjacentUnitArg.MyNetwork.MyInterNetwork and 
            (not adjacentUnitArg.MyNetwork.MyInterNetwork.IsBroken) then
            
                --# This next block is for debugging only.
                --# This can be delete wheh debugging is done.
                if false and debugAdjacencyCode then
                  LOG('Adjacency: AddLinkToInterNetwork:'
                      .. ' Using internetwork object on ' .. adjacentUnitArg.DebugId
                  )
                end
            
                --# Add it to the adjacent unit's network because 
                --# has a valid one and we don't.
                adjacentUnitArg.MyNetwork.MyInterNetwork:AddRemoteLinkToInterNetwork(
                    self.MyNetwork
                )
                useMyInterNetwork = false
            else
                --# This next block is for debugging only.
                --# This can be delete wheh debugging is done.
                if false and debugAdjacencyCode then
                  LOG('Adjacency: AddLinkToInterNetwork:'
                      .. ' Attaching new internetwork object to ' .. self.DebugId
                  )
                end
                
                --# We'll make a new network for this unit
                --# because neither unit had a network.
                --# This is the only time/place we create
                --# a new internetwork.                 
                self.MyNetwork.MyInterNetwork = ResourceInterNetwork(self.MyNetwork)
            end
        end
        --# Add the link to this unit's (possibly new) network if we didn''t
        --# already add it to the other unit's network.
    	if useMyInterNetwork then 
        
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if debugAdjacencyCode and not self.MyNetwork.MyInterNetwork then
              ERROR('Adjacency: AddLinkToInterNetwork:'
                  .. ' internetwork object missing on ' .. self.DebugId
              )
            end
            
            self.MyNetwork.MyInterNetwork:AddRemoteLinkToInterNetwork(
                adjacentUnitArg.MyNetwork
            ) 
        end
    end,
    
    
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Called by AddLinkToInterNetwork.
    --#**
    EnsureUnitHasValidNetwork = function(self)
        --# For safety, although not sure this is necessary.
        self.RemoveMeFromMyNetwork = false
        
        --# This attaches a new network object to the unit if 
        --# there isn't a valid one there already
        if (not self.MyNetwork) or self.MyNetwork.IsBroken then
            --# We'll make a new network for this unit
            self.MyNetwork = ResourceNetwork(self)
            
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAdjacencyCode then
              LOG('Adjacency: EnsureUnitHasValidNetwork:'
                  .. ' Attaching new network object to ' .. self.DebugId
                  .. ' with ID=' .. repr(self.MyNetwork.NetworkId)
              )
            end
        end
    end,
        
        
        
        
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Had to mod this to make sure tables are cleaned up.
    --#*  Base class version cancels bonuses and cleans up effects.
    --#*  
    --#**
    OnNotRemotelyAdjacentTo = function(self, adjacentUnitArg)

        --# Safety checks are gropuped together in another function
        if not self:CanDisconnectFrom(adjacentUnitArg, 'Remote') then return end
        
        --# Always remove a linking beam effect regardless
        --# of whether or not bonus is applied:
        --# Remove any adjacency beam effects between this pair.
        --# Also removes references to each other from all connection tables,
        --# which stops this function from being called a second time on the same pair.
        self:CleanUpAdjacentEffectsWith(adjacentUnitArg)
        
        --# Both units will have same network object because they were attached
        local internetwork = self.MyNetwork.MyInterNetwork or adjacentUnitArg.MyNetwork.MyInterNetwork 
        --# Contact the network to remove this link from network.
        --# First check if we have the correct variables to make the call.
    	if internetwork then
            internetwork:OnRemoteLinksRemovedFromInterNetwork()
    	end
  
    end,
    
    
    
    
    
   
    --#*
    --#*  Gilbot-X says:  
    --#* 
    --#*  Had to override this to make sure that tables are 
    --#*  cleaned up when a unit completes an upgrade and 
    --#*  destroys itself.
    --#*
    --#*  This function is defined in Unit.lua
    --#*  and ends by changing state to DeadState
    --#*  so I have to call my extra code before calling
    --#*  the base class code.
    --#**
    OnDestroy = function(self)
        --# Gilbot-X says:  
        --# I added this block of code so that when 
        --# units upgrade they get cleaned out of ResourceNetwork 
        --# tables and the ACU's auto-toggle tables.
        self:CleanUpRemoteAdjacency() 
             
        --# Finally the rest is GPG code.
        PipeLineUnit.OnDestroy(self)
    end,
    
    
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is required by DoStatValueUpdateFunction for my 'Stat Slider' mod.
    --#*  I need to be able to remove all remote adjacency, which normally happens
    --#*  when one of the units dies.  Here I remove it without any death
    --#*  so we can re-apply it with a larger radius.
    --#*
    --#** 
    CleanUpRemoteAdjacency = function(self)
    
        --# For each unit we have to remove
        for kAdjacentUnit, vAdjacentUnit in self.AdjacentUnits['Remote'] do
        
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAdjacencyCode then
              LOG('Adjacency: CleanUpRemoteAdjacency: Called by '
                  .. self.DebugId .. ' onto ' .. vAdjacentUnit.DebugId
              )
            end
        
            --# Call OnNotAdjacentTo in only one direction 
            --# to break the link.  Each call will cause a network split.
            self:OnNotRemotelyAdjacentTo(vAdjacentUnit)
            --# Wait before proceeding so the network's split operation can complete.
        end
    end,

    
    
    
    
    --#*
    --#* Gilbot-X says:
    --#*
    --#* When production is paused, remote-links are disabled.
    --#* This is an override of functions defined in unit.lua but I 
    --#* have also hooked these functions in this mod.
    --#**
    OnScriptBitSet = function(self, bit)
        if bit == 4 then --# production pause toggle
            --# Record state
            self.IsRemoteAdjacencyPaused = true
            --# These next two lines go together
            ForkThread(self.SwitchOffRemoteLinkGiving, self)
            self.WaitForSwitchOffRemoteLinkGiving = true
            --# These next two lines go together
            self.EnabledResourceDrains.Adjacency = false
            self:UpdateConsumptionValues()
        else
            --# Base class version will deal with shields etc.
            PipeLineUnit.OnScriptBitSet(self, bit)
        end
    end,

    --#*
    --#* Gilbot-X says:
    --#*
    --#* When production is unpaused, remote-links are re-established.
    --#* This is an override of functions defined in unit.lua but I 
    --#* have also hooked these functions in this mod.
    --#**
    OnScriptBitClear = function(self, bit)
        --# Gilbot-X says: I added this to call my code
        if bit == 4 then --# production pause toggle
            --# Record state
            self.IsRemoteAdjacencyPaused = false
            --# This thread will automatically wait until the other thread 
            --# running SwitchOffRemoteLinkGiving has marked that it has finished.
            ForkThread(self.FindAndCreateRemoteAdjacencyConnections, self)
            --# These next two lines go together
            self.EnabledResourceDrains.Adjacency = true
            self:UpdateConsumptionValues()
        else
            --# Base class version will deal with shields etc.
            PipeLineUnit.OnScriptBitClear(self, bit)
        end
    end,
    
    
    
     
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is required by the slider control.
    --#*  It must update whatever property, feature or variable that 
    --#*  the slider control declared in this unit's BP file was designed to adjust.
    --#**
    DoStatValueUpdateFunction = function(self, statType, newStatValue)
        
        if statType == 'ShieldMaxHealth' then 
            --# Update shield strength.  No need to store value.
            self.MyShield:SetMaxHealthAndRegenRate(newStatValue)
        
        elseif statType == 'AdjacencyExtensionDistance' then 

            --# Rebuild remote adjacency connections.
            
            --# First we need to do make sure our internetwork is split, 
            --# and then salvaged into separate parts, otherwise
            --# we won't be able to reconnect to any of them that
            --# are in reach of the new range.
            
            --# If the unit is paused, there is nothing to turn off
            if not self.IsRemoteAdjacencyPaused then
                --# Fork a thread to iteratively break each remote link 
                --# so there can be a pause after/between each 
                --# remote connection removal.
                ForkThread(self.SwitchOffRemoteLinkGiving, self)
                self.WaitForSwitchOffRemoteLinkGiving = true
            end
            
            --# Record new value from slider menu
            self.RemoteAdjacencyAllowedSeparationDistance = newStatValue
            
            --# Update red circle effect that illustrates adjacency range
            local gun = self:GetWeapon(1)
            gun:ChangeDamageRadius(newStatValue)
            gun:ChangeMaxRadius(newStatValue)
            
            --# If the unit is paused, there is nothing to turn on
            if not self.IsRemoteAdjacencyPaused then
                --# This thread will automatically wait until the other thread 
                --# running SwitchOffRemoteLinkGiving has marked that it has finished.
                ForkThread(self.FindAndCreateRemoteAdjacencyConnections, self)
            end
        end
        
    end,
    
    
        
    --#*
    --#*  Gilbot-X says: 
    --#*     
    --#*  I added this function to be called 
    --#*  from SwitchOffRemoteLinkGiving in this class.
    --#*  That function is always forked as a thread, and so this runs in that thread.
    --#*
    --#*  Unlike the calling function, this code gets executed 
    --#*  more than once: if networks are broken it gets called again.
    --#*
    --#*  Therefore this function separates the code into reusable sections
    --#*  and makes the code more readable.
    --#**
    CheckIfCanStillReceiveRemoteLinkWith = function(self, nearbyUnitArg)
        
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if false and debugAdjacencyCode then
          LOG('Adjacency: CheckIfCanStillReceiveRemoteLinkWith: Called by '
              .. self.DebugId .. ' onto ' .. nearbyUnitArg.DebugId
          )
        end
        
        --# Don't disconnect to a unit or network that we are already connected to.
        if not self:CheckIfAlreadyConnected(nearbyUnitArg, 'Remote') then
            --# Notify calling code what it wants to know;
            --# These two units cannot get a remote adjacency connection.
            return false
        end
        
        --# Check if other pipeline unit is paused.
        if nearbyUnitArg.IsRemoteAdjacencyPaused then 
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAdjacencyCode then
              LOG('Adjacency: CheckIfCanStillReceiveRemoteLinkWith: '
                  .. self.DebugId .. ' and ' 
                  .. nearbyUnitArg.DebugId .. ' are both switched off.'
              )
            end
            
            --# Notify calling code what it wants to know;
            --# These two units cannot get a remote adjacency connection.
            return false
        end 
        
        --# If these units are in each other's range then
        skirtSeparation = self:GetSkirtSeparationDistance(nearbyUnitArg)        
        if skirtSeparation <= nearbyUnitArg.RemoteAdjacencyAllowedSeparationDistance then
          
            if false and debugAdjacencyCode then
              LOG('Adjacency: CheckIfCanStillReceiveRemoteLinkWith: '
                  .. self.DebugId .. ' can remain connected to ' 
                  .. nearbyUnitArg.DebugId
              )
            end
            
            --# Notify calling code what it wants to know;
            --# These two units can get a remote adjacency connection.
            return true
        else
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAdjacencyCode then
              LOG('Adjacency: CheckIfCanStillReceiveRemoteLinkWith: '
                  .. self.DebugId .. ' is out of range of ' 
                  .. nearbyUnitArg.DebugId
              )
            end
            
            --# Notify calling code what it wants to know;
            --# These two units cannot get a remote adjacency connection.
            return false
        end 
        
    end,
    
    
    
      
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is required by DoStatValueUpdateFunction for my 'Stat Slider' mod.
    --#*  I need to be able to remove all remote adjacency, which normally happens
    --#*  when one of the units dies.  Here I remove it without any death
    --#*  so we can re-apply it with a larger radius.
    --#*
    --#** 
    SwitchOffRemoteLinkGiving = function(self)
    
        --# For each unit we have to remove
        for kAdjacentUnit, vAdjacentUnit in self.AdjacentUnits['Remote'] do
        
            if self:CheckIfCanStillReceiveRemoteLinkWith(vAdjacentUnit) then
                break
            end
            
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if false and debugAdjacencyCode then
              LOG('Adjacency: SwitchOffRemoteLinkGiving: Called by '
                  .. self.DebugId .. ' onto ' .. vAdjacentUnit.DebugId
              )
            end
        
            --# Call OnNotAdjacentTo in only one direction 
            --# to break the link.  Each call will cause a network split.
            self:OnNotRemotelyAdjacentTo(vAdjacentUnit)
            --# Wait before proceeding so the internetwork's split operation can complete.
            local waitMore = true
            while waitMore do 
                --# Waiting for network split/salvage peration to complete
                WaitTicks(5)
                --# Check if its finished (valid network that is not being reconstructed)
                if vAdjacentUnit.MyNetwork and not vAdjacentUnit.MyNetwork.IsBroken then
                    waitMore = vAdjacentUnit.MyNetwork.IsStillBeingReconstructed
                end
            end
        end
        
        --# Allow calling thread to continue
        self.WaitForSwitchOffRemoteLinkGiving = false
    end,
    
  
    
} --(end of RemotePipeLineUnit class)