--#*****************************************************************************
--#* New File : /mods/gilbotsimprovements/resourcenetwork.lua
--#*
--#* Modded By: Original version by Lakitrid, but rewritten by Gilbot-X
--#*
--#* Summary  : Network management system for adjacency links in a network
--#*            connected by 1 or more pipeline nodes.
--#*
--#*****************************************************************************
local GilbotUtils = 
    import('/mods/gilbotsimprovements/utils.lua')
local ApplyBonusCap = 
    import('/mods/gilbotsimprovements/adjacencybonustypes.lua').ApplyBonusCap


local debugAdjacencyCode = false

ResourceInterNetwork = Class
{

    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is called in function OnAdjacentTo in the 
    --#*  ExtendedAdjacencyUnit class.  All units from that class
    --#*  must have a reference to an object of this class once
    --#*  they become adjacent and connected to another unit in
    --#*  forming part of a network.  This class stores and manages
    --#*  all the necessary information about that network,
    --#*  including which units have joined it, and a record of
    --#*  which ones can give and receive each type of bonus.
    --#*  Functions of this class are called to calculate what
    --#*  bonuses should be spread arouind the network, and to
    --#*  switch on/off the distribution of those bonuses.
    --#*  Units that can join networks just have to call code
    --#*  in this class that either adds or removes them from the 
    --#*  network.
    --#**
    __init = function(self, firstNode, optionalArg)
        
        --# Initialise data structures        
        self:ClearNetworkTables()

        --# Safety: firstNode is not always supplied.
        if firstNode then self:AddLocalNetworkToInterNetwork(firstNode)  
        else
            if optionalArg == 'SpawningFromBroken' then
                self.IsStillBeingReconstructed = true
            else
                WARN('Adjacency: Internetwork object was created with ' 
                      .. 'no initial remote link node in constructor.')
            end
        end
    end,
    
   
  
   
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is only called by the constructor to initialise 
    --#*  and in the BreakNetworkAndSalvagePieces function to 
    --#*  make the network object safe (unusable because it contains no units).
    --#**
    ClearNetworkTables = function(self)
        --# Use this to keep a list of linking nodes in the internetwork
        self.AttachedNodesTable = {} 
    end,
        
        

    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is only called by AddToInterNetwork to initialise 
    --#*  network properties that require the first unit to be added 
    --#*  to the network.  The first unit gives us a reference to the
    --#*  army ID and thus we can contact the ACU if it is alive.
    --#**
    SetFirstNode = function(self, localNetworkArg)
        --# Keep a reference to the army so we can check for capturing
        self.Army = localNetworkArg.Army
        self.MyCommander = localNetworkArg.MyCommander
        
        --# The supervisor's first node is always a pipeline node
        --# because this is how the class is created.
        self.FirstNode = localNetworkArg
         
        --# If commander is alive...
        if self.MyCommander and not 
          (self.MyCommander:BeenDestroyed() or self.MyCommander:IsDead())
        then
            --# Put a reference to us in the ACU's table
            self.MyCommander:ReceiveACUMessage_AddNewInterNetwork(self)
        else
            --# Prevent any more checks to see if ACU is dead
            --# because we know it is already
            self.MyCommander = nil
        end
    end,
 
 
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is only for debugging.
    --#**
    TellACU_DumpAllActiveNetworks = function(self)

        --# If commander is alive...
        if self.MyCommander and not 
          (self.MyCommander:BeenDestroyed() or self.MyCommander:IsDead())
        then
            self.MyCommander:ReceiveACUMessage_DumpActiveNetworks()
        else
            --# Prevent any more checks to see if ACU is dead
            --# because we know it is already
            self.MyCommander = nil
        end
    end,
    
    
    
 
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is only called by ExtendedAdjacencyUnit nodes 
    --#*  when their OnAdjacentTo function is called, i.e. when
    --#*  two units are being connected to each other.
    --#*
    --#*  The creation of adjacency beam effects are 
    --#*  taken care of by the ExtendedAdjacencyUnit class.
    --#*  In this class, we just have to add appropriate references to the 
    --#*  two structures to a network.  There could be other units attached to 
    --#*  either of the two units being connected here, so we actually 
    --#*  merge one network into another if they are both already part of one.
    --#**
    AddRemoteLinkToInterNetwork = function(self, localNetworkArg2)
        
        --# The first unit must have a network to be running this function.
        --# If the other unit has no valid network,
        if (not localNetworkArg2.MyInterNetwork) or localNetworkArg2.MyInterNetwork.IsBroken then
            --# The other unit has just been built and has no 
            --# network supervisor yet, so it now will associate with ours.
            --# No need to merge.
            self:AddLocalNetworkToInterNetwork(localNetworkArg2)
            --# Change this so we only update if the 
            --# network that joined had adjacency givers or receivers
            if localNetworkArg2:DoesImpactInterNetworkBonus() then 
                self:PropogateBonusesThroughInterNetwork() 
            end
            
        --# They both have networks, so if they are not already the same 
        --# network we'll have to merge them both.
        elseif self ~= localNetworkArg2.MyInterNetwork then            
            --# Now merge the smaller of the two networks into the larger.
            --# Bonuses will automatically be switched on after the merge.
            --# Check who has most units before deciding who merges into who.
            local nodesAlreadyInMyInterNetwork = 
                table.getsize(self.AttachedNodesTable)
            
            local nodesAlreadyInOtherInterNetwork =  
                table.getsize(localNetworkArg2.MyInterNetwork.AttachedNodesTable)
                
            --# If this internetwork has more nodes..
            if nodesAlreadyInMyInterNetwork > nodesAlreadyInOtherInterNetwork
            then
                --# merge the other internetwork into me
                self:MergeInterNetworks(localNetworkArg2.MyInterNetwork)
            else
                --# we merge into them
                localNetworkArg2.MyInterNetwork:MergeInterNetworks(self)
            end
        end
        --# Otherwise if they have the same supervisor it means there 
        --# is a loop in the network.  Loops happen and are not a problem.

        --# This next line is for debugging only.
        --# This can be delete wheh debugging is done.
        if debugAdjacencyCode then self:TellACU_DumpAllActiveNetworks() end
    end,
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This can be called by RemotePipeLineUnit nodes when their 
    --#*  OnNotRemotelyAdjacentTo function is called.  This function then decides 
    --#*  if they can just drop either of the two units from the network, 
    --#*  or if the entire network needs to be broken up.
    --#* 
    --#*  Adjacency beam effects are taken care of by the ExtendedAdjacencyUnit
    --#*  class. We just have to remove all bonuses and references to these units 
    --#*  from the network.
    --#*  
    --#*  If there is no need to break up this network into two or more subnetworks,
    --#*  then the units remianing in the network get recalculated bonuses.
    --#*  If the network does get broken up into subnetworks, these subnetworks 
    --#*  will get their own new ResourceInterNetwork object with its own set of 
    --#*  recalculated adjacency bonuses.
    --#**
    OnRemoteLinksRemovedFromInterNetwork = function(self, breakInternetworkArg)
         
        --# The internetwork may have been broken on a previous call to this function
        --# but we allow subsequent calls to get this far anyway as it stops other 
        --# units in the network that died just a tick or so afterwards 
        --# from being added to newly spawned networks.
        if self.IsBroken then return end
        self.IsBroken = true
        if not self.FindNewRemoteBridgesThread then
            --# Leave the other nodes have reference to this network
            --# so they can get DontRejoinNetwork called, so we don't
            --# try to reinclude them when we salvage after the pause.
            self.FindNewRemoteBridgesThread = ForkThread(
                self.BreakInterNetworkAndCreateNewSubInterNetworks, self
            )
        end
    end,
    
    
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is forked as a thread from OnNotAdjacentTo defined above.
    --#*  We inspect the list of remaining units in the old network and 
    --#*  try to break it down into subnetworks that can have their own bonuses.
    --#**
    BreakInterNetworkAndCreateNewSubInterNetworks = function(self) 
      
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if debugAdjacencyCode then
          LOG('Adjacency: BreakInterNetworkAndCreateNewSubInterNetworks: '
           .. 'Breaking Internetwork id=' .. repr(self.InterNetworkId)
          )
        end
       
        --# Allow time for chain reactions to finish.
        --# These can wipe out entire networks in less than a second.
        --# Bigger explosions from nukes will take longer than a second
        --# and these may cause problems.
        WaitSeconds(1)

        --# By this point, all nodes in this network now have a reference
        --# that points to a broken network.  This network can no longer 
        --# give bonuses.  Units can no longer be added to it.
        --# The nodes that were in it now need to be reassigned.
        
        --# Only allow networks to be rebuilt if the
        --# army's ACU is still alive.
        if not GilbotUtils.GetCommanderFromArmyId(self.Army) then return end
        
        local newSubInternetworksTable = {}
        --# Use the list of pipeline nodes still in the broken network
        --# as a starting point to try rebuild new subnetworks.
        for arrayIndex, vNetwork in self.AttachedNodesTable do
            if vNetwork and (not vNetwork.IsBroken) 
              and not GilbotUtils.IsValueIn2DTable(newSubInternetworksTable, vNetwork) then
                --# Create a new subnetwork starting from this unit
                local newSubInterNetwork = {}
                table.insert(newSubInterNetwork, vNetwork)
                newSubInterNetwork = 
                    self:TryToAddMoreLocalNetworksToNewSubInterNetwork(vNetwork, newSubInterNetwork)
                table.insert(newSubInternetworksTable, newSubInterNetwork)
            end
        end
        
        --# Clear all our network tables now as this network object should 
        --# never be used again.
        self:ClearNetworkTables()
        
        --# Spawn any subnetworks we can.
        if self:VerifyValidSubInterNetworkTable(newSubInternetworksTable) then
            self:SpawnNewInternetworks(newSubInternetworksTable)
        end
        
        --# This next line is for debugging only.
        --# This can be delete wheh debugging is done.
        if debugAdjacencyCode then self:TellACU_DumpAllActiveNetworks() end
        
    end, 
             
            
    
   
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is called by BreakNetworkAndSalvagePieces defined above.
    --#*  We inspect the broken network to try to find and create
    --#*  subnetworks.  Each ExtendedAdjacencyUnit in the network has a list 
    --#*  it maintains of which other units it is directly connected to.
    --#*  We recursively search through these lists to expand new subnetworks
    --#*  until no more units can be added to them.
    --#**
    TryToAddMoreLocalNetworksToNewSubInterNetwork = function(self, networkArg, newSubInterNetwork)
        
        --# Iterate through all remote links this network has
        for unusedArrayIndex1, vRemotePipeLineUnit in networkArg.AttachedRemoteLinksTable do
        
            --# Look at all the other networks this remote link is connecting to remotely
            for unusedArrayIndex2, vAdjacentRemotePipeLineUnit 
                                          in vRemotePipeLineUnit.AdjacentUnits['Remote'] do
    
                --# if the pipeline on the other side isn't 
                --# a dead/recalaimed/captured remote-link pipeline pipeline
                if not vAdjacentRemotePipeLineUnit.DontRejoinNetwork 
                  --# check for a dead/reclaimed remote-link pipeline terminal
                  and (not vAdjacentRemotePipeLineUnit:IsDead()) 
                  --# check for captured remote-link pipeline terminal
                  and vAdjacentRemotePipeLineUnit:GetArmy() == self.Army
                  --# check that its network is not broken
                  and (not vAdjacentRemotePipeLineUnit.MyNetwork.IsBroken)
                  --# check its network is not in a subnetwork already.
                  --# it shouldn't be in a different subnetwork!!              
                  and not GilbotUtils.IsValueInTable(
                              newSubInterNetwork, 
                              vAdjacentRemotePipeLineUnit.MyNetwork
                          ) 
                  then
                    --# this can go in our subnetwork.
                    table.insert(newSubInterNetwork, vAdjacentRemotePipeLineUnit.MyNetwork)
                    --# Try to add whatever this newly added network
                    --# is connected to it in the same way with this recursive call.
                    newSubInterNetwork = 
                        self:TryToAddMoreLocalNetworksToNewSubInterNetwork(
                            vAdjacentRemotePipeLineUnit.MyNetwork, newSubInterNetwork)
                end
            end
            
        end
        
        --# return result
        return newSubInterNetwork
    end,
    
    
   
        
    --#*
    --#*  Gilbot-X says:
    --#*
    --#*  This is called by the function BreakInterNetworkAndCreateNewSubInterNetworks.
    --#*  It returns true if none of the units in the subnetworks tables
    --#*  have redundant entries.  Because a network can not be added to more than 
    --#*  one internetwork at a time, this function exists as a safety check to make sure 
    --#*  that never happens.
    --#*  
    --#*  ?? - I still get warnings that this function has caught and blocked some
    --#*  invalid networks from being spawned, so we need to keep this code.
    --#**
    VerifyValidSubInterNetworkTable = function(self, newSubInternetworksTable)
        --# No unit should appear twice in the same network.
        --# Examine all pairings of networks in the new subnetworks.

        --# For every network in the new sub internetworks..
        for kLocalNetworkList1, vLocalNetworkList1 in newSubInternetworksTable do
            for kNetwork1, vNetwork1 in vLocalNetworkList1 do
                
                --# Pair up against another one...
                for kLocalNetworkList2, vLocalNetworkList2 in newSubInternetworksTable do
                    for kNetwork2, vNetwork2 in vLocalNetworkList2 do
                        --# If a network gets paired up with itself ...
                        if vNetwork1 == vNetwork2 and 
                           (
                              --# and if they are from different subnetwork lists
                              (kLocalNetworkList1 ~= kLocalNetworkList2) 
                            or 
                              --# or they are from the same subnetwork list 
                              --# but not the same list index
                              (kNetwork1 ~= kNetwork2) 
                           ) 
                           then
                           
                              --# then network tables are not valid
                              --# because of redundancy/duplication.
                              WARN('Invalid network when breaking InterNetwork'
                                   .. ' id='            .. repr(self.NetworkId)
                                   .. ' because networks ' .. repr(vNetwork1)  
                                   .. ' and '           .. repr(vNetwork2) .. ' match.'
                              )
                            return false
                        end
                    end
                end
            end
    	end
        
        --# No unit should appear in more than one subnetwork
    	return true
    end,
        
  
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  This function is called by the function BreakInterNetworkAndCreateNewSubInterNetworks 
    --#*  defined above.  It takes a table where each entry is a list of unit nodes
    --#*  that will form a subnetwork, and it attaches them to a new network manager
    --#*  and gets the new bonuses recalculated and switched on.
    --#**
    SpawnNewInternetworks = function(self, newSubInternetworksTable)
    	
        --# Only allow networks to be rebuilt if the
        --# army's ACU is still alive.
        if not GilbotUtils.GetCommanderFromArmyId(self.Army) then return end
        
        --# For each subnetwork we found...
        for unusedArrayIndex1, localNetworkList in newSubInternetworksTable do
  
            --# Create a new network.
            local newInternetwork = ResourceInterNetwork(nil, 'SpawningFromBroken')
            
            --# Attach all its units
            for unusedArrayIndex2, vLocalNetwork in localNetworkList do
                newInternetwork:AddLocalNetworkToInterNetwork(vLocalNetwork)
            end
            
            --# Try to set the network going
            newInternetwork.IsStillBeingReconstructed = false
            newInternetwork:PropogateBonusesThroughInterNetwork()
    	end
    end,
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  This function is called by the function AddLinkToNetwork defined above. 
    --#* 
    --#*  It is called to merge another network into this one.  
    --#*  The calling code should ensure that the smaller network gets 
    --#*  chosen to be mergeged into the larger one. After the merge, 
    --#*  this ResourceNetwork object takes over the management of all 
    --#*  the units from both networks, and the smaller network is destroyed.
    --#**
    MergeInterNetworks = function(self, smallerInterNetworkArg)
        
        --# This next block is for debugging only.
        --# This can be delete wheh debugging is done.
        if debugAdjacencyCode then
          LOG('Adjacency: MergeInterNetworks:'
           .. ' Merging id=' .. repr(smallerInterNetworkArg.InterNetworkId)
           .. ' into id=' .. repr(self.InterNetworkId)
          )
        end
        
        --# Warn any remote-link pipelines that they need to wait 
        --# before connecting to this internetwork.        
        self.IsStillBeingReconstructed = true
  
        --# Merge the bonus givers lists from the other 
        --# network supervisor into the list for this one.
        local needToRecalculateBonuses = 
            self:MergeNetworkLists(self.AttachedNodesTable, 
                                   smallerInterNetworkArg.AttachedNodesTable)

        --# inform any remote-link pipelines waiting that they don't need 
        --# to wait anymore before connecting to this internetwork.    
        self.IsStillBeingReconstructed = false
        
        --# Propogate bonuses throughout this network if needed
        if needToRecalculateBonuses then 
            self:PropogateBonusesThroughInterNetwork() 
        end
    end,
	


    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  This function is called by the function MergeNetworks defined above. 
    --#*
    --#*  It merges tables that contain references to network objects 
    --#*  and remote-link pipeline objects in each internetwork. 
    --#*  It merges table2 into table1 and updates all items changing internetwork
    --#*  so that they have a reference to the internetwork they just merged into
    --#*  (instead of the old internetwork they just came from which will be destroyed next).
    --#**
    MergeNetworkLists = function(self, table1, table2)
    	--# Nothing to do as table 2 is empty
        if table.getn(table2) == 0 then
            return false
    	end
        --# Table 1 is empty so copy
        --# table 2 into table 1
    	if table.getn(table1) == 0 then
            for k, vUnit in table2 do
                table.insert(table1, vUnit)
            end
            return false
    	end
        
  	--# Have to do a merge.
        --# Iterate through the values to merge.
        local needToRecalculateBonuses = false
    	for k1, vNodeInTable2 in table2 do
        
            local foundInTable1 = false
            --# Iterate through the values to merge into
            --# to see if it's in both tables
            for k2, vNodeInTable1 in table1 do
                if vNodeInTable1 == vNodeInTable2 then
                    foundInTable1 = true
                end
            end
            --# Do the merge if the unit wasn't alredy in both tables
            if not foundInTable1 then
                --# Doing merge.
                --# Network objects and remote link objects both 
                --# get a reference to their new internetwork.
                vNodeInTable2:SetInterNetwork(self)
                
                --# Work out if we will have to recalulate the bonuses on 
                --# all networks after the merege.
                if vNodeInTable2:DoesImpactInterNetworkBonus() then
                    needToRecalculateBonuses = true
                end
                
                --# Give their new internetwork a reference to them
                table.insert(table1, vNodeInTable2)	
            end
    	end
        --# Tell calling code to update bonuses on all 
        --# member networks if necessary
        return needToRecalculateBonuses  	
    end,
    
 
	
        
 
    
    
    --#***************************
    --#*
    --#*  Adding and Removing from Network
    --#*  
    --#***********************************************
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  Add a network or remote link to this internetwork by giving it a reference 
    --#*  to this internetwork, and setting references in this internetwork back to 
    --#*  any subnetwork as a bonus giver or receiver, and to remote links
    --#*  for reconnection purposes.
    --#**
    AddLocalNetworkToInterNetwork = function(self, localNetworkArg)
        --# Perform last bit of network initialisation 
        --# if this is the first node added to the network. 
        if not self.FirstNode then self:SetFirstNode(localNetworkArg) end
        
        --# The added pipeline node needs a reference to its 
        --# new network so it can call us to tell us to update new bonuses
        localNetworkArg:SetInterNetwork(self)
        localNetworkArg.RemoveMeFromMyInterNetwork = false
        
        --# We keep a reference to the network 
        table.insert(self.AttachedNodesTable, localNetworkArg)

    end,
    
    
    
    

    
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  Nodes are either networks or remote-link pipelines.
    --#*  This function removes the node from this internetwork by 
    --#*  removing any references from the node to the internetwork, 
    --#*  and from the internetwork to the node.
    --#**
    RemoveLocalNetworkFromInterNetwork = function(self, localNetworkArg)   
        
        --# If the unit is leaving this network then it isn't in any network.
        if localNetworkArg and localNetworkArg.MyInterNetwork then 
            localNetworkArg:SetInterNetwork(nil) 
        end
        
        --# Remove it from accounting list
        self.AttachedNodesTable = 
            GilbotUtils.RemoveFromArrayByValue(self.AttachedNodesTable, localNetworkArg, true)
        

    end,
    

 
    
    
    --#***************************
    --#*
    --#*  Bonuses
    --#*  
    --#***********************************************
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  Tell all networks to do update. 
    --#*  Networks will call CalulateBonus in this object
    --#*  if they have a reference to it and the internetwork
    --#*  flags that it has at least more than one  
    --#**
    PropogateBonusesThroughInterNetwork = function(self)
    
        --# For each network on the internetwork, Tell them to recalculate
        --# bonuses for all of their bonus consuming nodes.
        --# Calls to CalulateBonus will be passed to this object,
        --# so all adjacency givers on the internetwork will be 
        --# included in the calculation.
        for unusedArrayIndex, vLocalNetwork in self.AttachedNodesTable do
            
            --# This next block is for debugging only.
            --# This can be delete wheh debugging is done.
            if debugAdjacencyCode then
              LOG('Adjacency: PropogateBonusesThroughInterNetwork:' 
                  .. ' On networkId=' .. repr(vLocalNetwork.NetworkId)
                  .. ' calling PropogateBonusesThroughThisNetworkOnly().' 
              )
            end
                    
            vLocalNetwork:PropogateBonusesThroughThisNetworkOnly()
        end
    end,
    
    
    
    --#*
    --#*  Gilbot-X says:
    --#*  
    --#*  Values for bonuses are recalculated from scratch here.
    --#*  This gets called by PropogateBonusesThroughNetwork 
    --#*  each time something is added to the network.
    --#**
    CalulateBonus = function(self, bonusTypeName, bonusConsumer)	
        local modValue = 0
        
        --# Added Safety Check
        if self.IsBroken then return end
        
        --# For each network on the internetwork
        for kNetworkId, vLocalNetwork in self.AttachedNodesTable do
            --# Get each giver on the network to boost up the bonus by its amount
            for k, vUnit in vLocalNetwork.AdjacencyGiversTable[bonusTypeName] do
                --# This safety check is required
                if vUnit and not vUnit:BeenDestroyed() then 
                    --# Add up the bonus modifiers from each giver for this category
                    modValue = modValue + vUnit:GetAdjacentBonus(bonusConsumer, bonusTypeName)
                else                    --# Clean up dead entry in table.
                    self:RemoveUnitFromNetwork(vUnit, true)
                end
                --# Put this in to stop an error that appears when 
                --# lots of buildings are being destroyed.
                if not vLocalNetwork.AdjacencyGiversTable then
                    WARN('ResourceInternetwork.lua: CalulateBonus: Exit loop because '
                      .. 'vLocalNetwork.AdjacencyGiversTable no longer defined.')
                    break
                end
            end
            --# Put this in to stop an error that appears when 
            --# lots of buildings are being destroyed.
            if not self.AttachedNodesTable then
                WARN('ResourceInternetwork.lua: CalulateBonus: Exit loop because '
                  .. 'self.AttachedNodesTable no longer defined.')
                break
            end
        end
        --# Apply cap limit to the bonus (limit depends on bonus type)
        return ApplyBonusCap(modValue, bonusTypeName)
        
    end,
    
}