--#****************************************************************************
--#**
--#**  New File :  /mods/gilbotsimprovements/statslidermenu.lua
--#**
--#**  Author(s):  Gilbot-X
--#**
--#**  Summary  :  Slider controls UI code
--#**  
--#***************************************************************************

local UIUtil = import('/lua/ui/uiutil.lua')
local LayoutHelpers = import('/lua/maui/layouthelpers.lua')
local Slider = import('/lua/maui/slider.lua').Slider
local Button = import('/lua/maui/button.lua').Button
local Bitmap = import('/lua/maui/bitmap.lua').Bitmap
local Tooltip = import('/lua/ui/game/tooltip.lua')

--# My own included files
local myModTextureFolderPath = '/mods/gilbotsimprovements/textures/'
local SupportedTypes = 
    import('/mods/gilbotsimprovements/statsliderdatatypes.lua').SupportedTypes
local GetCustomStatTypesFromUnit = 
    import('/mods/gilbotsimprovements/statsliderdatatypes.lua').GetCustomStatTypesFromUnit

--# These values are constants
local MinSliderValue = 1
local MaxSliderValue = 151
local SliderValueRange = MaxSliderValue-MinSliderValue

local debugThisFile = false


StatSliderMenu = Class { 

    --# These are private variables
    SelectedUnits = nil,
    EntityTables = nil,
    StatSliders = nil,
    StatSliderLabels = nil,
    StatSliderButton = nil,
    LastSlider = nil,
    IsMenuActive = false,


    --#*
    --#*  Constructor function for the class
    --#*  It sets up the menu with sliders, labels and buttons.
    --#**
    __init = function(self, unitsSelected)
        
        --# Reset these tables here just to be safe
        self.StatSliders = {}
        self.StatSliderLabels = {}
        self.EntityTables = {}
        self.EncounteredCustomStatValueTypes = {}
        
        self.xPos = 100
        self.yPos = 100
        self.yOffset = 40
        self.LastSlider = nil
            
        --# Store argument for later use    
        self.SelectedUnits = unitsSelected
        
        for k, selectedUnit in self.SelectedUnits do
        
            --# Check sync table as custom units 
            --# designed for this mod will pass data to use through it. 
            self.AttachAnySynchDataTo(selectedUnit)
   
            --# for each type of stat that can use a slider
            for kType, vType in SupportedTypes do
                if vType:Condition(selectedUnit) then
                    self:AddSliderForStatValue(selectedUnit, vType)
                end
            end
            
            --# for each type of custom stat this unit has
            for kType, vType in GetCustomStatTypesFromUnit(selectedUnit) do
                --# if its the first time we've seen it in this menu instance
                if not self.EncounteredCustomStatValueTypes[kType] then
                    if vType:Condition(selectedUnit) then
                        --# add a slider to the menu for it
                        self:AddSliderForStatValue(selectedUnit, vType)
                        --# keep a record of it
                        self.EncounteredCustomStatValueTypes[kType] = vType
                    end
                end
            end
            
            --# Create a button under the last slider
            if self.LastSlider then 
                self.IsMenuActive = true
                self.StatSliderButton = self:CreateButton(self.xPos+50, self.yPos-10)
                self.StatSliderButton:UseAlphaHitTest(true)
            end
                
        end
    end,
    
    
    --#*
    --#*  Check sync table as custom units 
    --#*  designed for this mod will pass data to use through it. 
    --#**
    AttachAnySynchDataTo = function(selectedUnit)
        if UnitData[selectedUnit:GetEntityId()] then
            selectedUnit.SliderSyncData =
                UnitData[selectedUnit:GetEntityId()].SliderControlledStatValues
            return selectedUnit
        end
    end,
    
    
    
    --#*
    --#*  Add a slider with appropriate parameters 
    --#*  for this stat type and link it to the unit. 
    --#**
    AddSliderForStatValue = function(self, selectedUnit, statDef)
  
        local unitbp = selectedUnit:GetBlueprint()
    
        --# Create the entity table
        if not self.EntityTables[statDef.TypeName] then
            self.EntityTables[statDef.TypeName] = {}
        end
        
        --# Get the variable name for where we will put the value
        --# inside the active unit object (remember we are on UI side here)
        local varName = statDef.VariableNameInUnit
        
        
        --# Record the value of the stat in the unit object,
        --# based on what we think it should be now.
        --# Note: We are doing this from the UI side here,
        --# we need to access SIM side to be sure.
        if not selectedUnit[varName] then
            --# If we'd done it before, leave it as what it was, otherwise 
            --# use the default value
            selectedUnit[varName] = statDef:GetDefaultValue(selectedUnit)
        end
 
        
        --# Check sync data for more up-to-date value of what the stat is now
        if selectedUnit.SliderSyncData then
            if selectedUnit.SliderSyncData[statDef.TypeName] then 
                --# This next line will overwrite what was just obtained in the block above,
                --# i.e. either what the slider menu remembered the value was 
                --# last time it was set here, or the default value if this is the first time
                --# the slider is setting it on this unit.
                selectedUnit[varName] = selectedUnit.SliderSyncData[statDef.TypeName].CurrentValue
            end
        end
        
        --# Put the entity ID and stat value of this unit into a table of changes to make.
        if not table.find(self.EntityTables[statDef.TypeName], selectedUnit:GetEntityId()) then
            self.EntityTables[statDef.TypeName][selectedUnit:GetEntityId()] = 
                selectedUnit[varName]
        end
        
        --# Create a slider for this stat if that hasn't been done already
        if not self.StatSliders[statDef.TypeName] then 
            
            --# A new slider gets a knob position based on the stat that creates it
            local sliderval = self:CalculateSliderValueForStatValue(selectedUnit, 
                selectedUnit[varName], statDef)
            
            --# Create slider and keep a ref to it
            self.StatSliders[statDef.TypeName] = self.CreateSlider(self.xPos, self.yPos, sliderval)
            
            --# Record that so far this is the last/lowest slider in the menu
            self.LastSlider = self.StatSliders[statDef.TypeName]
            
            --# Add a label for it, just above it.
            self.StatSliderLabels[statDef.TypeName] = 
                self.CreateSliderLabel(self.xPos+5, self.yPos-22, statDef.DisplayText)
            
            --# Change layout position for any following slider to be under this one
            self.yPos = self.yPos + self.yOffset
        end
    end, 
                    
    
    
    
    --#*
    --#*  Create slider UI control
    --#**
    CreateSlider = function(left, top, value)
        --# These sliders are used in the options dialog
        parent = import('/lua/ui/game/gamemain.lua').GetGameParent()
        iSlider = Slider(parent, false, MinSliderValue, MaxSliderValue, 
            UIUtil.SkinnableFile('/slider02/slider_btn_up.dds'), 
            UIUtil.SkinnableFile('/slider02/slider_btn_over.dds'),
            UIUtil.SkinnableFile('/slider02/slider_btn_down.dds'),
            UIUtil.SkinnableFile('/slider02/slider-back_bmp.dds'))  
        
        --# Set its position relative to where the mouse was clicked
        iSlider.Left:Set(left)
        iSlider.Top:Set(top)
        
        --# Set the slider position
        iSlider:SetValue(value)
        
        return iSlider
    end,
    
    
    --#*
    --#*  Create label for slider UI control
    --#**
    CreateSliderLabel= function(left, top, textArg)
        --# Get the backgroud bitmap from my mod folder
        parent = import('/lua/ui/game/gamemain.lua').GetGameParent()
        iBitmap = Bitmap(parent, myModTextureFolderPath .. 'label_back.dds')
            
        --# Set its position relative to where the mouse was clicked
        iBitmap.Left:Set(left)
        iBitmap.Top:Set(top)
        
        --# Add text to it
        local bitmapText = UIUtil.CreateText(iBitmap, textArg, 12, UIUtil.bodyFont)
        LayoutHelpers.AtCenterIn(bitmapText, iBitmap)
  
        return iBitmap
    end,
    
    
    --#*
    --#*  Create OK button for menu 
    --#**
    CreateButton = function(self, left, top)
        --# Use bitmaps already in game from 'create prifle' menu
        iButton = Button(self.LastSlider,
            UIUtil.SkinnableFile('/menus02/profile-select_btn_up.dds'), 
            UIUtil.SkinnableFile('/menus02/profile-select_btn_over.dds'),
            UIUtil.SkinnableFile('/menus02/profile-select_btn_down.dds'),
            UIUtil.SkinnableFile('/menus02/profile-select_btn_dis.dds'))  
        
        --# Set its position relative to where the mouse was clicked
        iButton.Left:Set(left)
        iButton.Top:Set(top)
        
        --# Add text
        local buttonText = UIUtil.CreateText(iButton, "OK", 12, UIUtil.bodyFont)
        LayoutHelpers.AtCenterIn(buttonText, iButton)
    
        --# This gives a mouseover highlight effect 
        iButton.HandleEvent = function(control, event)
            if event.Type == 'MouseEnter' then
                --Need to define tooltip text in the tooltip file if you want to have one
                --Tooltip.CreateMouseoverDisplay(iButton, "update stats ", 5, true)
                buttonText:SetColor('ff000000')
            elseif event.Type == 'MouseExit' then
                --Tooltip.DestroyMouseoverDisplay()
                buttonText:SetColor(UIUtil.fontColor())
            end
            Button.HandleEvent(control, event)
        end
    
        --# Call function in our menu class when clicked
        iButton.OwnerMenu = self
        iButton.OnClick = function(self, modifiers)
            self.OwnerMenu:FinishSlider()
        end
        
        return iButton
    end,
    
    
    
    
    --#*
    --#*  This is called after the button is pressed.
    --#*  The user is finished usng the sliders to set new values.
    --#*  We can update the units now.
    --#**
    FinishSlider = function(self)

        --# We are going to call 'update-functions' on all 
        --# relevant units, so first we must iteratively
        --# get 'update-function' arguments for all the units
        for k, selectedUnit in self.SelectedUnits do
        
            --# for each type of stat that can use a slider
            for kType, vType in SupportedTypes do
            
                --# check if this unit has that stat 
                if vType:Condition(selectedUnit) then
                    self:PutValueInUpdatesTable(selectedUnit, vType)
                end
            end

            --# Now do the same for any custom stat types this unit has
            for kType, vType in GetCustomStatTypesFromUnit(selectedUnit) do
                if vType:Condition(selectedUnit) then
                    self:PutValueInUpdatesTable(selectedUnit, vType)
                end
            end
            
        end
        
        --# Now call the update functions!!!
        --# for each type of stat that can use a slider
        for kStatTypeName, vEntityTable in self.EntityTables do

            --# if there was a slider for this type of stat value 
            --# then link to sim via callback to set stat values on unit
            if SupportedTypes[kStatTypeName] then 
                self:DoSimCallback(SupportedTypes[kStatTypeName])
            else
                self:DoSimCallback(self.EncounteredCustomStatValueTypes[kStatTypeName])
            end
        end
        
        --# We are finsihed so destroy the menu
        self:Destroy()
    
    end,
    
    
    
    --#*
    --#* 
    --#*  This is called by the FinishSlider function above.
    --#**
    DoSimCallback = function(self, statDef)
  
        local functionName = 'SetStatValue'
        
        --# Shield size has its own callback function
        --# because the size change is animated
        if statDef.TypeName == 'ShieldSize' then
            functionName = 'SetShieldSize'
        end
      
        --# Invoke sim side code.
        --# Not sure if you can pass a nested table as argument.
        --# You definitely can't pass functions or classes.
        --# You can definitely pass flat tables.
        SimCallback( {  Func=functionName,
                        Args={ 
                          EntityTable= self.EntityTables[statDef.TypeName],
                          StatType = statDef.TypeName,
                          ResourceDrainID = statDef.ResourceDrainID,
                          BPDefaultValueLocation = statDef.BPDefaultValueLocation,
                          BPDefaultValueName = statDef.BPDefaultValueName,
                          UpdateConsumptionImmediately = statDef.UpdateConsumptionImmediately
                        }
                      }
                    )
    end,
    
    
    
    --#*
    --#* 
    --#*  This is called by the FinishSlider function above.
    --#**
    Destroy = function(self)
    
        --# Time to clean up and dispose of the menu.
        --# First get rid of the slider controls and their labels...
        for statName, sliderControl in self.StatSliders do 
            if sliderControl then 
                --# Destroy the slider
                sliderControl:Destroy()
                sliderControl=false
                --# Destroy the label
                self.StatSliderLabels[statName]:Destroy()
                self.StatSliderLabels[statName] = false
                --# Wipe the table
                self.EntityTables[statName] = false
            end
        end
        
        --# Then get rid of the OK button
        self.StatSliderButton:Destroy()
        self.StatSliderButton = false
        
        --# By this point everything should be destroyed!
        self.IsMenuActive = false
        
    end, 
    
    
    
    
    --#*
    --#* 
    --#*  This is called by the FinishSlider function above.
    --#**
    PutValueInUpdatesTable = function(self, selectedUnit, statDef)
          --# Get the variable name for where we will put the stat value
          --# inside the active unit object (remember we are on UI side here)
          local varName = statDef.VariableNameInUnit
          
          --# Get the new stat value the user set with the slider
          selectedUnit[varName] = 
              self:CalculateStatValueFromSlider(selectedUnit, 
                                                self.StatSliders[statDef.TypeName]:GetValue(), 
                                                statDef)
          
          --# Update our table so we can synchronise all units after this
          self.EntityTables[statDef.TypeName][selectedUnit:GetEntityId()] = selectedUnit[varName] 
    end,
    

    

    --#*
    --#*  Work out the starting position of the slider control's knob
    --#*  from last shield size, intel radius, or other stat value in the unit.
    --#*  This is called by the class constructor 
    --#**
    CalculateSliderValueForStatValue = function(self, selectedUnit, statValue, statDef)
        
        --# Get min and max values for this stat
        local minStatValue = statDef:GetMinValue(selectedUnit) or 0
        local maxStatValue = statDef:GetMaxValue(selectedUnit) or 0
        
        --# Perform scaling and return calculated result
        local sliderValuesPerStatUnit = SliderValueRange / (maxStatValue-minStatValue) 
        local sliderValue = MinSliderValue + (sliderValuesPerStatUnit * (statValue - minStatValue))
        return sliderValue
    end,
    
    
    
    --#*
    --#*  Get shield size, intel radius, or other stat value, from the slider
    --#*  This is called by FinishSlider.
    --#**
    CalculateStatValueFromSlider = function(self, selectedUnit, currentSliderValue, statDef)
        --# Get min and max values for this stat
        local minStatValue = statDef:GetMinValue(selectedUnit) or 0
        local maxStatValue = statDef:GetMaxValue(selectedUnit) or 0
        
        --# Perform scaling and return calculated result
        local statUnitsPerSliderUnits = (maxStatValue-minStatValue) / SliderValueRange
        local statValueFromSlider = minStatValue + 
            (statUnitsPerSliderUnits * (currentSliderValue - MinSliderValue))

        return statValueFromSlider
    end,
    
}
